@@ -3,7 +3,7 @@
  *
  * handles debugfs commands
  *
- * Copyright (C) 2004 Oracle.  All rights reserved.
+ * Copyright (C) 2004, 2008 Oracle.  All rights reserved.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
@@ -20,10 +20,12 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 021110-1307, USA.
  *
- * Authors: Sunil Mushran, Manish Singh
  */
 
-#include <main.h>
+#include "main.h"
+#include "ocfs2/image.h"
+#include "ocfs2/byteorder.h"
+
 #define SYSTEM_FILE_NAME_MAX	40
 #define MAX_BLOCKS		50
 
@@ -67,6 +69,7 @@ static void do_locate (char **args);
 static void do_fs_locks (char **args);
 static void do_bmap (char **args);
 static void do_icheck (char **args);
+static void do_dlm_locks (char **args);
 
 dbgfs_gbls gbls;
 
@@ -77,6 +80,7 @@ static Command commands[] = {
 	{ "chroot",	do_chroot },
 	{ "close",	do_close },
 	{ "curdev",	do_curdev },
+	{ "dlm_locks",	do_dlm_locks },
 	{ "dump",	do_dump },
 	{ "extent",	do_extent },
 	{ "fs_locks",	do_fs_locks },
@@ -297,7 +301,8 @@ static int process_inodestr_args(char **args, int count, uint64_t *blkno)
 /* open the device, read the block from the device and get the
  * blocksize from the offset of the ocfs2_super_block.
  */
-static errcode_t get_blocksize(char* dev, uint64_t offset, uint64_t *blocksize)
+static errcode_t get_blocksize(char* dev, uint64_t offset, uint64_t *blocksize,
+			       int super_no)
 {
 	errcode_t ret;
 	uint64_t blkno;
@@ -305,6 +310,7 @@ static errcode_t get_blocksize(char* dev, uint64_t offset, uint64_t *blocksize)
 	char *buf = NULL;
 	io_channel* channel = NULL;
 	struct ocfs2_dinode *di = NULL;
+	struct ocfs2_image_hdr *hdr;
 
 	ret = io_open(dev, OCFS2_FLAG_RO, &channel);
 	if (ret)
@@ -322,6 +328,19 @@ static errcode_t get_blocksize(char* dev, uint64_t offset, uint64_t *blocksize)
 	if (ret)
 		goto bail;
 
+	if (gbls.imagefile) {
+		ret = io_read_block(channel, 0, 1, buf);
+		if (ret)
+			goto bail;
+		hdr = (struct ocfs2_image_hdr *)buf;
+		ocfs2_image_swap_header(hdr);
+		if (super_no > hdr->hdr_superblkcnt) {
+			ret = OCFS2_ET_IO;
+			goto bail;
+		}
+		offset = hdr->hdr_superblocks[super_no-1] * hdr->hdr_fsblksz;
+	}
+
 	blkno = offset / OCFS2_MIN_BLOCKSIZE;
 	ret = io_read_block(channel, blkno, 1, buf);
 	if (ret)
@@ -345,35 +364,44 @@ static int process_open_args(char **args,
 			     uint64_t *superblock, uint64_t *blocksize)
 {
 	errcode_t ret = 0;
-	uint32_t s;
-	char *ptr;
-	uint64_t byte_off[OCFS2_MAX_BACKUP_SUPERBLOCKS], blksize;
-	int num, ind = 2;
+	uint32_t s = 0;
+	char *ptr, *dev;
+	uint64_t byte_off[OCFS2_MAX_BACKUP_SUPERBLOCKS];
+	uint64_t blksize = 0;
+	int num, argc, c;
 
-	if (!args[ind])
+	for (argc = 0; (args[argc]); ++argc);
+	dev = strdup(args[1]);
+	optind = 0;
+	while ((c = getopt(argc, args, "is:")) != EOF) {
+		switch (c) {
+			case 'i':
+				gbls.imagefile = 1;
+				break;
+			case 's':
+				s = strtoul(optarg, &ptr, 0);
+				break;
+			default:
+				return 1;
+				break;
+		}
+	}
+
+	if (!s)
 		return 0;
 
-	if (args[ind] && !strcmp(args[ind], "-s"))
-		ind++;
-	else
-		return -1;
-
-	if(!args[ind])
-		return -1;
-
-	num = ocfs2_get_backup_super_offset(NULL,
-					    byte_off, ARRAY_SIZE(byte_off));
+	num = ocfs2_get_backup_super_offsets(NULL, byte_off,
+					     ARRAY_SIZE(byte_off));
 	if (!num)
 		return -1;
 
-	s = strtoul(args[ind], &ptr, 0);
 	if (s < 1 || s > num) {
 		fprintf (stderr, "Backup super block is outside of valid range"
 			 "(between 1 and %d)\n", num);
 		return -1;
 	}
 
-	ret = get_blocksize(args[1], byte_off[s-1], &blksize);
+	ret = get_blocksize(dev, byte_off[s-1], &blksize, s);
 	if (ret) {
 		com_err(args[0],ret, "Can't get the blocksize from the device"
 			" by the num %u\n", s);
@@ -426,21 +454,34 @@ static errcode_t find_block_offset(ocfs2_filesys *fs,
 	errcode_t ret = 0;
 	char *buf = NULL;
 	int i;
-	uint32_t clstoff;
+	uint32_t clstoff, clusters;
 	uint32_t tmp;
 
 	clstoff = ocfs2_blocks_to_clusters(fs, blkoff);
 
 	for (i = 0; i < el->l_next_free_rec; ++i) {
 		rec = &(el->l_recs[i]);
+		clusters = ocfs2_rec_clusters(el->l_tree_depth, rec);
 
-		/* TODO Fix to handle sparse trees */
-		if (clstoff >= (rec->e_cpos + rec->e_clusters))
+		/*
+		 * For a sparse file, we may find an empty record.
+		 * Just skip it.
+		 */
+		if (!clusters)
+			continue;
+
+		if (clstoff >= (rec->e_cpos + clusters))
 			continue;
 
 		if (!el->l_tree_depth) {
-			tmp = blkoff - ocfs2_clusters_to_blocks(fs, clstoff);
-			dump_logical_blkno(out, rec->e_blkno + tmp);
+			if (clstoff < rec->e_cpos) {
+				dump_logical_blkno(out, 0);
+			} else {
+				tmp = blkoff -
+					ocfs2_clusters_to_blocks(fs,
+								 rec->e_cpos);
+				dump_logical_blkno(out, rec->e_blkno + tmp);
+			}
 			goto bail;
 		}
 
@@ -453,7 +494,7 @@ static errcode_t find_block_offset(ocfs2_filesys *fs,
 		ret = ocfs2_read_extent_block(fs, rec->e_blkno, buf);
 		if (ret) {
 			com_err(gbls.cmd, ret, "while reading extent %"PRIu64,
-				rec->e_blkno);
+				(uint64_t)rec->e_blkno);
 			goto bail;
 		}
 
@@ -482,11 +523,23 @@ static errcode_t traverse_extents (ocfs2_filesys *fs, struct ocfs2_extent_list *
 	errcode_t ret = 0;
 	char *buf = NULL;
 	int i;
+	uint32_t clusters;
 
 	dump_extent_list (out, el);
 
 	for (i = 0; i < el->l_next_free_rec; ++i) {
 		rec = &(el->l_recs[i]);
+		clusters = ocfs2_rec_clusters(el->l_tree_depth, rec);
+
+		/*
+		 * In a unsuccessful insertion, we may shift a tree
+		 * add a new branch for it and do no insertion. So we
+		 * may meet a extent block which have
+		 * clusters == 0, this should only be happen
+		 * in the last extent rec. */
+		if (!clusters && i == el->l_next_free_rec - 1)
+			break;
+
 		if (el->l_tree_depth) {
 			ret = ocfs2_malloc_block(gbls.fs->fs_io, &buf);
 			if (ret)
@@ -574,15 +627,20 @@ static void do_open (char **args)
 		do_close (NULL);
 
 	if (dev == NULL || process_open_args(args, &superblock, &block_size)) {
-		fprintf (stderr, "usage: %s <device> [-s num]\n", args[0]);
+		fprintf (stderr, "usage: %s <device> [-i] [-s num]\n", args[0]);
+		gbls.imagefile = 0;
 		return ;
 	}
 
 	flags = gbls.allow_write ? OCFS2_FLAG_RW : OCFS2_FLAG_RO;
         flags |= OCFS2_FLAG_HEARTBEAT_DEV_OK;
+	if (gbls.imagefile)
+		flags |= OCFS2_FLAG_IMAGE_FILE;
+
 	ret = ocfs2_open(dev, flags, superblock, block_size, &gbls.fs);
 	if (ret) {
 		gbls.fs = NULL;
+		gbls.imagefile = 0;
 		com_err(args[0], ret, "while opening context for device %s",
 			dev);
 		return ;
@@ -653,6 +711,7 @@ static void do_close (char **args)
 	if (ret)
 		com_err(args[0], ret, "while closing context");
 	gbls.fs = NULL;
+	gbls.imagefile = 0;
 
 	if (gbls.blockbuf)
 		ocfs2_free(&gbls.blockbuf);
@@ -767,11 +826,12 @@ static void do_help (char **args)
 	printf ("close\t\t\t\t\tClose a device\n");
 	printf ("curdev\t\t\t\t\tShow current device\n");
 	printf ("decode <lockname#> ...\t\t\tDecode block#(s) from the lockname(s)\n");
+	printf ("dlm_locks [-l] lockname\t\t\tShow live dlm locking state\n");
 	printf ("dump [-p] <filespec> <outfile>\t\tDumps file to outfile on a mounted fs\n");
 	printf ("encode <filespec>\t\t\tShow lock name\n");
 	printf ("extent <block#>\t\t\t\tShow extent block\n");
 	printf ("findpath <block#>\t\t\tList one pathname of the inode/lockname\n");
-	printf ("fs_locks [-l]\t\t\t\tShow live fs locking state\n");
+	printf ("fs_locks [-l] [-B]\t\t\tShow live fs locking state\n");
 	printf ("group <block#>\t\t\t\tShow chain group\n");
 	printf ("hb\t\t\t\t\tShows the used heartbeat blocks\n");
 	printf ("help, ?\t\t\t\t\tThis information\n");
@@ -781,7 +841,7 @@ static void do_help (char **args)
 	printf ("logdump <slot#>\t\t\t\tPrints journal file for the node slot\n");
 	printf ("ls [-l] <filespec>\t\t\tList directory\n");
 	printf ("ncheck <block#> ...\t\t\tList all pathnames of the inode(s)/lockname(s)\n");
-	printf ("open <device> [-s backup#]\t\t\t\tOpen a device\n");
+	printf ("open <device> [-i] [-s backup#]\t\tOpen a device\n");
 	printf ("quit, q\t\t\t\t\tExit the program\n");
 	printf ("rdump [-v] <filespec> <outdir>\t\tRecursively dumps from src to a dir on a mounted filesystem\n");
 	printf ("slotmap\t\t\t\t\tShow slot map\n");
@@ -935,7 +995,7 @@ static void do_stat (char **args)
 	else if ((inode->i_flags & OCFS2_CHAIN_FL))
 		ret = traverse_chains(gbls.fs, &(inode->id2.i_chain), out);
 	else if (S_ISLNK(inode->i_mode) && !inode->i_clusters)
-		dump_fast_symlink(out, inode->id2.i_symlink);
+		dump_fast_symlink(out, (char *)inode->id2.i_symlink);
 	else if (inode->i_flags & OCFS2_DEALLOC_FL)
 		dump_truncate_log(out, &(inode->id2.i_dealloc));
 	else
@@ -1175,26 +1235,30 @@ static void do_slotmap (char **args)
 {
 	FILE *out;
 	errcode_t ret;
-	char *buf = NULL;
-	uint32_t len = gbls.fs->fs_blocksize;
+	int num_slots;
+	struct ocfs2_slot_map_extended *se = NULL;
+	struct ocfs2_slot_map *sm = NULL;
 
 	if (check_device_open())
 		return ;
 
-	/* read in the first block of the slot_map file */
-	ret = read_whole_file(gbls.fs, gbls.slotmap_blkno, &buf, &len);
+	num_slots = OCFS2_RAW_SB(gbls.fs->fs_super)->s_max_slots;
+	if (ocfs2_uses_extended_slot_map(OCFS2_RAW_SB(gbls.fs->fs_super)))
+		ret = ocfs2_read_slot_map_extended(gbls.fs, num_slots, &se);
+	else
+		ret = ocfs2_read_slot_map(gbls.fs, num_slots, &sm);
 	if (ret) {
 		com_err(args[0], ret, "while reading slotmap system file");
 		goto bail;
 	}
 
 	out = open_pager(gbls.interactive);
-	dump_slots (out, buf, len);
+	dump_slots (out, se, sm, num_slots);
 	close_pager (out);
 
 bail:
-	if (buf)
-		ocfs2_free(&buf);
+	if (sm)
+		ocfs2_free(&sm);
 
 	return ;
 }
@@ -1382,6 +1446,54 @@ static void do_locate(char **args)
 	find_inode_paths(gbls.fs, args, findall, count, blkno, stdout);
 }
 
+/*
+ * do_dlm_locks()
+ *
+ */
+static void do_dlm_locks(char **args)
+{
+	FILE *out;
+	int dump_lvbs = 0;
+	int i;
+	struct locknames *lock;
+	struct list_head locklist;
+	struct list_head *iter, *iter2;
+
+	if (check_device_open())
+		return;
+
+	INIT_LIST_HEAD(&locklist);
+
+	i = 1;
+	if (args[i] && strlen(args[i])) {
+		if (!strcmp("-l", args[i])) {
+			dump_lvbs = 1;
+			i++;
+		}
+
+		for ( ; args[i] && strlen(args[i]); ++i) {
+			lock = calloc(1, sizeof(struct locknames));
+			if (!lock)
+				break;
+			INIT_LIST_HEAD(&lock->list);
+			strncpy(lock->name, args[i], sizeof(lock->name));
+			list_add_tail(&lock->list, &locklist);
+		}
+	}
+
+	out = open_pager(gbls.interactive);
+	dump_dlm_locks(gbls.fs->uuid_str, out, dump_lvbs, &locklist);
+	close_pager(out);
+
+	if (!list_empty(&locklist)) {
+		list_for_each_safe(iter, iter2, &locklist) {
+			lock = list_entry(iter, struct locknames, list);
+			list_del(iter);
+			free(lock);
+		}
+	}
+}
+
 /*
  * do_dump_fs_locks()
  *
@@ -1390,15 +1502,21 @@ static void do_fs_locks(char **args)
 {
 	FILE *out;
 	int dump_lvbs = 0;
+	int only_busy = 0;
+	int i = 0;
 
 	if (check_device_open())
 		return;
 
-	if (args[1] && !strcmp("-l", args[1]))
-		dump_lvbs = 1;
+	while (args[++i]) {
+		if (!strcmp("-l", args[i]))
+			dump_lvbs = 1;
+		else if (!strcmp("-B", args[i]))
+			only_busy = 1;
+	}
 
 	out = open_pager(gbls.interactive);
-	dump_fs_locks(gbls.fs->uuid_str, out, dump_lvbs);
+	dump_fs_locks(gbls.fs->uuid_str, out, dump_lvbs, only_busy);
 	close_pager(out);
 }
 
