@@ -23,9 +23,11 @@
  * Authors: Sunil Mushran
  */
 
-#include <main.h>
 #include <stdint.h>
 
+#include "main.h"
+#include "ocfs2/byteorder.h"
+
 extern dbgfs_gbls gbls;
 
 /*
@@ -39,10 +41,13 @@ void dump_super_block(FILE *out, struct ocfs2_super_block *sb)
 	GString *compat = NULL;
 	GString *incompat = NULL;
 	GString *rocompat = NULL;
+	GString *tunefs_flag = NULL;
+	time_t lastcheck;
 
 	compat = g_string_new(NULL);
 	incompat = g_string_new(NULL);
 	rocompat = g_string_new(NULL);
+	tunefs_flag = g_string_new(NULL);
 
 	fprintf(out, "\tRevision: %u.%u\n", sb->s_major_rev_level, sb->s_minor_rev_level);
 	fprintf(out, "\tMount Count: %u   Max Mount Count: %u\n", sb->s_mnt_count,
@@ -50,28 +55,33 @@ void dump_super_block(FILE *out, struct ocfs2_super_block *sb)
 
 	fprintf(out, "\tState: %u   Errors: %u\n", sb->s_state, sb->s_errors);
 
-	str = ctime((time_t*)&sb->s_lastcheck);
+	lastcheck = (time_t)sb->s_lastcheck;
+	str = ctime(&lastcheck);
 	fprintf(out, "\tCheck Interval: %u   Last Check: %s", sb->s_checkinterval, str);
 
 	fprintf(out, "\tCreator OS: %u\n", sb->s_creator_os);
 
 	get_compat_flag(sb->s_feature_compat, compat);
 	get_incompat_flag(sb->s_feature_incompat, incompat);
+	get_tunefs_flag(sb->s_feature_incompat,
+			sb->s_tunefs_flag, tunefs_flag);
 	get_rocompat_flag(sb->s_feature_ro_compat, rocompat);
 
 	fprintf(out, "\tFeature Compat: %u %s\n", sb->s_feature_compat,
 		compat->str);
 	fprintf(out, "\tFeature Incompat: %u %s\n", sb->s_feature_incompat,
 		incompat->str);
+	fprintf(out, "\tTunefs Incomplete: %u %s\n", sb->s_tunefs_flag,
+		tunefs_flag->str);
 	fprintf(out, "\tFeature RO compat: %u %s\n", sb->s_feature_ro_compat,
 		rocompat->str);
 
 	fprintf(out, "\tRoot Blknum: %"PRIu64"   System Dir Blknum: %"PRIu64"\n",
-		sb->s_root_blkno,
-		sb->s_system_dir_blkno);
+		(uint64_t)sb->s_root_blkno,
+		(uint64_t)sb->s_system_dir_blkno);
 
 	fprintf(out, "\tFirst Cluster Group Blknum: %"PRIu64"\n",
-		sb->s_first_cluster_group);
+		(uint64_t)sb->s_first_cluster_group);
 
 	fprintf(out, "\tBlock Size Bits: %u   Cluster Size Bits: %u\n",
 	       sb->s_blocksize_bits, sb->s_clustersize_bits);
@@ -83,6 +93,14 @@ void dump_super_block(FILE *out, struct ocfs2_super_block *sb)
 	for (i = 0; i < 16; i++)
 		fprintf(out, "%02X", sb->s_uuid[i]);
 	fprintf(out, "\n");
+	if (ocfs2_userspace_stack(sb))
+		fprintf(out,
+			"\tCluster stack: %s\n"
+			"\tCluster name: %s\n",
+			sb->s_cluster_info.ci_stack,
+			sb->s_cluster_info.ci_cluster);
+	else
+		fprintf(out, "\tCluster stack: classic o2cb\n");
 
 	g_string_free(compat, 1);
 	g_string_free(incompat, 1);
@@ -148,6 +166,7 @@ void dump_inode(FILE *out, struct ocfs2_dinode *in)
 	uint16_t mode;
 	GString *flags = NULL;
 	char tmp_str[30];
+	time_t tm;
 
 	if (S_ISREG(in->i_mode))
 		str = "Regular";
@@ -195,7 +214,7 @@ void dump_inode(FILE *out, struct ocfs2_dinode *in)
 		g_string_append (flags, "Dealloc ");
 
 	fprintf(out, "\tInode: %"PRIu64"   Mode: 0%0o   Generation: %u (0x%x)\n",
-	        in->i_blkno, mode, in->i_generation, in->i_generation);
+		(uint64_t)in->i_blkno, mode, in->i_generation, in->i_generation);
 
 	fprintf(out, "\tFS Generation: %u (0x%x)\n", in->i_fs_generation,
 		in->i_fs_generation);
@@ -208,18 +227,18 @@ void dump_inode(FILE *out, struct ocfs2_dinode *in)
 	fprintf(out, "\tUser: %d (%s)   Group: %d (%s)   Size: %"PRIu64"\n",
 	       in->i_uid, (pw ? pw->pw_name : "unknown"),
 	       in->i_gid, (gr ? gr->gr_name : "unknown"),
-	       in->i_size);
+	       (uint64_t)in->i_size);
 
 	fprintf(out, "\tLinks: %u   Clusters: %u\n", in->i_links_count, in->i_clusters);
 
-	str = ctime((time_t*)&in->i_ctime);
-	fprintf(out, "\tctime: 0x%"PRIx64" -- %s", in->i_ctime, str);
-	str = ctime((time_t*)&in->i_atime);
-	fprintf(out, "\tatime: 0x%"PRIx64" -- %s", in->i_atime, str);
-	str = ctime((time_t*)&in->i_mtime);
-	fprintf(out, "\tmtime: 0x%"PRIx64" -- %s", in->i_mtime, str);
-	str = ctime((time_t*)&in->i_dtime);
-	fprintf(out, "\tdtime: 0x%"PRIx64" -- %s", in->i_dtime, str);
+	tm = (time_t)in->i_ctime;
+	fprintf(out, "\tctime: 0x%"PRIx64" -- %s", (uint64_t)tm, ctime(&tm));
+	tm = (time_t)in->i_atime;
+	fprintf(out, "\tatime: 0x%"PRIx64" -- %s", (uint64_t)tm, ctime(&tm));
+	tm = (time_t)in->i_mtime;
+	fprintf(out, "\tmtime: 0x%"PRIx64" -- %s", (uint64_t)tm, ctime(&tm));
+	tm = (time_t)in->i_dtime;
+	fprintf(out, "\tdtime: 0x%"PRIx64" -- %s", (uint64_t)tm, ctime(&tm));
 
 	fprintf(out, "\tctime_nsec: 0x%08"PRIx32" -- %u\n",
 		in->i_ctime_nsec, in->i_ctime_nsec);
@@ -228,7 +247,7 @@ void dump_inode(FILE *out, struct ocfs2_dinode *in)
 	fprintf(out, "\tmtime_nsec: 0x%08"PRIx32" -- %u\n",
 		in->i_mtime_nsec, in->i_mtime_nsec);
 
-	fprintf(out, "\tLast Extblk: %"PRIu64"\n", in->i_last_eb_blk);
+	fprintf(out, "\tLast Extblk: %"PRIu64"\n", (uint64_t)in->i_last_eb_blk);
 	if (in->i_suballoc_slot == (uint16_t)OCFS2_INVALID_SLOT)
 		strcpy(tmp_str, "Global");
 	else
@@ -246,6 +265,8 @@ void dump_inode(FILE *out, struct ocfs2_dinode *in)
 		if (in->id1.journal1.ij_flags & OCFS2_JOURNAL_DIRTY_FL)
 			fprintf(out, "Dirty ");
 		fprintf(out, "\n");
+		fprintf(out, "\tRecovery Generation: %u\n",
+			in->id1.journal1.ij_recovery_generation);
 	}
 
 	if (flags)
@@ -279,7 +300,7 @@ void dump_chain_list (FILE *out, struct ocfs2_chain_list *cl)
 		rec = &(cl->cl_recs[i]);
 		fprintf(out, "\t%-2d   %-10u   %-10u   %-10u   %"PRIu64"\n",
 			i, rec->c_total, (rec->c_total - rec->c_free),
-			rec->c_free, rec->c_blkno);
+			rec->c_free, (uint64_t)rec->c_blkno);
 	}
 
 bail:
@@ -290,6 +311,7 @@ void dump_extent_list (FILE *out, struct ocfs2_extent_list *ext)
 {
 	struct ocfs2_extent_rec *rec;
 	int i;
+	uint32_t clusters;
 
 	fprintf(out, "\tTree Depth: %u   Count: %u   Next Free Rec: %u\n",
 		ext->l_tree_depth, ext->l_count, ext->l_next_free_rec);
@@ -297,12 +319,26 @@ void dump_extent_list (FILE *out, struct ocfs2_extent_list *ext)
 	if (!ext->l_next_free_rec)
 		goto bail;
 
-	fprintf(out, "\t## %-11s   %-12s   %-s\n", "Offset", "Clusters", "Block#");
+	if (ext->l_tree_depth)
+		fprintf(out, "\t## %-11s   %-12s   %-s\n", "Offset",
+			"Clusters", "Block#");
+	else
+		fprintf(out, "\t## %-11s   %-12s   %-13s   %s\n", "Offset",
+			"Clusters", "Block#", "Flags");
 
 	for (i = 0; i < ext->l_next_free_rec; ++i) {
 		rec = &(ext->l_recs[i]);
-		fprintf(out, "\t%-2d %-11u   %-12u   %"PRIu64"\n",
-		       	i, rec->e_cpos, rec->e_clusters, rec->e_blkno);
+		clusters = ocfs2_rec_clusters(ext->l_tree_depth, rec);
+
+		if (ext->l_tree_depth)
+			fprintf(out, "\t%-2d %-11u   %-12u   %"PRIu64"\n",
+				i, rec->e_cpos, clusters,
+				(uint64_t)rec->e_blkno);
+		else
+			fprintf(out,
+				"\t%-2d %-11u   %-12u   %-13"PRIu64"   0x%x\n",
+				i, rec->e_cpos, clusters,
+				(uint64_t)rec->e_blkno,	rec->e_flags);
 	}
 
 bail:
@@ -319,7 +355,7 @@ void dump_extent_block (FILE *out, struct ocfs2_extent_block *blk)
 		 blk->h_suballoc_bit, blk->h_suballoc_slot);
 
 	fprintf (out, "\tBlknum: %"PRIu64"   Next Leaf: %"PRIu64"\n",
-		 blk->h_blkno, blk->h_next_leaf_blk);
+		 (uint64_t)blk->h_blkno, (uint64_t)blk->h_next_leaf_blk);
 
 	return ;
 }
@@ -337,7 +373,7 @@ void dump_group_descriptor (FILE *out, struct ocfs2_group_desc *grp,
 		fprintf (out, "\tGroup Chain: %u   Parent Inode: %"PRIu64"  "
 			 "Generation: %u\n",
 			 grp->bg_chain,
-			 grp->bg_parent_dinode,
+			 (uint64_t)grp->bg_parent_dinode,
 			 grp->bg_generation);
 		fprintf(out, "\t##   %-15s   %-6s   %-6s   %-6s   %-6s   %-6s\n",
 			"Block#", "Total", "Used", "Free", "Contig", "Size");
@@ -346,7 +382,7 @@ void dump_group_descriptor (FILE *out, struct ocfs2_group_desc *grp,
 	find_max_contig_free_bits(grp, &max_contig_free_bits);
 
 	fprintf(out, "\t%-2d   %-15"PRIu64"   %-6u   %-6u   %-6u   %-6u   %-6u\n",
-		index, grp->bg_blkno, grp->bg_bits,
+		index, (uint64_t)grp->bg_blkno, grp->bg_bits,
 		(grp->bg_bits - grp->bg_free_bits_count),
 		grp->bg_free_bits_count, max_contig_free_bits, grp->bg_size);
 
@@ -369,7 +405,8 @@ int  dump_dir_entry (struct ocfs2_dir_entry *rec, int offset, int blocksize,
 	rec->name[rec->name_len] = '\0';
 
 	if (!ls->long_opt) {
-		fprintf(ls->out, "\t%-15"PRIu64" %-4u %-4u %-2u %s\n", rec->inode,
+		fprintf(ls->out, "\t%-15"PRIu64" %-4u %-4u %-2u %s\n",
+			(uint64_t)rec->inode,
 			rec->rec_len, rec->name_len, rec->file_type, rec->name);
 	} else {
 		memset(ls->buf, 0, ls->fs->fs_blocksize);
@@ -380,8 +417,9 @@ int  dump_dir_entry (struct ocfs2_dir_entry *rec, int offset, int blocksize,
 		inode_time_to_str(di->i_mtime, timestr, sizeof(timestr));
 
 		fprintf(ls->out, "\t%-15"PRIu64" %10s %3u %5u %5u %15"PRIu64" %s %s\n",
-		       	rec->inode, perms, di->i_links_count, di->i_uid, di->i_gid,
-			di->i_size, timestr, rec->name);
+			(uint64_t)rec->inode, perms, di->i_links_count,
+			di->i_uid, di->i_gid,
+			(uint64_t)di->i_size, timestr, rec->name);
 	}
 
 	rec->name[rec->name_len] = tmp;
@@ -585,23 +623,27 @@ void dump_jbd_unknown (FILE *out, uint64_t start, uint64_t end)
  * dump_slots()
  *
  */
-void dump_slots (FILE *out, char *buf, uint32_t len)
+void dump_slots (FILE *out, struct ocfs2_slot_map_extended *se,
+                 struct ocfs2_slot_map *sm, int num_slots)
 {
-	int16_t *slots = (int16_t *)buf;
-	uint32_t i;
-	uint32_t num_slots = (len / sizeof(uint16_t));
+	int i;
+        unsigned int node_num;
 	
 	fprintf (out, "\t%5s   %5s\n", "Slot#", "Node#");
 	
 	for (i = 0; i < num_slots; ++i) {
-		uint16_t slot = le16_to_cpu(slots[i]);
-		if (slot == (uint16_t)OCFS2_INVALID_SLOT)
-			continue;
+		if (se) {
+			if (!se->se_slots[i].es_valid)
+				continue;
+			node_num = se->se_slots[i].es_node_num;
+		} else {
+			if (sm->sm_slots[i] == (uint16_t)OCFS2_INVALID_SLOT)
+				continue;
+			node_num = sm->sm_slots[i];
+		}
 
-		fprintf (out, "\t%5d   %5u\n", i, slot);
+		fprintf (out, "\t%5d   %5u\n", i, node_num);
 	}
-
-	return ;
 }
 
 void dump_hb (FILE *out, char *buf, uint32_t len)
@@ -618,8 +660,8 @@ void dump_hb (FILE *out, char *buf, uint32_t len)
 		if (hb->hb_seq)
 			fprintf (out, "\t%4u: %4u %016"PRIx64" %016"PRIx64" "
 				 "%08"PRIx32"\n", i,
-				 hb->hb_node, hb->hb_seq, hb->hb_generation,
-				 hb->hb_cksum);
+				 hb->hb_node, (uint64_t)hb->hb_seq,
+				 (uint64_t)hb->hb_generation, hb->hb_cksum);
 	}
 
 	return ;
