@@ -48,7 +48,8 @@ void dump_jbd_superblock (FILE *out, journal_superblock_t *jsb);
 void dump_jbd_block (FILE *out, journal_header_t *header, uint64_t blknum);
 void dump_jbd_metadata (FILE *out, int type, char *buf, uint64_t blknum);
 void dump_jbd_unknown (FILE *out, uint64_t start, uint64_t end);
-void dump_slots (FILE *out, char *buf, uint32_t len);
+void dump_slots (FILE *out, struct ocfs2_slot_map_extended *se,
+                 struct ocfs2_slot_map *sm, int num_slots);
 void dump_fast_symlink (FILE *out, char *link);
 void dump_hb (FILE *out, char *buf, uint32_t len);
 void dump_inode_path (FILE *out, uint64_t blkno, char *path);
