@@ -3,7 +3,7 @@
  *
  * Function prototypes, macros, etc. for related 'C' files
  *
- * Copyright (C) 2004 Oracle.  All rights reserved.
+ * Copyright (C) 2004, 2008 Oracle.  All rights reserved.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
@@ -20,7 +20,6 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 021110-1307, USA.
  *
- * Authors: Sunil Mushran
  */
 
 #ifndef __MAIN_H__
@@ -34,6 +33,7 @@
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/statfs.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <pwd.h>
@@ -56,9 +56,8 @@
 
 #include <linux/types.h>
 
-#include "ocfs2.h"
-#include "ocfs2_fs.h"
-#include "ocfs1_fs_compat.h"
+#include "ocfs2/ocfs2.h"
+#include "ocfs2-kernel/ocfs1_fs_compat.h"
 
 enum {
 	CONFIG,
@@ -69,6 +68,7 @@ enum {
 typedef struct _dbgfs_glbs {
 	char *progname;
 	int allow_write;
+	int imagefile;
 	int interactive;
 	char *device;
 	ocfs2_filesys *fs;
@@ -87,6 +87,7 @@ typedef struct _dbgfs_glbs {
 
 typedef struct _dbgfs_opts {
 	int allow_write;
+	int imagefile;
 	int no_prompt;
 	uint32_t sb_num;
 	char *cmd_file;
@@ -156,12 +157,13 @@ typedef struct _dbgfs_opts {
 
 /* remaining headers */
 #include <commands.h>
-#include <kernel-list.h>
+#include "ocfs2-kernel/kernel-list.h"
 #include <utils.h>
 #include <journal.h>
 #include <find_block_inode.h>
 #include <find_inode_paths.h>
 #include <dump_fs_locks.h>
+#include <dump_dlm_locks.h>
 #include <dump.h>
 
 #endif		/* __MAIN_H__ */
