@@ -23,7 +23,7 @@
  * Authors: Sunil Mushran, Manish Singh
  */
 
-#include <main.h>
+#include "main.h"
 #include <sys/types.h>
 #include <dirent.h>
 
@@ -47,10 +47,11 @@ static void usage (char *progname)
 	g_print ("usage: %s -l [<logentry> ... [allow|off|deny]] ...\n", progname);
 	g_print ("usage: %s -d, --decode <lockres>\n", progname);
 	g_print ("usage: %s -e, --encode <lock type> <block num> <generation>\n", progname);
-	g_print ("usage: %s [-f cmdfile] [-R request] [-s backup#] [-V] [-w] [-n] [-?] [device]\n", progname);
+	g_print ("usage: %s [-f cmdfile] [-R request] [-i] [-s backup#] [-V] [-w] [-n] [-?] [device]\n", progname);
 	g_print ("\t-f, --file <cmdfile>\t\tExecute commands in cmdfile\n");
 	g_print ("\t-R, --request <command>\t\tExecute a single command\n");
 	g_print ("\t-s, --superblock <backup#>\tOpen the device using a backup superblock\n");
+	g_print ("\t-i, --image\t\t\tOpen an o2image file\n");
 	g_print ("\t-w, --write\t\t\tOpen in read-write mode instead of the default of read-only\n");
 	g_print ("\t-V, --version\t\t\tShow version\n");
 	g_print ("\t-n, --noprompt\t\t\tHide prompt\n");
@@ -189,6 +190,7 @@ static void get_options(int argc, char **argv, dbgfs_opts *opts)
 		{ "decode", 0, 0, 'd' },
 		{ "encode", 0, 0, 'e' },
 		{ "superblock", 0, 0, 's' },
+		{ "image", 0, 0, 'i' },
 		{ 0, 0, 0, 0}
 	};
 
@@ -196,7 +198,8 @@ static void get_options(int argc, char **argv, dbgfs_opts *opts)
 		if (decodemode || encodemode || logmode)
 			break;
 
-		c = getopt_long(argc, argv, "lf:R:deV?wns:", long_options, NULL);
+		c = getopt_long(argc, argv, "lf:R:deV?wns:i",
+				long_options, NULL);
 		if (c == -1)
 			break;
 
@@ -225,6 +228,10 @@ static void get_options(int argc, char **argv, dbgfs_opts *opts)
 			encodemode++;
 			break;
 
+		case 'i':
+			opts->imagefile = 1;
+			break;
+
 		case 'l':
 			logmode++;
 			break;
@@ -321,14 +328,15 @@ static int set_logmode_proc(struct log_entry *entry)
 	return 0;
 }
 
-#define LOG_CTL_SYSFS_DIR "/sys/o2cb/logmask"
-#define LOG_CTL_SYSFS_FORMAT LOG_CTL_SYSFS_DIR "/%s"
-static int set_logmode_sysfs(struct log_entry *entry)
+#define LOG_CTL_SYSFS_DIR_OLD "/sys/o2cb/logmask"
+#define LOG_CTL_SYSFS_DIR "/sys/fs/o2cb/logmask"
+#define LOG_CTL_SYSFS_FORMAT "%s/%s"
+static int set_logmode_sysfs(const char *path, struct log_entry *entry)
 {
 	FILE *f;
 	char *logpath;
 
-	logpath = g_strdup_printf(LOG_CTL_SYSFS_FORMAT, entry->mask);
+	logpath = g_strdup_printf(LOG_CTL_SYSFS_FORMAT, path, entry->mask);
 	f = fopen(logpath, "w");
 	g_free(logpath);
 	if (!f) {
@@ -343,12 +351,12 @@ static int set_logmode_sysfs(struct log_entry *entry)
 	return 0;
 }
 
-static int get_logmode_sysfs(const char *name)
+static int get_logmode_sysfs(const char *path, const char *name)
 {
 	char *logpath;
 	char *current_mask;
 
-	logpath = g_strdup_printf(LOG_CTL_SYSFS_FORMAT, name);
+	logpath = g_strdup_printf(LOG_CTL_SYSFS_FORMAT, path, name);
 	if (g_file_get_contents(logpath, &current_mask,
 				NULL, NULL)) {
 		fprintf(stdout, "%s %s", name, current_mask);
@@ -378,7 +386,7 @@ static void run_logmode_proc(void)
 	}
 }
 
-static void run_logmode_sysfs(void)
+static void run_logmode_sysfs(const char *path)
 {
 	GList *tmp;
 	DIR *dir;
@@ -387,15 +395,15 @@ static void run_logmode_sysfs(void)
 	if (loglist) {
 		tmp = loglist;
 		while (tmp) {
-			if (set_logmode_sysfs(tmp->data))
+			if (set_logmode_sysfs(path, tmp->data))
 				break;
 			tmp = tmp->next;
 		}
 	} else {
-		dir = opendir(LOG_CTL_SYSFS_DIR);
+		dir = opendir(path);
 		if (dir) {
 			while ((d = readdir(dir)) != NULL)
-				get_logmode_sysfs(d->d_name);
+				get_logmode_sysfs(path, d->d_name);
 			closedir(dir);
 		}
 	}
@@ -407,7 +415,10 @@ static void run_logmode(void)
 
 	if (!stat(LOG_CTL_SYSFS_DIR, &stat_buf) &&
 	    S_ISDIR(stat_buf.st_mode))
-		run_logmode_sysfs();
+		run_logmode_sysfs(LOG_CTL_SYSFS_DIR);
+        else if (!stat(LOG_CTL_SYSFS_DIR_OLD, &stat_buf) &&
+	    S_ISDIR(stat_buf.st_mode))
+		run_logmode_sysfs(LOG_CTL_SYSFS_DIR_OLD);
 	else if (!stat(LOG_CTL_PROC, &stat_buf) &&
 		 S_ISREG(stat_buf.st_mode))
 		run_logmode_proc();
@@ -458,6 +469,7 @@ int main (int argc, char **argv)
 	}
 
 	gbls.allow_write = opts.allow_write;
+	gbls.imagefile = opts.imagefile;
 	if (!opts.cmd_file)
 		gbls.interactive++;
 
