@@ -3,7 +3,7 @@
  *
  * utility functions
  *
- * Copyright (C) 2004 Oracle.  All rights reserved.
+ * Copyright (C) 2004, 2008 Oracle.  All rights reserved.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
@@ -20,32 +20,42 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 021110-1307, USA.
  *
- * Authors: Sunil Mushran
  */
 
-#include <main.h>
-#include <bitops.h>
+#include "main.h"
+#include "ocfs2/bitops.h"
 
 extern dbgfs_gbls gbls;
 
+/*
+ * Tests features in a flag field and adds to a printable string.  Expects
+ * the flag field named 'flag' and the string named 'str'.  Primarily for
+ * readability.
+ */
+#define test_feature(_flag, _flagstr) do {		\
+	typeof(flag) _f = (_flag);			\
+	if (flag & _f) {				\
+		g_string_append(str, (_flagstr));	\
+		g_string_append_c(str, ' ');		\
+		flag &= ~_f;				\
+	}						\
+} while (0)
+
 void get_incompat_flag(uint32_t flag, GString *str)
 {
-	if (flag & OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV)
-		g_string_append(str, "Heartbeat ");
 
-	if (flag & OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG)
-		g_string_append(str, "AbortedResize ");
+	test_feature(OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV, "Heartbeat");
+	test_feature(OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG, "AbortedResize");
+	test_feature(OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT, "Local");
+	test_feature(OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC, "Sparse");
+	test_feature(OCFS2_FEATURE_INCOMPAT_EXTENDED_SLOT_MAP,
+		     "ExtendedSlotMap");
+	test_feature(OCFS2_FEATURE_INCOMPAT_USERSPACE_STACK,
+		     "UserspaceClusterStack");
+	test_feature(OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG, "AbortedTunefs");
 
-	if (flag & OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT)
-		g_string_append(str, "Local ");
-
-	if (flag & OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC)
-		g_string_append(str, "Sparse ");
-
-	if (flag & ~(OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV |
-		     OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG |
-		     OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT |
-		     OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC))
+        /* test_feature() clears out known flags */
+	if (flag)
 		g_string_append(str, "Unknown ");
 
 	if (!str->len)
@@ -54,12 +64,28 @@ void get_incompat_flag(uint32_t flag, GString *str)
 	return;
 }
 
+void get_tunefs_flag(uint32_t incompat_flag, uint16_t flag, GString *str)
+{
+	if (!(incompat_flag & OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG)) {
+		g_string_append(str, "None");
+		return;
+	}
+
+	if (flag & OCFS2_TUNEFS_INPROG_REMOVE_SLOT)
+		g_string_append(str, "RemoveSlot ");
+
+	if (flag & ~OCFS2_TUNEFS_INPROG_REMOVE_SLOT)
+		g_string_append(str, "Unknown ");
+
+	return;
+}
+
 void get_compat_flag(uint32_t flag, GString *str)
 {
-	if (flag & OCFS2_FEATURE_COMPAT_BACKUP_SB)
-		g_string_append(str, "BackupSuper ");
+	test_feature(OCFS2_FEATURE_COMPAT_BACKUP_SB, "BackupSuper");
 
-	if (flag & ~(OCFS2_FEATURE_COMPAT_BACKUP_SB))
+        /* test_feature() clears out known flags */
+        if (flag)
 		g_string_append(str, "Unknown ");
 
 	if (!str->len)
@@ -70,7 +96,10 @@ void get_compat_flag(uint32_t flag, GString *str)
 
 void get_rocompat_flag(uint32_t flag, GString *str)
 {
-	if (flag)
+	test_feature(OCFS2_FEATURE_RO_COMPAT_UNWRITTEN, "Unwritten");
+
+        /* test_feature() clears out known flags */
+        if (flag)
 		 g_string_append(str, "Unknown ");
 
 	if (!str->len)
@@ -422,12 +451,6 @@ errcode_t dump_file(ocfs2_filesys *fs, uint64_t ino, int fd, char *out_file,
 		goto bail;
 	}
 
-	ret = ocfs2_extent_map_init(fs, ci);
-	if (ret) {
-		com_err(gbls.cmd, ret, "while initializing extent map");
-		goto bail;
-	}
-
 	buflen = 1024 * 1024;
 
 	ret = ocfs2_malloc_blocks(fs->fs_io,
@@ -496,12 +519,6 @@ errcode_t read_whole_file(ocfs2_filesys *fs, uint64_t ino, char **buf,
 		goto bail;
 	}
 
-	ret = ocfs2_extent_map_init(fs, ci);
-	if (ret) {
-		com_err(gbls.cmd, ret, "while initializing extent map");
-		goto bail;
-	}
-
 	if (!*buflen) {
 		*buflen = (((ci->ci_inode->i_size + fs->fs_blocksize - 1) >>
 			    OCFS2_RAW_SB(fs->fs_super)->s_blocksize_bits) <<
@@ -838,3 +855,70 @@ void find_max_contig_free_bits(struct ocfs2_group_desc *gd, int *max_contig_free
 			*max_contig_free_bits = free_bits;
 	}
 }
+
+#define SYSFS_BASE		"/sys/kernel/"
+#define DEBUGFS_PATH		SYSFS_BASE "debug"
+#define DEBUGFS_ALTERNATE_PATH	"/debug"
+#define DEBUGFS_MAGIC		0x64626720
+errcode_t get_debugfs_path(char *debugfs_path, int len)
+{
+	errcode_t ret;
+	int err;
+	struct stat64 stat_buf;
+	struct statfs64 statfs_buf;
+	char *path = DEBUGFS_PATH;
+
+	err = stat64(SYSFS_BASE, &stat_buf);
+	if (err)
+		path = DEBUGFS_ALTERNATE_PATH;
+
+	ret = stat64(path, &stat_buf);
+	if (ret || !S_ISDIR(stat_buf.st_mode))
+		return O2CB_ET_SERVICE_UNAVAILABLE;
+
+	ret = statfs64(path, &statfs_buf);
+	if (ret || (statfs_buf.f_type != DEBUGFS_MAGIC))
+		return O2CB_ET_SERVICE_UNAVAILABLE;
+
+	strncpy(debugfs_path, path, len);
+
+	return 0;
+}
+
+errcode_t open_debugfs_file(const char *debugfs_path, const char *dirname,
+			    const char *uuid, const char *filename, FILE **fd)
+{
+	errcode_t ret = 0;
+	char path[PATH_MAX];
+
+	ret = snprintf(path, PATH_MAX - 1, "%s/%s/%s/%s",
+		       debugfs_path, dirname, uuid, filename);
+	if ((ret <= 0) || (ret == (PATH_MAX - 1)))
+		return O2CB_ET_INTERNAL_FAILURE;
+
+	*fd = fopen(path, "r");
+	if (!*fd) {
+		switch (errno) {
+		default:
+			ret = O2CB_ET_INTERNAL_FAILURE;
+			break;
+
+		case ENOTDIR:
+		case ENOENT:
+		case EISDIR:
+			ret = O2CB_ET_SERVICE_UNAVAILABLE;
+			break;
+
+		case EACCES:
+		case EPERM:
+		case EROFS:
+			ret = O2CB_ET_PERMISSION_DENIED;
+			break;
+		}
+		goto out;
+	}
+
+	ret = 0;
+out:
+	return ret;
+}
