@@ -52,8 +52,9 @@
 #include <stdint.h>
 #include <string.h>
 #include <inttypes.h>
+#include <signal.h>
 
-#include "ocfs2.h"
+#include "ocfs2/ocfs2.h"
 
 #include "fsck.h"
 #include "icount.h"
@@ -69,6 +70,43 @@
 int verbose = 0;
 
 static char *whoami = "fsck.ocfs2";
+static o2fsck_state _ost;
+static int cluster_locked = 0;
+
+static void handle_signal(int sig)
+{
+	switch (sig) {
+	case SIGTERM:
+	case SIGINT:
+		printf("\nProcess Interrupted.\n");
+
+		if (cluster_locked && _ost.ost_fs->fs_dlm_ctxt) {
+			ocfs2_release_cluster(_ost.ost_fs);
+			cluster_locked = 0;
+		}
+
+		if (_ost.ost_fs->fs_dlm_ctxt)
+			ocfs2_shutdown_dlm(_ost.ost_fs, whoami);
+
+		if (_ost.ost_fs)
+			ocfs2_close(_ost.ost_fs);
+
+		exit(1);
+	}
+
+	return ;
+}
+
+/* Call this with SIG_BLOCK to block and SIG_UNBLOCK to unblock */
+static void block_signals(int how)
+{
+     sigset_t sigs;
+
+     sigfillset(&sigs);
+     sigdelset(&sigs, SIGTRAP);
+     sigdelset(&sigs, SIGSEGV);
+     sigprocmask(how, &sigs, NULL);
+}
 
 static void print_usage(void)
 {
@@ -182,6 +220,12 @@ static errcode_t write_out_superblock(o2fsck_state *ost)
 	if (sb->s_feature_incompat & OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG)
 		sb->s_feature_incompat &= ~OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG;
 
+	if (sb->s_feature_incompat & OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG) {
+		sb->s_feature_incompat &=
+				 ~OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG;
+		sb->s_tunefs_flag = 0;
+	}
+
 	if (ost->ost_num_clusters)
 		di->i_clusters = ost->ost_num_clusters;
 
@@ -192,32 +236,6 @@ static errcode_t write_out_superblock(o2fsck_state *ost)
 	return ocfs2_write_super(ost->ost_fs);
 }
 
-static errcode_t update_backup_super(o2fsck_state *ost)
-{
-	errcode_t ret;
-	int num;
-	struct ocfs2_dinode *di = ost->ost_fs->fs_super;
-	struct ocfs2_super_block *sb = OCFS2_RAW_SB(di);
-	uint64_t blocks[OCFS2_MAX_BACKUP_SUPERBLOCKS];
-
-	if (!OCFS2_HAS_COMPAT_FEATURE(sb, OCFS2_FEATURE_COMPAT_BACKUP_SB))
-		return 0;
-
-	num = ocfs2_get_backup_super_offset(ost->ost_fs,
-					    blocks, ARRAY_SIZE(blocks));
-	if (!num)
-		return 0;
-
-	ret = ocfs2_refresh_backup_super(ost->ost_fs, blocks, num);
-	if (ret) {
-		com_err(whoami, ret, "while refreshing backup superblocks.");
-		goto bail;
-	}
-
-bail:
-	return ret;
-}
-
 static void scale_time(time_t secs, unsigned *scaled, char **units)
 {
 	if (secs < 60) {
@@ -262,6 +280,9 @@ static int fs_is_clean(o2fsck_state *ost, char *filename)
 	else if ((OCFS2_RAW_SB(ost->ost_fs->fs_super)->s_feature_incompat &
 		  OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG))
 		strcpy(reason, "incomplete volume resize detected");
+	else if ((OCFS2_RAW_SB(ost->ost_fs->fs_super)->s_feature_incompat &
+		  OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG))
+		strcpy(reason, "incomplete tunefs operation detected");
 	else if (sb->s_state & OCFS2_ERROR_FS)
 		strcpy(reason, "contains a file system with errors");
 	else if (sb->s_max_mnt_count > 0 &&
@@ -309,7 +330,7 @@ static int fs_is_clean(o2fsck_state *ost, char *filename)
 
 static void print_label(o2fsck_state *ost)
 {
-	char *label = OCFS2_RAW_SB(ost->ost_fs->fs_super)->s_label;
+	unsigned char *label = OCFS2_RAW_SB(ost->ost_fs->fs_super)->s_label;
 	size_t i, max = sizeof(OCFS2_RAW_SB(ost->ost_fs->fs_super)->s_label);
 
 	for(i = 0; i < max && label[i]; i++) {
@@ -348,8 +369,8 @@ static void mark_magical_clusters(o2fsck_state *ost)
 
 static void version(void)
 {
-	char url[] = "$URL: http://oss.oracle.com/projects/ocfs2-tools/src/tags/ocfs2-tools-1.2.4/fsck.ocfs2/fsck.c $";
-       	char rev[] = "$Rev: 1284 $";
+	char url[] = "$URL$";
+       	char rev[] = "$Rev$";
 	char noise[] = "fsck.ocfs2/fsck.c";
 	char *found;
 
@@ -447,7 +468,7 @@ static errcode_t recover_backup_super(o2fsck_state *ost,
 	if (sb_num < 1 || sb_num > OCFS2_MAX_BACKUP_SUPERBLOCKS)
 		return -1;
 
-	ocfs2_get_backup_super_offset(NULL, offsets, ARRAY_SIZE(offsets));
+	ocfs2_get_backup_super_offsets(NULL, offsets, ARRAY_SIZE(offsets));
 
 	/* iterate all the blocksize to get the right one. */
 	for (blksize = OCFS2_MIN_BLOCKSIZE;
@@ -470,7 +491,7 @@ static errcode_t recover_backup_super(o2fsck_state *ost,
 	    	   "Recover superblock information from backup block"
 		   "#%"PRIu64"?", sb)) {
 		fs->fs_super->i_blkno = OCFS2_SUPER_BLOCK_BLKNO;
-		ret = ocfs2_write_super(fs);
+		ret = ocfs2_write_primary_super(fs);
 		if (ret)
 			goto bail;
 	}
@@ -487,11 +508,70 @@ bail:
 	return ret;
 }
 
+static errcode_t recover_cluster_info(o2fsck_state *ost)
+{
+	errcode_t ret;
+	struct o2cb_cluster_desc disk = {NULL, NULL}, running = {NULL, NULL};
+
+	ret = o2cb_running_cluster_desc(&running);
+	if (ret)
+		goto bail;
+
+	ret = ocfs2_fill_cluster_desc(ost->ost_fs, &disk);
+	if (ret)
+		goto bail;
+
+	/*
+	 * If the disk matches the running cluster, there is nothing we
+	 * can fix.
+	 */
+	if ((!running.c_stack && !disk.c_stack) ||
+	    (running.c_stack && running.c_cluster &&
+	     disk.c_stack && disk.c_cluster &&
+	     !strcmp(running.c_stack, disk.c_stack) &&
+	     !strcmp(running.c_cluster, disk.c_cluster)))
+		goto bail;
+
+	/* recover the backup information to superblock. */
+	if (prompt(ost, PN, PR_RECOVER_CLUSTER_INFO,
+		   "The running cluster is using the %s stack%s%s, but "
+		   "the filesystem is configured for the %s stack%s%s.  "
+		   "Thus, %s cannot determine whether the filesystem is in "
+		   "use.  %s can reconfigure the filesystem to use the "
+		   "currently running cluster configuration.  DANGER: "
+		   "YOU MUST BE ABSOLUTELY SURE THAT NO OTHER NODE IS "
+		   "USING THIS FILESYSTEM BEFORE MODIFYING ITS CLUSTER "
+		   "CONFIGURATION.  Recover cluster configuration "
+		   "information the running cluster?",
+		   running.c_stack ? running.c_stack : "classic o2cb",
+		   running.c_stack ? " with the cluster name " : "",
+		   running.c_stack ? running.c_cluster : "",
+		   disk.c_stack ? disk.c_stack : "classic o2cb",
+		   disk.c_stack ? " with the cluster name " : "",
+		   disk.c_stack ? disk.c_cluster : "", whoami, whoami)) {
+		ret = ocfs2_set_cluster_desc(ost->ost_fs, &running);
+		if (ret)
+			goto bail;
+	}
+
+	/* no matter whether the user recover the superblock or not above,
+	 * we should return 0 in case the superblock can be opened
+	 * without the recovery.
+	 */
+	ret = 0;
+
+bail:
+	o2cb_free_cluster_desc(&running);
+	o2cb_free_cluster_desc(&disk);
+
+	return ret;
+}
+
 int main(int argc, char **argv)
 {
 	char *filename;
 	int64_t blkno, blksize;
-	o2fsck_state _ost, *ost = &_ost;
+	o2fsck_state *ost = &_ost;
 	int c, open_flags = OCFS2_FLAG_RW | OCFS2_FLAG_STRICT_COMPAT_CHECK;
 	int sb_num = 0;
 	int fsck_mask = FSCK_OK;
@@ -605,6 +685,16 @@ int main(int argc, char **argv)
 
 	filename = argv[optind];
 
+	if (signal(SIGTERM, handle_signal) == SIG_ERR) {
+		com_err(whoami, 0, "Could not set SIGTERM");
+		exit(1);
+	}
+
+	if (signal(SIGINT, handle_signal) == SIG_ERR) {
+		com_err(whoami, 0, "Could not set SIGINT");
+		exit(1);
+	}
+
 	/* recover superblock should be called at first. */
 	if (sb_num) {
 		ret = recover_backup_super(ost, filename, sb_num);
@@ -630,17 +720,33 @@ int main(int argc, char **argv)
 			goto close;
 		}
 
-		ret = ocfs2_initialize_dlm(ost->ost_fs);
+		block_signals(SIG_BLOCK);
+		ret = ocfs2_initialize_dlm(ost->ost_fs, whoami);
+		if (ret == O2CB_ET_INVALID_STACK_NAME) {
+			block_signals(SIG_UNBLOCK);
+			ret = recover_cluster_info(ost);
+			if (ret) {
+				com_err(whoami, ret,
+					"while recovering cluster information");
+				goto close;
+			}
+			block_signals(SIG_BLOCK);
+			ret = ocfs2_initialize_dlm(ost->ost_fs, whoami);
+		}
 		if (ret) {
+			block_signals(SIG_UNBLOCK);
 			com_err(whoami, ret, "while initializing the DLM");
 			goto close;
 		}
 
 		ret = ocfs2_lock_down_cluster(ost->ost_fs);
 		if (ret) {
+			block_signals(SIG_UNBLOCK);
 			com_err(whoami, ret, "while locking down the cluster");
 			goto close;
 		}
+		cluster_locked = 1;
+		block_signals(SIG_UNBLOCK);
 	}
 
 	printf("Checking OCFS2 filesystem in %s:\n", filename);
@@ -655,6 +761,15 @@ int main(int argc, char **argv)
 	printf("  max slots:          %u\n\n", 
 	       OCFS2_RAW_SB(ost->ost_fs->fs_super)->s_max_slots);
 
+	if (open_flags & OCFS2_FLAG_RW) {
+		ret = o2fsck_check_journals(ost);
+		if (ret) {
+			printf("fsck saw unrecoverable errors in the journal "
+				"files and will not continue.\n");
+			goto unlock;
+		}
+	}
+
 	ret = maybe_replay_journals(ost, filename, open_flags, blkno, blksize);
 	if (ret) {
 		printf("fsck encountered unrecoverable errors while "
@@ -730,22 +845,21 @@ done:
 		ret = write_out_superblock(ost);
 		if (ret)
 			com_err(whoami, ret, "while writing back the "
-				"superblock");
-		else {
-			ret = update_backup_super(ost);
-			if (ret)
-				com_err(whoami, ret,
-					"while updating backup superblock.");
-		}
+				"superblock(s)");
 	}
 
 unlock:
+	block_signals(SIG_BLOCK);
 	if (ost->ost_fs->fs_dlm_ctxt)
 		ocfs2_release_cluster(ost->ost_fs);
+	cluster_locked = 0;
+	block_signals(SIG_UNBLOCK);
 
 close:
+	block_signals(SIG_BLOCK);
 	if (ost->ost_fs->fs_dlm_ctxt)
-		ocfs2_shutdown_dlm(ost->ost_fs);
+		ocfs2_shutdown_dlm(ost->ost_fs, whoami);
+	block_signals(SIG_UNBLOCK);
 
 	ret = ocfs2_close(ost->ost_fs);
 	if (ret) {
