@@ -41,11 +41,11 @@
 #include <string.h>
 #include <inttypes.h>
 
-#include "byteorder.h"
+#include "ocfs2/byteorder.h"
+#include "ocfs2/ocfs2.h"
+
 #include "fsck.h"
 #include "journal.h"
-#include "jbd.h"
-#include "ocfs2.h"
 #include "pass1.h"
 #include "problem.h"
 #include "util.h"
@@ -238,11 +238,11 @@ static errcode_t lookup_journal_block(ocfs2_filesys *fs,
 				      int check_dup)
 {
 	errcode_t ret;
-	int contig;
+	uint64_t contig;
 	int was_set;
 
 	ret = ocfs2_extent_map_get_blocks(ji->ji_cinode, blkoff, 1, blkno,
-					  &contig);
+					  &contig, NULL);
 	if (ret) {
 		com_err(whoami, ret, "while looking up logical block "
 			"%"PRIu64" in slot %d's journal", blkoff, ji->ji_slot);
@@ -277,7 +277,7 @@ static errcode_t read_journal_block(ocfs2_filesys *fs,
 	if (err)
 		return err;
 
-	err = io_read_block(fs->fs_io, blkno, 1, buf);
+	err = ocfs2_read_blocks(fs, blkno, 1, buf);
 	if (err)
 		com_err(whoami, err, "while reading block %"PRIu64" of slot "
 			"%d's journal", blkno, ji->ji_slot);
@@ -482,12 +482,6 @@ static errcode_t prep_journal_info(ocfs2_filesys *fs, int slot,
 	      OCFS2_JOURNAL_DIRTY_FL))
 		goto out;
 
-	err = ocfs2_extent_map_init(fs, ji->ji_cinode);
-	if (err) {
-		com_err(whoami, err, "while initializing extent map");
-		goto out;
-	}
-
 	err = lookup_journal_block(fs, ji, 0, &ji->ji_jsb_block, 1);
 	if (err)
 		goto out;
@@ -529,7 +523,8 @@ errcode_t o2fsck_should_replay_journals(ocfs2_filesys *fs, int *should)
 	uint64_t blkno;
 	errcode_t ret;
 	ocfs2_cached_inode *cinode = NULL;
-	int contig, is_dirty;
+	int is_dirty;
+	uint64_t contig;
 	journal_superblock_t *jsb;
 
 	*should = 0;
@@ -564,12 +559,6 @@ errcode_t o2fsck_should_replay_journals(ocfs2_filesys *fs, int *should)
 			goto out;
 		}
 
-		ret = ocfs2_extent_map_init(fs, cinode);
-		if (ret) {
-			com_err(whoami, ret, "while initializing extent map");
-			goto out;
-		}
-
 		is_dirty = cinode->ci_inode->id1.journal1.ij_flags &
 			   OCFS2_JOURNAL_DIRTY_FL;
 		verbosef("slot %d JOURNAL_DIRTY_FL: %d\n", i, is_dirty);
@@ -577,7 +566,7 @@ errcode_t o2fsck_should_replay_journals(ocfs2_filesys *fs, int *should)
 			continue;
 
 		ret = ocfs2_extent_map_get_blocks(cinode, 0, 1, &blkno,
-						  &contig);
+						  &contig, NULL);
 		if (ret) {
 			com_err(whoami, ret, "while looking up the journal "
 				"super block in slot %d's journal", i);
@@ -735,3 +724,83 @@ out:
 
 	return ret;
 }
+
+static errcode_t check_journal_super(ocfs2_filesys *fs,
+				     ocfs2_cached_inode *ci)
+{
+	errcode_t ret;
+	uint64_t contig;
+	uint64_t blkno;
+	char *buf = NULL;
+
+	ret = ocfs2_malloc_blocks(fs->fs_io, 1, &buf);
+	if (ret)
+		goto out;
+
+	ret = ocfs2_extent_map_get_blocks(ci, 0, 1, &blkno, &contig, NULL);
+	if (ret)
+		goto out;
+
+	ret = ocfs2_read_journal_superblock(fs, blkno, buf);
+out:
+	return ret;
+}
+
+/* When we remove slot in tunefs.ocfs2, there may be some panic and
+ * we may corrupt some journal files, so we have to check whether the
+ * journal file is corrupted and recreate it.
+ */
+errcode_t o2fsck_check_journals(o2fsck_state *ost)
+{
+	errcode_t ret = 0;
+	uint64_t blkno;
+	uint32_t num_clusters = 0;
+	ocfs2_filesys *fs = ost->ost_fs;
+	char fname[OCFS2_MAX_FILENAME_LEN];
+	uint16_t i, max_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
+	ocfs2_cached_inode *ci = NULL;
+
+	for (i = 0; i < max_slots; i++) {
+		ret = ocfs2_lookup_system_inode(fs, JOURNAL_SYSTEM_INODE, i,
+						&blkno);
+		if (ret)
+			goto out;
+
+		ret = ocfs2_read_cached_inode(fs, blkno, &ci);
+		if (ret)
+			goto out;
+
+		if (ci->ci_inode->i_clusters > 0) {
+			/* check whether the file contains valid super block. */
+			ret = check_journal_super(fs, ci);
+			if (!ret) {
+				/* record the valid cluster size. */
+				num_clusters = ci->ci_inode->i_clusters;
+				continue;
+			}
+		}
+
+		if (num_clusters == 0) {
+			/* none of the journal is valid, servere errors. */
+			ret = OCFS2_ET_JOURNAL_TOO_SMALL;
+			goto out;
+		}
+
+		sprintf(fname,
+			ocfs2_system_inodes[JOURNAL_SYSTEM_INODE].si_name, i);
+		if (!prompt(ost, PY, PR_JOURNAL_FILE_INVALID,
+			    "journal file %s is invalid, regenerate it?",
+			    fname))
+			continue;
+
+		ret = ocfs2_make_journal(fs, blkno, num_clusters);
+		if (ret)
+			goto out;
+	}
+
+out:
+	if (ci)
+		ocfs2_free_cached_inode(fs, ci);
+	return ret;
+}
+
