@@ -49,8 +49,8 @@
 #include <inttypes.h>
 #include <time.h>
 
-#include "ocfs2.h"
-#include "bitops.h"
+#include "ocfs2/ocfs2.h"
+#include "ocfs2/bitops.h"
 
 #include "dirblocks.h"
 #include "dirparents.h"
@@ -91,20 +91,85 @@ static void find_max_free_bits(struct ocfs2_group_desc *gd, int *max_free_bits)
 	}
 }
 
+/* check whether the group really exists in the specified chain of
+ * the specified allocator file.
+ */
+static errcode_t check_group_parent(ocfs2_filesys *fs, uint64_t group,
+				    uint64_t ino, uint16_t chain,int *exist)
+{
+	errcode_t ret;
+	uint64_t gd_blkno;
+	char *buf = NULL, *gd_buf = NULL;
+	struct ocfs2_dinode *di = NULL;
+	struct ocfs2_group_desc *gd = NULL;
+	struct ocfs2_chain_rec *cr = NULL;
+
+	ret = ocfs2_malloc_block(fs->fs_io, &buf);
+	if (ret)
+		goto out;
+
+	ret = ocfs2_read_inode(fs, ino, buf);
+	if (ret) {
+		goto out;
+	}
+
+	di = (struct ocfs2_dinode *)buf;
+
+	if (!(di->i_flags & OCFS2_VALID_FL) ||
+	    !(di->i_flags & OCFS2_BITMAP_FL) ||
+	    !(di->i_flags & OCFS2_CHAIN_FL))
+		goto out;
+
+	if (di->id1.bitmap1.i_total == 0)
+		goto out;
+
+	if (di->id2.i_chain.cl_next_free_rec <= chain)
+		goto out;
+
+	cr = &di->id2.i_chain.cl_recs[chain];
+
+	ret = ocfs2_malloc_block(fs->fs_io, &gd_buf);
+	if (ret)
+		goto out;
+
+	gd_blkno = cr->c_blkno;
+	while (gd_blkno) {
+		if (gd_blkno ==  group) {
+			*exist = 1;
+			break;
+		}
+
+		ret = ocfs2_read_group_desc(fs, gd_blkno, gd_buf);
+		if (ret)
+			goto out;
+		gd = (struct ocfs2_group_desc *)gd_buf;
+
+		gd_blkno = gd->bg_next_group;
+	}
+
+out:
+	if (gd_buf)
+		ocfs2_free(&gd_buf);
+	if (buf)
+		ocfs2_free(&buf);
+	return ret;
+}
+
 static errcode_t repair_group_desc(o2fsck_state *ost,
 				   struct ocfs2_dinode *di,
 				   struct chain_state *cs,
 				   struct ocfs2_group_desc *bg,
-				   uint64_t blkno)
+				   uint64_t blkno,
+				   int *clear_ref)
 {
 	errcode_t ret = 0;
 	int changed = 0;
 	int max_free_bits = 0;
 
 	verbosef("checking desc at %"PRIu64"; blkno %"PRIu64" size %u bits %u "
-		 "free_bits %u chain %u generation %u\n", blkno, bg->bg_blkno,
-		 bg->bg_size, bg->bg_bits, bg->bg_free_bits_count, 
-		 bg->bg_chain, bg->bg_generation);
+		 "free_bits %u chain %u generation %u\n", blkno,
+		 (uint64_t)bg->bg_blkno, bg->bg_size, bg->bg_bits,
+		 bg->bg_free_bits_count, bg->bg_chain, bg->bg_generation);
 
 	if (bg->bg_generation != ost->ost_fs_generation &&
 	    prompt(ost, PY, PR_GROUP_GEN,
@@ -121,21 +186,44 @@ static errcode_t repair_group_desc(o2fsck_state *ost,
 	/* XXX maybe for advanced pain we could check to see if these 
 	 * kinds of descs have valid generations for the inodes they
 	 * reference */
-	if ((bg->bg_parent_dinode != di->i_blkno) &&
-	    prompt(ost, PY, PR_GROUP_PARENT,
+	if ((bg->bg_parent_dinode != di->i_blkno)) {
+		int exist = 0;
+		ret = check_group_parent(ost->ost_fs, bg->bg_blkno,
+					 bg->bg_parent_dinode,
+					 bg->bg_chain, &exist);
+
+		/* If we finds that the group really exists in the specified
+		 * chain of the specified alloc inode, then this may be a
+		 * duplicated group and we may need to remove it from current
+		 * inode.
+		 */
+		if (!ret && exist && prompt(ost, PY, PR_GROUP_DUPLICATE,
+		   "Group descriptor at block %"PRIu64" is "
+		   "referenced by inode %"PRIu64" but thinks its parent inode "
+		   "is %"PRIu64" and we can also see it in that inode."
+		    " So it may be duplicated.  Remove it from this inode?",
+		    blkno, (uint64_t)di->i_blkno,
+		    (uint64_t)bg->bg_parent_dinode)) {
+			*clear_ref = 1;
+			goto out;
+		}
+
+		if (prompt(ost, PY, PR_GROUP_PARENT,
 		   "Group descriptor at block %"PRIu64" is "
 		   "referenced by inode %"PRIu64" but thinks its parent inode "
 		   "is %"PRIu64".  Fix the descriptor's parent inode?", blkno,
-		   di->i_blkno, bg->bg_parent_dinode)) {
-		bg->bg_parent_dinode = di->i_blkno;
-		changed = 1;
+		   (uint64_t)di->i_blkno, (uint64_t)bg->bg_parent_dinode)) {
+			bg->bg_parent_dinode = di->i_blkno;
+			changed = 1;
+		}
+
 	}
 
 	if ((bg->bg_blkno != blkno) &&
 	    prompt(ost, PY, PR_GROUP_BLKNO,
 		   "Group descriptor read from block %"PRIu64" "
 		   "claims to be located at block %"PRIu64".  Update its "
-		   "recorded block location?", blkno, di->i_blkno)) {
+		   "recorded block location?", blkno, (uint64_t)di->i_blkno)) {
 		bg->bg_blkno = blkno;
 		changed = 1;
 	}
@@ -172,14 +260,15 @@ static errcode_t repair_group_desc(o2fsck_state *ost,
 			com_err(whoami, ret, "while writing a group "
 				"descriptor to block %"PRIu64" somewhere in "
 				"chain %d in group allocator inode %"PRIu64, 
-				bg->bg_blkno, cs->cs_chain_no, di->i_blkno);
+				(uint64_t)bg->bg_blkno, cs->cs_chain_no,
+				(uint64_t)di->i_blkno);
 			ost->ost_saw_error = 1;
 		}
 	}
 
 	cs->cs_total_bits += bg->bg_bits;
 	cs->cs_free_bits += bg->bg_free_bits_count;
-
+out:
 	return ret;
 }
 
@@ -245,7 +334,7 @@ static void unlink_group_desc(o2fsck_state *ost,
 					"descriptor to block %"PRIu64" "
 					"somewhere in chain %d in group "
 					"allocator inode %"PRIu64, 
-					next_desc, i, di->i_blkno);
+					next_desc, i, (uint64_t)di->i_blkno);
 				ost->ost_saw_error = 1;
 				goto out;
 			}
@@ -280,7 +369,7 @@ static void unlink_group_desc(o2fsck_state *ost,
 	if (ret) {
 		/* XXX ugh, undo the bitmap math? */
 		com_err(whoami, ret, "while writing inode alloc inode "
-			    "%"PRIu64, di->i_blkno);
+			    "%"PRIu64, (uint64_t)di->i_blkno);
 		ost->ost_saw_error = 1;
 		goto out;
 	}
@@ -306,7 +395,7 @@ static void mark_group_used(o2fsck_state *ost, struct chain_state *cs,
 }
 
 /*
- * Due to a glitch in mkfs, cl->cl_cpg for the GLOBAL BITMAP could be
+ * Due to a glitch in old mkfs, cl->cl_cpg for the GLOBAL BITMAP could be
  * less than the max possible for volumes having just one cluster
  * group. Fix.
  */
@@ -347,13 +436,14 @@ static errcode_t maybe_fix_clusters_per_group(o2fsck_state *ost,
 
 	if (prompt(ost, PY, PR_CHAIN_CPG,
 		   "Global bitmap at block %"PRIu64" has clusters per group "
-		   "set to %u instead of %u. Fix?", di->i_blkno, cl->cl_cpg,
-		   new_cl_cpg)) {
+		   "set to %u instead of %u. Fix?", (uint64_t)di->i_blkno,
+		   cl->cl_cpg, new_cl_cpg)) {
 		cl->cl_cpg = new_cl_cpg;
 		ret = ocfs2_write_inode(ost->ost_fs, di->i_blkno, (char *)di);
 		if (ret) {
 			com_err(whoami, ret, "while writing inode alloc inode "
-				"%"PRIu64" to fix cl_cpg", di->i_blkno);
+				"%"PRIu64" to fix cl_cpg",
+				(uint64_t)di->i_blkno);
 			ost->ost_saw_error = 1;
 			ret = 0;
 		}
@@ -383,7 +473,7 @@ static errcode_t check_chain(o2fsck_state *ost,
 	int depth = 0, clear_ref = 0;
 
 	verbosef("free %u total %u blkno %"PRIu64"\n", chain->c_free,
-		 chain->c_total, chain->c_blkno);
+		 chain->c_total, (uint64_t)chain->c_blkno);
 
 	while(1) {
 		/* fetch the next reference */
@@ -425,8 +515,8 @@ static errcode_t check_chain(o2fsck_state *ost,
 				   "%"PRIu64" contains a reference at depth "
 				   "%d to block %"PRIu64" which is out "
 				   "of range. Truncate this chain?",
-				   cs->cs_chain_no, di->i_blkno, depth,
-				   blkno))  {
+				   cs->cs_chain_no, (uint64_t)di->i_blkno,
+				   depth, blkno))  {
 
 				clear_ref = 1;
 				break;
@@ -442,8 +532,8 @@ static errcode_t check_chain(o2fsck_state *ost,
 				   "%"PRIu64" contains a reference at depth "
 				   "%d to block %"PRIu64" which doesn't have "
 				   "a valid checksum.  Truncate this chain?",
-				   cs->cs_chain_no, di->i_blkno, depth,
-				   blkno))  {
+				   cs->cs_chain_no, (uint64_t)di->i_blkno,
+				   depth, blkno))  {
 
 				clear_ref = 1;
 				break;
@@ -458,7 +548,7 @@ static errcode_t check_chain(o2fsck_state *ost,
 				"descriptor from block %"PRIu64" as pointed "
 				"to by chain %d in allocator at inode "
 				"%"PRIu64" at depth %d", blkno, 
-				cs->cs_chain_no, di->i_blkno, depth);
+				cs->cs_chain_no, (uint64_t)di->i_blkno, depth);
 			goto out;
 		}
 
@@ -474,10 +564,19 @@ static errcode_t check_chain(o2fsck_state *ost,
 			break;
 		}
 
-		ret = repair_group_desc(ost, di, cs, bg2, blkno);
+		ret = repair_group_desc(ost, di, cs, bg2, blkno, &clear_ref);
 		if (ret)
 			goto out;
 
+		/* we found a duplicate chain, so we need to clear them from
+		 * current chain.
+		 *
+		 * Please note that all the groups below this group will also
+		 * be removed from this chain because this is the mechanism
+		 * of removing slots in tunefs.ocfs2.
+		 */
+		if (clear_ref)
+			break;
 
 		/* the loop will now start by reading bg1->next_group */
 		memcpy(buf1, buf2, ost->ost_fs->fs_blocksize);
@@ -500,8 +599,8 @@ static errcode_t check_chain(o2fsck_state *ost,
 					"descriptor at depth %d in chain %d "
 					"in group allocator inode %"PRIu64" "
 					"to block %"PRIu64, depth,
-					cs->cs_chain_no, di->i_blkno,
-					bg1->bg_blkno);
+					cs->cs_chain_no, (uint64_t)di->i_blkno,
+					(uint64_t)bg1->bg_blkno);
 				ost->ost_saw_error = 1;
 			}
 		}
@@ -514,7 +613,8 @@ static errcode_t check_chain(o2fsck_state *ost,
 			   "has %u bits marked free out of %d total bits "
 			   "but the block groups in the chain have %u "
 			   "free out of %u total.  Fix this by updating "
-			   "the chain record?", cs->cs_chain_no, di->i_blkno,
+			   "the chain record?", cs->cs_chain_no,
+			   (uint64_t)di->i_blkno,
 			   chain->c_free, chain->c_total, cs->cs_free_bits,
 			   cs->cs_total_bits)) {
 			chain->c_total = cs->cs_total_bits;
@@ -547,21 +647,23 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 	if (memcmp(di->i_signature, OCFS2_INODE_SIGNATURE,
 		   strlen(OCFS2_INODE_SIGNATURE))) {
 		printf("Allocator inode %"PRIu64" doesn't have an inode "
-		       "signature.  fsck won't repair this.\n", di->i_blkno);
+		       "signature.  fsck won't repair this.\n",
+		       (uint64_t)di->i_blkno);
 		ret = OCFS2_ET_BAD_INODE_MAGIC;
 		goto out;
 	}
 
 	if (!(di->i_flags & OCFS2_VALID_FL)) {
 		printf("Allocator inode %"PRIu64" is not active.  fsck won't "
-		       "repair this.\n", di->i_blkno);
+		       "repair this.\n", (uint64_t)di->i_blkno);
 		ret = OCFS2_ET_INODE_NOT_VALID;
 		goto out;
 	}
 
 	if (!(di->i_flags & OCFS2_CHAIN_FL)) {
 		printf("Allocator inode %"PRIu64" doesn't have the CHAIN_FL "
-			"flag set.  fsck won't repair this.\n", di->i_blkno);
+		       "flag set.  fsck won't repair this.\n",
+		       (uint64_t)di->i_blkno);
 		/* not _entirely_ accurate, but pretty close. */
 		ret = OCFS2_ET_INODE_NOT_VALID;
 		goto out;
@@ -586,7 +688,7 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 			   "Chain %d in allocator inode %"PRIu64" "
 			   "contains an initial block reference to %"PRIu64" "
 			   "which is out of range.  Clear this reference?",
-			   i, di->i_blkno, cr->c_blkno)) {
+			   i, (uint64_t)di->i_blkno, (uint64_t)cr->c_blkno)) {
 
 			cr->c_blkno = 0;
 			changed = 1;
@@ -598,7 +700,7 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 	    prompt(ost, PY, PR_CHAIN_COUNT,
 		   "Allocator inode %"PRIu64" claims to have %u "
 		   "chains, but the maximum is %u. Fix the inode's count?",
-		   di->i_blkno, cl->cl_count, max_count)) {
+		   (uint64_t)di->i_blkno, cl->cl_count, max_count)) {
 		cl->cl_count = max_count;
 		changed = 1;
 	}
@@ -611,7 +713,8 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 			   "Allocator inode %"PRIu64" claims %u "
 			   "as the next free chain record, but fsck believes "
 			   "the largest valid value is %u.  Clamp the next "
-			   "record value?", di->i_blkno, cl->cl_next_free_rec,
+			   "record value?", (uint64_t)di->i_blkno,
+			   cl->cl_next_free_rec,
 			   max_count)) {
 			cl->cl_next_free_rec = cl->cl_count;
 			changed = 1;
@@ -647,7 +750,8 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 			   "Chain %d in allocator inode %"PRIu64" "
 			   "is empty.  Remove it from the chain record "
 			   "array in the inode and shift further chains "
-			   "into its place?", cs.cs_chain_no, di->i_blkno)) {
+			   "into its place?", cs.cs_chain_no,
+			   (uint64_t)di->i_blkno)) {
 
 			if (!trust_next_free) {
 				printf("Can't remove the chain becuase "
@@ -685,7 +789,7 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 			   "Allocator inode %"PRIu64" has %u bits "
 			   "marked used out of %d total bits but the chains "
 			   "have %u used out of %u total.  Fix this by "
-			   "updating the inode counts?", di->i_blkno,
+			   "updating the inode counts?", (uint64_t)di->i_blkno,
 			   di->id1.bitmap1.i_used, di->id1.bitmap1.i_total,
 			   total - free, total)) {
 			   di->id1.bitmap1.i_used = total - free;
@@ -702,7 +806,8 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 		   "Allocator inode %"PRIu64" has %"PRIu32" clusters "
 		   "represented in its allocator chains but has an "
 		   "i_clusters value of %"PRIu32". Fix this by updating "
-		   "i_clusters?", di->i_blkno, total, di->i_clusters)) {
+		   "i_clusters?", (uint64_t)di->i_blkno,
+		   total, di->i_clusters)) {
 		di->i_clusters = total;
 		changed = 1;
 	}
@@ -713,8 +818,9 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 		   "Allocator inode %"PRIu64" has %"PRIu32" clusters "
 		   "represented in its allocator chain which accounts for "
 		   "%"PRIu64" total bytes, but its i_size is %"PRIu64". "
-		   "Fix this by updating i_size?", di->i_blkno,
-		   di->id1.bitmap1.i_total, chain_bytes, di->i_size)) {
+		   "Fix this by updating i_size?", (uint64_t)di->i_blkno,
+		   di->id1.bitmap1.i_total, chain_bytes,
+		   (uint64_t)di->i_size)) {
 		di->i_size = chain_bytes;
 		changed = 1;
 	}
@@ -723,7 +829,7 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 		ret = ocfs2_write_inode(ost->ost_fs, di->i_blkno, (char *)di);
 		if (ret) {
 			com_err(whoami, ret, "while writing inode alloc inode "
-				    "%"PRIu64, di->i_blkno);
+				    "%"PRIu64, (uint64_t)di->i_blkno);
 			ost->ost_saw_error = 1;
 			ret = 0;
 		}
@@ -786,7 +892,7 @@ static errcode_t verify_bitmap_descs(o2fsck_state *ost,
 	ret = verify_chain_alloc(ost, di, buf1, buf2, allowed, forbidden);
 	if (ret) {
 		com_err(whoami, ret, "while looking up chain allocator inode "
-			"%"PRIu64, di->i_blkno);
+			"%"PRIu64, (uint64_t)di->i_blkno);
 		goto out;
 	}
 
@@ -911,7 +1017,7 @@ static errcode_t verify_bitmap_descs(o2fsck_state *ost,
 		ret = ocfs2_write_inode(ost->ost_fs, di->i_blkno, (char *)di);
 		if (ret) {
 			com_err(whoami, ret, "while writing inode alloc inode "
-				    "%"PRIu64, di->i_blkno);
+				    "%"PRIu64, (uint64_t)di->i_blkno);
 			ost->ost_saw_error = 1;
 			goto out;
 		}
@@ -973,25 +1079,36 @@ errcode_t o2fsck_pass0(o2fsck_state *ost)
 	}
 
 	verbosef("found inode alloc %"PRIu64" at block %"PRIu64"\n",
-		 di->i_blkno, blkno);
+		 (uint64_t)di->i_blkno, blkno);
 
 	ret = maybe_fix_clusters_per_group(ost, di);
 	if (ret)
 		goto out;
 
+	/*
+	 * during resize, we may update the global bitmap but fails to
+	 * to update i_clusters in superblock, so ask the user which one
+	 * to use before checking.
+	 */
+	if (fs->fs_super->i_clusters != di->i_clusters) {
+		if (prompt(ost, PY, PR_SUPERBLOCK_CLUSTERS,
+			   "Superblock has clusters set to %u instead of %u "
+			   "recorded in global_bitmap, it may be caused by an "
+			   "unsuccessful resize. Trust global_bitmap?",
+			   fs->fs_super->i_clusters, di->i_clusters)) {
+			ost->ost_num_clusters = di->i_clusters;
+			fs->fs_clusters = di->i_clusters;
+			fs->fs_blocks = ocfs2_clusters_to_blocks(fs,
+							 fs->fs_clusters);
+		}
+	}
+
 	ret = verify_bitmap_descs(ost, di, blocks + ost->ost_fs->fs_blocksize,
 				  blocks + (ost->ost_fs->fs_blocksize * 2));
 
 	if (ret)
 		goto out;
 
-	if (fs->fs_super->i_clusters != di->i_clusters) {
-		if (prompt(ost, PY, PR_SUPERBLOCK_CLUSTERS,
-			   "Superblock has clusters set to %u instead of %u. Fix?",
-			   fs->fs_super->i_clusters, di->i_clusters))
-			ost->ost_num_clusters = di->i_clusters;
-	}
-
 	printf("Pass 0b: Checking inode allocation chains\n");
 
 	/* first the global inode alloc and then each of the node's
@@ -1015,7 +1132,7 @@ errcode_t o2fsck_pass0(o2fsck_state *ost)
 		}
 
 		verbosef("found inode alloc %"PRIu64" at block %"PRIu64"\n",
-			 di->i_blkno, blkno);
+			 (uint64_t)di->i_blkno, blkno);
 
 		ret = verify_chain_alloc(ost, di,
 					 blocks + ost->ost_fs->fs_blocksize,
@@ -1068,7 +1185,7 @@ errcode_t o2fsck_pass0(o2fsck_state *ost)
 		}
 
 		verbosef("found extent alloc %"PRIu64" at block %"PRIu64"\n",
-			 di->i_blkno, blkno);
+			 (uint64_t)di->i_blkno, blkno);
 
 		ret = verify_chain_alloc(ost, di,
 					 blocks + ost->ost_fs->fs_blocksize,
