@@ -35,7 +35,7 @@
 #include <inttypes.h>
 #include <time.h>
 
-#include "ocfs2.h"
+#include "ocfs2/ocfs2.h"
 
 #include "dirparents.h"
 #include "icount.h"
@@ -97,7 +97,7 @@ static int expected_dots(o2fsck_dirblock_entry *dbe, int offset)
 
 static errcode_t fix_dirent_dots(o2fsck_state *ost, o2fsck_dirblock_entry *dbe,
 				 struct ocfs2_dir_entry *dirent, int offset, 
-				 int left, int *flags)
+				 int left, unsigned int *flags)
 {
 	int expect_dots = expected_dots(dbe, offset);
 	int changed_len = 0;
@@ -155,7 +155,7 @@ static errcode_t fix_dirent_dots(o2fsck_state *ost, o2fsck_dirblock_entry *dbe,
             prompt(ost, PY, PR_DIRENT_DOT_INODE,
 		   "The '.' entry in directory inode %"PRIu64" "
 		   "points to inode %"PRIu64" instead of itself.  Fix "
-		   "the '.' entry?", dbe->e_ino, dirent->inode)) {
+		   "the '.' entry?", dbe->e_ino, (uint64_t)dirent->inode)) {
 		dirent->inode = dbe->e_ino;
 		*flags |= OCFS2_DIRENT_CHANGED;
 	}
@@ -213,7 +213,7 @@ static int dirent_leaves_partial(struct ocfs2_dir_entry *dirent, int left)
  */
 static void fix_dirent_lengths(struct ocfs2_dir_entry *dirent,
 			       int left, struct ocfs2_dir_entry *prev,
-			       int *flags)
+			       unsigned int *flags)
 {
 	/* 
 	 * as described above we can't reconstruct either value if it is
@@ -286,7 +286,7 @@ out:
 
 static void fix_dirent_name(o2fsck_state *ost, o2fsck_dirblock_entry *dbe,
 			    struct ocfs2_dir_entry *dirent, int offset,
-			    int *flags)
+			    unsigned int *flags)
 {
 	char *chr = dirent->name;
 	int len = dirent->name_len, fix = 0;
@@ -318,13 +318,13 @@ static void fix_dirent_name(o2fsck_state *ost, o2fsck_dirblock_entry *dbe,
 
 static void fix_dirent_inode(o2fsck_state *ost, o2fsck_dirblock_entry *dbe,
 			     struct ocfs2_dir_entry *dirent, int offset,
-			     int *flags)
+			     unsigned int *flags)
 {
 	if (ocfs2_block_out_of_range(ost->ost_fs, dirent->inode) &&
 	    prompt(ost, PY, PR_DIRENT_INODE_RANGE,
 		   "Directory entry '%.*s' refers to inode "
 		   "number %"PRIu64" which is out of range, clear the entry?",
-		   dirent->name_len, dirent->name, dirent->inode)) {
+		   dirent->name_len, dirent->name, (uint64_t)dirent->inode)) {
 
 		dirent->inode = 0;
 		*flags |= OCFS2_DIRENT_CHANGED;
@@ -335,7 +335,7 @@ static void fix_dirent_inode(o2fsck_state *ost, o2fsck_dirblock_entry *dbe,
 	    prompt(ost, PY, PR_DIRENT_INODE_FREE,
 		   "Directory entry '%.*s' refers to inode number "
 		   "%"PRIu64" which isn't allocated, clear the entry?", 
-		   dirent->name_len, dirent->name, dirent->inode)) {
+		   dirent->name_len, dirent->name, (uint64_t)dirent->inode)) {
 		dirent->inode = 0;
 		*flags |= OCFS2_DIRENT_CHANGED;
 	}
@@ -368,7 +368,7 @@ static errcode_t fix_dirent_filetype(o2fsck_state *ost,
 				     o2fsck_dirblock_entry *dbe,
 				     struct ocfs2_dir_entry *dirent,
 				     int offset,
-				     int *flags)
+				     unsigned int *flags)
 {
 	uint8_t expected_type;
 	errcode_t ret;
@@ -410,7 +410,7 @@ check:
 		"entry's type to match the inode's?",
 		dirent->name_len, dirent->name, 
 		file_type_string(dirent->file_type), dirent->file_type,
-		dirent->inode,
+		(uint64_t)dirent->inode,
 		file_type_string(expected_type), expected_type)) {
 
 		dirent->file_type = expected_type;
@@ -430,7 +430,7 @@ static errcode_t fix_dirent_linkage(o2fsck_state *ost,
 				    o2fsck_dirblock_entry *dbe,
 				    struct ocfs2_dir_entry *dirent,
 				    int offset,
-				    int *flags)
+				    unsigned int *flags)
 {
 	int expect_dots = expected_dots(dbe, offset);
 	o2fsck_dir_parent *dp;
@@ -446,7 +446,7 @@ static errcode_t fix_dirent_linkage(o2fsck_state *ost,
 	ret = ocfs2_bitmap_test(ost->ost_dir_inodes, dirent->inode, &is_dir);
 	if (ret)
 		com_err(whoami, ret, "while checking for inode %"PRIu64" in "
-			"the dir bitmap", dirent->inode);
+			"the dir bitmap", (uint64_t)dirent->inode);
 	if (!is_dir)
 		goto out;
 
@@ -454,7 +454,7 @@ static errcode_t fix_dirent_linkage(o2fsck_state *ost,
 	if (dp == NULL) {
 		ret = OCFS2_ET_INTERNAL_FAILURE;
 		com_err(whoami, ret, "no dir parents recorded for inode "
-			"%"PRIu64, dirent->inode);
+			"%"PRIu64, (uint64_t)dirent->inode);
 		goto out;
 	}
 
@@ -470,7 +470,7 @@ static errcode_t fix_dirent_linkage(o2fsck_state *ost,
 		"claim to be the parent of subdir '%.*s' (inode %"PRIu64"). "
 		"Clear this directory entry and leave the previous parent of "
 		"the subdir's inode intact?", dbe->e_ino, 
-		dirent->name_len, dirent->name, dirent->inode)) {
+		dirent->name_len, dirent->name, (uint64_t)dirent->inode)) {
 
 		dirent->inode = 0;
 		*flags |= OCFS2_DIRENT_CHANGED;
@@ -499,7 +499,7 @@ static errcode_t fix_dirent_dups(o2fsck_state *ost,
 				 o2fsck_dirblock_entry *dbe,
 				 struct ocfs2_dir_entry *dirent,
 				 o2fsck_strings *strings,
-				 int *flags)
+				 unsigned int *flags)
 {
 	errcode_t ret = 0;
 	char *new_name = NULL;
@@ -594,7 +594,7 @@ static int corrupt_dirent_lengths(struct ocfs2_dir_entry *dirent, int left)
 		return 0;
 
 	verbosef("corrupt dirent: %"PRIu64" rec_len %u name_len %u\n",
-		dirent->inode, dirent->rec_len, dirent->name_len);
+		 (uint64_t)dirent->inode, dirent->rec_len, dirent->name_len);
 
 	return 1;
 }
@@ -630,7 +630,7 @@ static unsigned pass2_dir_block_iterate(o2fsck_dirblock_entry *dbe,
 		}
 
 		verbosef("dir inode %"PRIu64" i_size %"PRIu64"\n",
-			 dbe->e_ino, di->i_size);
+			 dbe->e_ino, (uint64_t)di->i_size);
 
 	}
 
@@ -727,7 +727,7 @@ static unsigned pass2_dir_block_iterate(o2fsck_dirblock_entry *dbe,
 			goto next;
 
 		verbosef("dirent %.*s refs ino %"PRIu64"\n", dirent->name_len,
-				dirent->name, dirent->inode);
+				dirent->name, (uint64_t)dirent->inode);
 		o2fsck_icount_delta(dd->ost->ost_icount_refs, dirent->inode, 1);
 next:
 		offset += dirent->rec_len;
