@@ -33,7 +33,7 @@
 #include <string.h>
 #include <limits.h>
 
-#include <ocfs2.h>
+#include "ocfs2/ocfs2.h"
 
 #include "dirparents.h"
 #include "fsck.h"
@@ -95,7 +95,7 @@ static void check_root(o2fsck_state *ost)
 	ost->ost_fs->fs_root_blkno = blkno;
 	sb->s_root_blkno = blkno;
 
-	ret = ocfs2_write_super(ost->ost_fs);
+	ret = ocfs2_write_primary_super(ost->ost_fs);
 	if (ret) {
 		com_err(whoami, ret, "while writing the super block with a "
 			"new root directory inode");
