@@ -29,7 +29,7 @@
 #include <inttypes.h>
 #include <limits.h>
 
-#include <ocfs2.h>
+#include "ocfs2/ocfs2.h"
 
 #include "fsck.h"
 #include "icount.h"
@@ -57,7 +57,7 @@ static void check_link_counts(o2fsck_state *ost,
 	    prompt(ost, PY, PR_INODE_NOT_CONNECTED,
 		   "Inode %"PRIu64" isn't referenced by any "
 		   "directory entries.  Move it to lost+found?", 
-		   di->i_blkno)) {
+		   (uint64_t)di->i_blkno)) {
 		o2fsck_reconnect_file(ost, blkno);
 		refs = o2fsck_icount_get(ost->ost_icount_refs, blkno);
 	}
@@ -77,13 +77,14 @@ static void check_link_counts(o2fsck_state *ost,
 	if (in_inode != di->i_links_count)
 		com_err(whoami, OCFS2_ET_INTERNAL_FAILURE, "fsck's thinks "
 			"inode %"PRIu64" has a link count of %"PRIu16" but on "
-			"disk it is %"PRIu16, di->i_blkno, in_inode, 
+			"disk it is %"PRIu16, (uint64_t)di->i_blkno, in_inode,
 			di->i_links_count);
 
 	if (prompt(ost, PY, PR_INODE_COUNT,
 		   "Inode %"PRIu64" has a link count of %"PRIu16" on "
 		   "disk but directory entry references come to %"PRIu16". "
-		   "Update the count on disk to match?", di->i_blkno, in_inode, 
+		   "Update the count on disk to match?",
+		   (uint64_t)di->i_blkno, in_inode,
 		   refs)) {
 		di->i_links_count = refs;
 		o2fsck_icount_set(ost->ost_icount_in_inodes, di->i_blkno,
@@ -114,14 +115,15 @@ static int replay_orphan_iterate(struct ocfs2_dir_entry *dirent,
 
 	if (!prompt(ost, PY, PR_INODE_ORPHANED,
 		   "Inode %"PRIu64" was found in the orphan directory. "
-		   "Delete its contents and unlink it?", dirent->inode)) {
+		   "Delete its contents and unlink it?",
+		   (uint64_t)dirent->inode)) {
 		goto out;
 	}
 
 	ret = ocfs2_truncate(ost->ost_fs, dirent->inode, 0);
 	if (ret) {
 		com_err(whoami, ret, "while truncating orphan inode %"PRIu64,
-			dirent->inode);
+			(uint64_t)dirent->inode);
 		ret_flags |= OCFS2_DIRENT_ABORT;
 		goto out;
 	}
@@ -129,7 +131,7 @@ static int replay_orphan_iterate(struct ocfs2_dir_entry *dirent,
 	ret = ocfs2_delete_inode(ost->ost_fs, dirent->inode);
 	if (ret) {
 		com_err(whoami, ret, "while deleting orphan inode %"PRIu64
-			"after truncating it", dirent->inode);
+			"after truncating it", (uint64_t)dirent->inode);
 		ret_flags |= OCFS2_DIRENT_ABORT;
 		goto out;
 	}
@@ -152,6 +154,46 @@ out:
 	return ret_flags;
 }
 
+static errcode_t create_orphan_dir(o2fsck_state *ost, char *fname)
+{
+	errcode_t ret;
+	uint64_t blkno;
+	ocfs2_filesys *fs = ost->ost_fs;
+
+	/* create inode for system file */
+	ret = ocfs2_new_system_inode(fs, &blkno,
+			ocfs2_system_inodes[ORPHAN_DIR_SYSTEM_INODE].si_mode,
+			ocfs2_system_inodes[ORPHAN_DIR_SYSTEM_INODE].si_iflags);
+	if (ret)
+		goto bail;
+
+	ret = ocfs2_expand_dir(fs, blkno, fs->fs_sysdir_blkno);
+	if (ret)
+		goto bail;
+
+	/* Add the inode to the system dir */
+	ret = ocfs2_link(fs, fs->fs_sysdir_blkno, fname, blkno,
+			 OCFS2_FT_DIR);
+	if (ret == OCFS2_ET_DIR_NO_SPACE) {
+		ret = ocfs2_expand_dir(fs, fs->fs_sysdir_blkno,
+				       fs->fs_sysdir_blkno);
+		if (!ret)
+			ret = ocfs2_link(fs, fs->fs_sysdir_blkno,
+					 fname, blkno, OCFS2_FT_DIR);
+	}
+
+	if (ret)
+		goto bail;
+
+	/* we have created an orphan dir under system dir and updated the disk,
+	 * so we have to update the refs in ost accordingly.
+	 */
+	o2fsck_icount_delta(ost->ost_icount_refs, fs->fs_sysdir_blkno, 1);
+	o2fsck_icount_delta(ost->ost_icount_in_inodes, fs->fs_sysdir_blkno, 1);
+bail:
+	return ret;
+}
+
 static errcode_t replay_orphan_dir(o2fsck_state *ost)
 {
 	errcode_t ret = OCFS2_ET_CORRUPT_SUPERBLOCK;
@@ -171,8 +213,25 @@ static errcode_t replay_orphan_dir(o2fsck_state *ost)
 
 		ret = ocfs2_lookup(ost->ost_fs, ost->ost_fs->fs_sysdir_blkno,
 				   name, bytes, NULL, &ino);
-		if (ret)
-			goto out;
+		if (ret) {
+			if (ret != OCFS2_ET_FILE_NOT_FOUND)
+				goto out;
+
+			/* orphan dir is missing, it may be caused by an
+			 * unsuccessful removing slots in tunefs.ocfs2.
+			 * so create it.
+			 */
+	   		if (prompt(ost, PY, PR_ORPHAN_DIR_MISSING,
+				   "%s is missing in system directory. "
+				   "Create it?", name)) {
+				ret = create_orphan_dir(ost, name);
+				if (ret) {
+					com_err(whoami, ret, "while creating"
+						"orphan directory %s", name);
+					continue;
+				}
+			}
+		}
 
 		ret = ocfs2_dir_iterate(ost->ost_fs, ino,
 					OCFS2_DIRENT_FLAG_EXCLUDE_DOTS, NULL,
@@ -219,7 +278,7 @@ errcode_t o2fsck_pass4(o2fsck_state *ost)
 	struct ocfs2_dinode *di;
 	char *buf = NULL;
 	errcode_t ret;
-	uint64_t blkno, start;
+	uint64_t blkno = 0, start;
 
 	printf("Pass 4a: checking for orphaned inodes\n");
 
