@@ -27,7 +27,7 @@
  */
 #include <inttypes.h>
 #include <string.h>
-#include "ocfs2.h"
+#include "ocfs2/ocfs2.h"
 
 #include "util.h"
 
@@ -40,14 +40,14 @@ void o2fsck_write_inode(o2fsck_state *ost, uint64_t blkno,
 	if (blkno != di->i_blkno) {
 		com_err(whoami, OCFS2_ET_INTERNAL_FAILURE, "when asked to "
 			"write an inode with an i_blkno of %"PRIu64" to block "
-			"%"PRIu64, di->i_blkno, blkno);
+			"%"PRIu64, (uint64_t)di->i_blkno, blkno);
 		return;
 	}
 
 	ret = ocfs2_write_inode(ost->ost_fs, blkno, (char *)di);
 	if (ret) {
-		com_err(whoami, ret, "while writing inode %"PRIu64, 
-		        di->i_blkno);
+		com_err(whoami, ret, "while writing inode %"PRIu64,
+			(uint64_t)di->i_blkno);
 		ost->ost_saw_error = 1;
 	}
 }
@@ -70,6 +70,13 @@ void o2fsck_mark_clusters_allocated(o2fsck_state *ost, uint32_t cluster,
 		o2fsck_mark_cluster_allocated(ost, cluster++);
 }
 
+void o2fsck_mark_cluster_unallocated(o2fsck_state *ost, uint32_t cluster)
+{
+	int was_set;
+
+	ocfs2_bitmap_clear(ost->ost_allocated_clusters, cluster, &was_set);
+}
+
 errcode_t o2fsck_type_from_dinode(o2fsck_state *ost, uint64_t ino,
 				  uint8_t *type)
 {
@@ -103,24 +110,17 @@ out:
 	return ret;
 }
 
-#define BITCOUNT(x)     (((BX_(x)+(BX_(x)>>4)) & 0x0F0F0F0F) % 255)
-#define BX_(x)          ((x) - (((x)>>1)&0x77777777) \
-			     - (((x)>>2)&0x33333333) \
-			     - (((x)>>3)&0x11111111))
-unsigned long o2fsck_bitcount(void *bytes, size_t len)
+size_t o2fsck_bitcount(unsigned char *bytes, size_t len)
 {
-	uint32_t val;
-	unsigned long total = 0;
-	size_t this;
+	static unsigned char nibble_count[16] = {
+		0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4
+	};
+	size_t count = 0;
 
-	while(len) {
-		val = 0;
-		this = ocfs2_min(len, sizeof(val));
-		memcpy(&val, bytes, this);
-		total += BITCOUNT(val);
-		len -= this;
+	for (; len--; bytes++) {
+		count += nibble_count[*bytes >> 4];
+		count += nibble_count[*bytes & 0xf];
 	}
-	return total;
+
+	return count;
 }
-#undef BITCOUNT
-#undef BX_
