@@ -29,7 +29,7 @@
 #include <string.h>
 #include <inttypes.h>
 
-#include "ocfs2.h"
+#include "ocfs2/ocfs2.h"
 
 static errcode_t ocfs2_chain_alloc_with_io(ocfs2_filesys *fs,
 					   ocfs2_cached_inode *cinode,
@@ -140,7 +140,7 @@ static void ocfs2_init_inode(ocfs2_filesys *fs, struct ocfs2_dinode *di,
 		di->i_links_count = 2;
 	else
 		di->i_links_count = 1;
-	strcpy(di->i_signature, OCFS2_INODE_SIGNATURE);
+	strcpy((char *)di->i_signature, OCFS2_INODE_SIGNATURE);
 	di->i_atime = di->i_ctime = di->i_mtime = time(NULL);
 	di->i_dtime = 0;
 
@@ -181,7 +181,7 @@ static void ocfs2_init_eb(ocfs2_filesys *fs,
 			  struct ocfs2_extent_block *eb,
 			  uint64_t gd_blkno, uint64_t blkno)
 {
-	strcpy(eb->h_signature, OCFS2_EXTENT_BLOCK_SIGNATURE);
+	strcpy((char *)eb->h_signature, OCFS2_EXTENT_BLOCK_SIGNATURE);
 	eb->h_fs_generation = fs->fs_super->i_fs_generation;
 	eb->h_blkno = blkno;
 	eb->h_suballoc_slot = 0;
@@ -279,7 +279,7 @@ errcode_t ocfs2_delete_inode(ocfs2_filesys *fs, uint64_t ino)
 	errcode_t ret;
 	char *buf;
 	struct ocfs2_dinode *di;
-	int slot;
+	int16_t slot;
 	ocfs2_cached_inode **inode_alloc;
 
 	ret = ocfs2_malloc_block(fs->fs_io, &buf);
@@ -292,13 +292,17 @@ errcode_t ocfs2_delete_inode(ocfs2_filesys *fs, uint64_t ino)
 	di = (struct ocfs2_dinode *)buf;
 	slot = di->i_suballoc_slot;
 
-	if (slot == OCFS2_INVALID_SLOT)
+	if (slot == OCFS2_INVALID_SLOT) {
 		inode_alloc = &fs->fs_system_inode_alloc;
-	else
+		ret = ocfs2_load_allocator(fs, GLOBAL_INODE_ALLOC_SYSTEM_INODE,
+					   0, inode_alloc);
+	}
+	else {
 		inode_alloc = &fs->fs_inode_allocs[slot];
+		ret = ocfs2_load_allocator(fs, INODE_ALLOC_SYSTEM_INODE, slot,
+					   inode_alloc);
+	}
 
-	ret = ocfs2_load_allocator(fs, INODE_ALLOC_SYSTEM_INODE, slot,
-				   inode_alloc);
 	if (ret)
 		goto out;
 
