@@ -10,12 +10,12 @@
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
  * License, version 2,  as published by the Free Software Foundation.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public
  * License along with this program; if not, write to the
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
@@ -25,8 +25,8 @@
 
 #define _XOPEN_SOURCE 600 /* Triggers magic in features.h */
 #define _LARGEFILE64_SOURCE
-                                                                                                                                                         
-#include "ocfs2.h"
+
+#include "ocfs2/ocfs2.h"
 
 #define DEFAULT_DLMFS_PATH	"/dlm/"
 
@@ -101,31 +101,129 @@ bail:
 	return ret;
 }
 
-errcode_t ocfs2_initialize_dlm(ocfs2_filesys *fs)
+errcode_t ocfs2_fill_cluster_desc(ocfs2_filesys *fs,
+				  struct o2cb_cluster_desc *desc)
+{
+	errcode_t ret = 0;
+	struct ocfs2_super_block *sb = OCFS2_RAW_SB(fs->fs_super);
+
+	if (!ocfs2_userspace_stack(sb)) {
+		desc->c_stack = NULL;
+		desc->c_cluster = NULL;
+		return 0;
+	}
+
+	ret = ocfs2_malloc0(OCFS2_STACK_LABEL_LEN + 1, &desc->c_stack);
+	if (ret)
+		return ret;
+
+	ret = ocfs2_malloc0(OCFS2_CLUSTER_NAME_LEN + 1, &desc->c_cluster);
+	if (ret) {
+		ocfs2_free(&desc->c_stack);
+		return ret;
+	}
+
+	memcpy(desc->c_stack, sb->s_cluster_info.ci_stack,
+	       OCFS2_STACK_LABEL_LEN);
+	memcpy(desc->c_cluster, sb->s_cluster_info.ci_cluster,
+	       OCFS2_CLUSTER_NAME_LEN);
+
+	return 0;
+}
+
+errcode_t ocfs2_set_cluster_desc(ocfs2_filesys *fs,
+				 struct o2cb_cluster_desc *desc)
+{
+	errcode_t ret;
+	struct ocfs2_super_block *sb = OCFS2_RAW_SB(fs->fs_super);
+
+	if (desc->c_stack) {
+		if (!desc->c_stack[0] || !desc->c_cluster ||
+		    !desc->c_cluster[0]) {
+			ret = OCFS2_ET_INVALID_ARGUMENT;
+			goto out;
+		}
+
+		if (!ocfs2_uses_extended_slot_map(sb)) {
+			sb->s_feature_incompat |=
+				OCFS2_FEATURE_INCOMPAT_EXTENDED_SLOT_MAP;
+			ret = ocfs2_format_slot_map(fs);
+			if (ret)
+				goto out;
+		}
+		sb->s_feature_incompat |=
+			OCFS2_FEATURE_INCOMPAT_USERSPACE_STACK;
+		memcpy(sb->s_cluster_info.ci_stack, desc->c_stack,
+		       OCFS2_STACK_LABEL_LEN);
+		memcpy(sb->s_cluster_info.ci_cluster, desc->c_cluster,
+		       OCFS2_CLUSTER_NAME_LEN);
+	} else {
+		sb->s_feature_incompat &=
+			~OCFS2_FEATURE_INCOMPAT_USERSPACE_STACK;
+	}
+
+	ret = ocfs2_write_super(fs);
+
+out:
+	return ret;
+}
+
+errcode_t ocfs2_initialize_dlm(ocfs2_filesys *fs, const char *service)
 {
 	struct o2dlm_ctxt *dlm_ctxt = NULL;
 	errcode_t ret = 0;
+	struct o2cb_cluster_desc cluster;
+	struct o2cb_region_desc desc;
+	char *stack_path;
 
-	ret = ocfs2_start_heartbeat(fs);
+	ret = ocfs2_fill_cluster_desc(fs, &cluster);
 	if (ret)
 		goto bail;
 
-	ret = o2dlm_initialize(DEFAULT_DLMFS_PATH, fs->uuid_str, &dlm_ctxt);
+	ret = ocfs2_fill_heartbeat_desc(fs, &desc);
+	if (ret)
+		goto bail;
+
+	desc.r_service = (char *)service;
+	desc.r_persist = 0;
+	ret = o2cb_begin_group_join(&cluster, &desc);
+	if (ret)
+		goto bail;
+
+	/*
+	 * NULL c_stack means o2cb, means use DLMFS, means
+	 * pass DLMFS_PATH.  If we're using a userspace stack, pass NULL.
+	 */
+	if (cluster.c_stack)
+		stack_path = NULL;
+	else
+		stack_path = DEFAULT_DLMFS_PATH;
+	ret = o2dlm_initialize(stack_path, fs->uuid_str, &dlm_ctxt);
 	if (ret) {
 		/* What to do with an error code? */
-		ocfs2_stop_heartbeat(fs);
+
+		/* Ignore the result of complete_group_join, as we want
+		 * to propagate our o2dlm_initialize() error */
+		o2cb_complete_group_join(&cluster, &desc, ret);
 		goto bail;
 	}
 
-	fs->fs_dlm_ctxt = dlm_ctxt;
+	ret = o2cb_complete_group_join(&cluster, &desc, 0);
+
+	if (!ret)
+		fs->fs_dlm_ctxt = dlm_ctxt;
+	else
+		o2dlm_destroy(dlm_ctxt);
 
 bail:
 	return ret;
 }
 
-errcode_t ocfs2_shutdown_dlm(ocfs2_filesys *fs)
+errcode_t ocfs2_shutdown_dlm(ocfs2_filesys *fs, const char *service)
 {
 	errcode_t ret;
+	struct o2cb_cluster_desc cluster;
+	struct o2cb_region_desc desc;
 
 	ret = o2dlm_destroy(fs->fs_dlm_ctxt);
 	if (ret)
@@ -133,7 +231,17 @@ errcode_t ocfs2_shutdown_dlm(ocfs2_filesys *fs)
 
 	fs->fs_dlm_ctxt = NULL;
 
-	ret = ocfs2_stop_heartbeat(fs);
+	ret = ocfs2_fill_cluster_desc(fs, &cluster);
+	if (ret)
+		goto bail;
+
+	ret = ocfs2_fill_heartbeat_desc(fs, &desc);
+	if (ret)
+		goto bail;
+
+	desc.r_service = (char *)service;
+	desc.r_persist = 0;
+	ret = o2cb_group_leave(&cluster, &desc);
 
 bail:
 	return ret;
@@ -162,10 +270,10 @@ errcode_t ocfs2_super_unlock(ocfs2_filesys *fs)
 			     0, lock_name);
 
 	ret = o2dlm_unlock(fs->fs_dlm_ctxt, lock_name);
-	
+
 	return ret;
 }
-	
+
 errcode_t ocfs2_meta_lock(ocfs2_filesys *fs,
 			  ocfs2_cached_inode *ci,
 			  enum o2dlm_lock_level level,
@@ -203,6 +311,8 @@ errcode_t ocfs2_meta_unlock(ocfs2_filesys *fs,
 #include <stdlib.h>
 #include <libgen.h>
 
+#define DEBUG_SERVICE "debug"
+
 static void print_usage(void)
 {
 	fprintf(stderr, "Usage: dlm <filename>\n");
@@ -237,7 +347,7 @@ int main(int argc, char *argv[])
 		goto out;
 	}
 
-	ret = ocfs2_initialize_dlm(fs);
+	ret = ocfs2_initialize_dlm(fs, DEBUG_SERVICE);
 	if (ret) {
 		com_err(progname, ret, "while initializing dlm");
 		goto out;
@@ -263,7 +373,7 @@ int main(int argc, char *argv[])
 
 out:
 	if (fs->fs_dlm_ctxt) {
-		ret = ocfs2_shutdown_dlm(fs);
+		ret = ocfs2_shutdown_dlm(fs, DEBUG_SERVICE);
 		if (ret)
 			com_err(progname, ret, "while shutting down dlm");
 	}
