@@ -35,7 +35,7 @@
 #include <unistd.h>
 #endif
 
-#include "ocfs2.h"
+#include "ocfs2/ocfs2.h"
 
 /*
  * ocfs2_expand_dir()
@@ -50,7 +50,7 @@ errcode_t ocfs2_expand_dir(ocfs2_filesys *fs,
 	uint64_t used_blks;
 	uint64_t totl_blks;
 	uint64_t new_blk;
-	int contig;
+	uint64_t contig;
 	char *buf = NULL;
 
 	if (!(fs->fs_flags & OCFS2_FLAG_RW))
@@ -74,6 +74,7 @@ errcode_t ocfs2_expand_dir(ocfs2_filesys *fs,
 
 	if (used_blks >= totl_blks) {
 		ocfs2_free_cached_inode(fs, cinode);
+		cinode = NULL;
 
 		/* extend the directory */
 		ret = ocfs2_extend_allocation(fs, dir, 1);
@@ -85,13 +86,9 @@ errcode_t ocfs2_expand_dir(ocfs2_filesys *fs,
 			goto bail;
 	}
 
-	ret = ocfs2_extent_map_init(fs, cinode);
-	if (ret)
-		goto bail;
-
 	/* get the next free block */
 	ret = ocfs2_extent_map_get_blocks(cinode, used_blks, 1,
-					  &new_blk, &contig);
+					  &new_blk, &contig, NULL);
 	if (ret) 
 		goto bail;
 
