@@ -11,12 +11,12 @@
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
  * License, version 2,  as published by the Free Software Foundation.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public
  * License along with this program; if not, write to the
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
@@ -32,7 +32,8 @@
 #include <string.h>
 #include <inttypes.h>
 
-#include "ocfs2.h"
+#include "ocfs2/byteorder.h"
+#include "ocfs2/ocfs2.h"
 
 static void ocfs2_swap_extent_list_primary(struct ocfs2_extent_list *el)
 {
@@ -49,7 +50,10 @@ static void ocfs2_swap_extent_list_secondary(struct ocfs2_extent_list *el)
 		struct ocfs2_extent_rec *rec = &el->l_recs[i];
 
 		rec->e_cpos = bswap_32(rec->e_cpos);
-		rec->e_clusters = bswap_32(rec->e_clusters);
+		if (el->l_tree_depth)
+			rec->e_int_clusters = bswap_32(rec->e_int_clusters);
+		else
+			rec->e_leaf_clusters = bswap_16(rec->e_leaf_clusters);
 		rec->e_blkno = bswap_64(rec->e_blkno);
 	}
 }
@@ -114,7 +118,7 @@ errcode_t ocfs2_read_extent_block_nocheck(ocfs2_filesys *fs,
 	if (ret)
 		return ret;
 
-	ret = io_read_block(fs->fs_io, blkno, 1, blk);
+	ret = ocfs2_read_blocks(fs, blkno, 1, blk);
 	if (ret)
 		goto out;
 
@@ -248,7 +252,7 @@ static int extent_iterate_el(struct ocfs2_extent_list *el,
 				iret |= update_eb_rec(ctxt, &before,
 						      &el->l_recs[i]);
 
-			if (el->l_recs[i].e_clusters &&
+			if (el->l_recs[i].e_int_clusters &&
 			   (el->l_recs[i].e_cpos >= ctxt->last_eb_cpos)) {
 				/*
 				 * Only set last_eb_blkno if current extent
@@ -261,6 +265,12 @@ static int extent_iterate_el(struct ocfs2_extent_list *el,
 			}
 
 		} else {
+			/*
+			 * For a sparse file, we may find an empty record
+			 * in the left most record. Just skip it.
+			 */
+			if (!i && !el->l_recs[i].e_leaf_clusters)
+				continue;
 			iret |= (*ctxt->func)(ctxt->fs, &el->l_recs[i],
 					      el->l_tree_depth,
 					      ctxt->ccount, ref_blkno,
@@ -268,7 +278,8 @@ static int extent_iterate_el(struct ocfs2_extent_list *el,
 			if (iret & OCFS2_EXTENT_CHANGED)
 				iret |= update_leaf_rec(ctxt, &before,
 							&el->l_recs[i]);
-			ctxt->ccount += el->l_recs[i].e_clusters;
+			ctxt->ccount += ocfs2_rec_clusters(el->l_tree_depth,
+							   &el->l_recs[i]);
 		}
 		if (iret & (OCFS2_EXTENT_ABORT | OCFS2_EXTENT_ERROR))
 			break;
@@ -276,7 +287,8 @@ static int extent_iterate_el(struct ocfs2_extent_list *el,
 
 	if (iret & OCFS2_EXTENT_CHANGED) {
 		for (i = 0; i < el->l_count; i++) {
-			if (el->l_recs[i].e_clusters)
+			if (ocfs2_rec_clusters(el->l_tree_depth,
+					       &el->l_recs[i]))
 				continue;
 			el->l_next_free_rec = i;
 			break;
@@ -334,9 +346,12 @@ static int extent_iterate_eb(struct ocfs2_extent_rec *eb_rec,
 	if (flags & (OCFS2_EXTENT_ABORT | OCFS2_EXTENT_ERROR))
 		iret |= flags & (OCFS2_EXTENT_ABORT | OCFS2_EXTENT_ERROR);
 
-	/* if the list was changed and we still have recs then we need
-	 * to write the changes to disk */
-	if (changed & OCFS2_EXTENT_CHANGED && el->l_next_free_rec) {
+	/*
+	 * If the list was changed, we should write the changes to disk.
+	 * Note:
+	 * For a sparse file, we may have an empty extent block.
+	 */
+	if (changed & OCFS2_EXTENT_CHANGED) {
 		ctxt->errcode = ocfs2_write_extent_block(ctxt->fs,
 							 eb_rec->e_blkno,
 						ctxt->eb_bufs[tree_depth]);
@@ -406,7 +421,7 @@ errcode_t ocfs2_extent_iterate_inode(ocfs2_filesys *fs,
 
 		for (i = 1; i < el->l_tree_depth; i++) {
 			ctxt.eb_bufs[i] = ctxt.eb_bufs[0] +
-				fs->fs_blocksize;
+				i * fs->fs_blocksize;
 		}
 	}
 	else
@@ -435,7 +450,7 @@ errcode_t ocfs2_extent_iterate_inode(ocfs2_filesys *fs,
 	}
 
 out_abort:
-	if (iret & OCFS2_EXTENT_CHANGED)
+	if (!ret && (iret & OCFS2_EXTENT_CHANGED))
 		ret = ocfs2_write_inode(fs, inode->i_blkno, (char *)inode);
 
 out_eb_bufs:
@@ -489,6 +504,7 @@ struct block_context {
 	int (*func)(ocfs2_filesys *fs,
 		    uint64_t blkno,
 		    uint64_t bcount,
+		    uint16_t ext_flags,
 		    void *priv_data);
 	int flags;
 	struct ocfs2_dinode *inode;
@@ -508,15 +524,16 @@ static int block_iterate_func(ocfs2_filesys *fs,
 	uint64_t blkno, bcount, bend;
 	int iret = 0;
 
-	bcount = ocfs2_clusters_to_blocks(fs, ccount);
-	bend = bcount + ocfs2_clusters_to_blocks(fs, rec->e_clusters);
+	bcount = ocfs2_clusters_to_blocks(fs, rec->e_cpos);
+	bend = bcount + ocfs2_clusters_to_blocks(fs,
+					ocfs2_rec_clusters(tree_depth, rec));
 
 	for (blkno = rec->e_blkno; bcount < bend; blkno++, bcount++) {
 		if (((bcount * fs->fs_blocksize) >= ctxt->inode->i_size) &&
 		    !(ctxt->flags & OCFS2_BLOCK_FLAG_APPEND))
 			break;
 
-		iret = (*ctxt->func)(fs, blkno, bcount,
+		iret = (*ctxt->func)(fs, blkno, bcount, rec->e_flags,
 				     ctxt->priv_data);
 		if (iret & OCFS2_BLOCK_ABORT)
 			break;
@@ -531,6 +548,7 @@ errcode_t ocfs2_block_iterate_inode(ocfs2_filesys *fs,
 				    int (*func)(ocfs2_filesys *fs,
 						uint64_t blkno,
 						uint64_t bcount,
+						uint16_t ext_flags,
 						void *priv_data),
 				    void *priv_data)
 {
@@ -556,6 +574,7 @@ errcode_t ocfs2_block_iterate(ocfs2_filesys *fs,
 			      int (*func)(ocfs2_filesys *fs,
 					  uint64_t blkno,
 					  uint64_t bcount,
+					  uint16_t ext_flags,
 					  void *priv_data),
 			      void *priv_data)
 {
@@ -627,12 +646,14 @@ static int walk_extents_func(ocfs2_filesys *fs,
 		fprintf(stdout, " ");
 	fprintf(stdout, "(%08"PRIu32", %08"PRIu32", %08"PRIu64") |"
 			" + %08"PRIu32" = %08"PRIu32" / %08"PRIu32"\n",
-		rec->e_cpos, rec->e_clusters,
-		rec->e_blkno, ccount, ccount + rec->e_clusters,
+		rec->e_cpos, ocfs2_rec_clusters(tree_depth, rec),
+		rec->e_blkno, ccount,
+		ccount + ocfs2_rec_clusters(tree_depth, rec),
 		wi->di->i_clusters);
 
 	if (!tree_depth &&
-	    ((ccount + rec->e_clusters) == wi->di->i_clusters))
+	    ((ccount + ocfs2_rec_clusters(tree_depth, rec)) ==
+							 wi->di->i_clusters))
 		fprintf(stdout, "TOTAL: %u\n", wi->di->i_clusters);
 
 	return 0;
@@ -649,6 +670,7 @@ struct walk_block {
 static int walk_blocks_func(ocfs2_filesys *fs,
 			    uint64_t blkno,
 			    uint64_t bcount,
+			    uint16_t ext_flags,
 			    void *priv_data)
 {
 	struct walk_block *wb = priv_data;
@@ -756,7 +778,7 @@ int main(int argc, char *argv[])
 		return 1;
 	}
 	filename = argv[optind];
-	
+
 	if (!(walk_blocks + walk_extents)) {
 		fprintf(stderr,
 			"No operation specified\n");
