@@ -32,7 +32,7 @@
 #include <limits.h>
 #include <inttypes.h>
 
-#include "ocfs2.h"
+#include "ocfs2/ocfs2.h"
 
 struct read_whole_context {
 	char		*buf;
@@ -45,12 +45,17 @@ struct read_whole_context {
 static int read_whole_func(ocfs2_filesys *fs,
 			   uint64_t blkno,
 			   uint64_t bcount,
+			   uint16_t ext_flags,
 			   void *priv_data)
 {
 	struct read_whole_context *ctx = priv_data;
 
-	ctx->errcode = io_read_block(fs->fs_io, blkno,
-				     1, ctx->ptr);
+	if (ext_flags & OCFS2_EXT_UNWRITTEN) {
+		memset(ctx->ptr, 0, fs->fs_blocksize);
+		ctx->errcode = 0;
+	} else
+		ctx->errcode = ocfs2_read_blocks(fs, blkno, 1, ctx->ptr);
+
 	if (ctx->errcode)
 		return OCFS2_BLOCK_ABORT;
 
@@ -128,11 +133,12 @@ errcode_t ocfs2_file_read(ocfs2_cached_inode *ci, void *buf, uint32_t count,
 	errcode_t	ret = 0;
 	char		*ptr = (char *) buf;
 	uint32_t	wanted_blocks;
-	uint32_t	contig_blocks;
+	uint64_t	contig_blocks;
 	uint64_t	v_blkno;
 	uint64_t	p_blkno;
 	uint32_t	tmp;
 	uint64_t	num_blocks;
+	uint16_t	extent_flags;
 
 	/* o_direct requires aligned io */
 	tmp = fs->fs_blocksize - 1;
@@ -155,16 +161,26 @@ errcode_t ocfs2_file_read(ocfs2_cached_inode *ci, void *buf, uint32_t count,
 
 	while(wanted_blocks) {
 		ret = ocfs2_extent_map_get_blocks(ci, v_blkno, 1,
-						  &p_blkno, &contig_blocks);
+						  &p_blkno, &contig_blocks,
+						  &extent_flags);
 		if (ret)
 			return ret;
 
 		if (contig_blocks > wanted_blocks)
 			contig_blocks = wanted_blocks;
 
-		ret = io_read_block(fs->fs_io, p_blkno, contig_blocks, ptr);
-		if (ret)
-			return ret;
+		if (!p_blkno || extent_flags & OCFS2_EXT_UNWRITTEN) {
+			/*
+			 * we meet with a hole or an unwritten extent,
+			 * so just empty the content.
+			 */
+			memset(ptr, 0, contig_blocks * fs->fs_blocksize);
+		} else {
+			ret = ocfs2_read_blocks(fs, p_blkno, contig_blocks,
+						ptr);
+			if (ret)
+				return ret;
+		}
 
 		*got += (contig_blocks <<
 			 OCFS2_RAW_SB(fs->fs_super)->s_blocksize_bits);
@@ -184,6 +200,38 @@ errcode_t ocfs2_file_read(ocfs2_cached_inode *ci, void *buf, uint32_t count,
 	return ret;
 }
 
+/*
+ * Emtpy the blocks on the disk.
+ */
+static errcode_t empty_blocks(ocfs2_filesys *fs,
+			      uint64_t start_blk,
+			      uint64_t num_blocks)
+{
+	errcode_t ret;
+	char *buf = NULL;
+
+	ret = ocfs2_malloc_block(fs->fs_io, &buf);
+	if (ret)
+		goto bail;
+
+	memset(buf, 0, fs->fs_blocksize);
+
+	while (num_blocks) {
+		ret = io_write_block(fs->fs_io, start_blk, 1, buf);
+		if (ret)
+			goto bail;
+
+		num_blocks--;
+		start_blk++;
+	}
+
+bail:
+	if (buf)
+		ocfs2_free(&buf);
+
+	return ret;
+}
+
 errcode_t ocfs2_file_write(ocfs2_cached_inode *ci, void *buf, uint32_t count,
 			   uint64_t offset, uint32_t *wrote)
 {
@@ -191,12 +239,18 @@ errcode_t ocfs2_file_write(ocfs2_cached_inode *ci, void *buf, uint32_t count,
 	errcode_t	ret = 0;
 	char		*ptr = (char *) buf;
 	uint32_t	wanted_blocks;
-	uint32_t	contig_blocks;
+	uint64_t	contig_blocks;
 	uint64_t	v_blkno;
-	uint64_t	p_blkno;
+	uint64_t	p_blkno, p_start, p_end;
+	uint64_t	begin_blocks = 0, end_blocks = 0;
 	uint32_t	tmp;
 	uint64_t	num_blocks;
 	int		bs_bits = OCFS2_RAW_SB(fs->fs_super)->s_blocksize_bits;
+	uint64_t	ino = ci->ci_blkno;
+	uint32_t	n_clusters, cluster_begin, cluster_end;
+	uint64_t	bpc = fs->fs_clustersize/fs->fs_blocksize;
+	int		insert = 0;
+	uint16_t	extent_flags = 0;
 
 	/* o_direct requires aligned io */
 	tmp = fs->fs_blocksize - 1;
@@ -218,17 +272,122 @@ errcode_t ocfs2_file_write(ocfs2_cached_inode *ci, void *buf, uint32_t count,
 
 	while(wanted_blocks) {
 		ret = ocfs2_extent_map_get_blocks(ci, v_blkno, 1,
-						  &p_blkno, &contig_blocks);
+						  &p_blkno, &contig_blocks,
+						  &extent_flags);
 		if (ret)
 			return ret;
 
 		if (contig_blocks > wanted_blocks)
 			contig_blocks = wanted_blocks;
 
+		begin_blocks = 0;
+		end_blocks = 0;
+		p_end = 0;
+	 	if (!p_blkno) {
+			/*
+			 * We meet with a hole here, so we allocate clusters
+			 * and empty the both ends in case.
+			 *
+			 * We will postpone the extent insertion after we
+			 * successfully write the extent block, so that and
+			 * problems happens in block writing would not affect
+			 * the file.
+			 */
+			cluster_begin = ocfs2_blocks_to_clusters(fs, v_blkno);
+			cluster_end = ocfs2_blocks_to_clusters(fs,
+						v_blkno + contig_blocks -1);
+			n_clusters = cluster_end - cluster_begin + 1;
+			ret = ocfs2_new_clusters(fs, 1, n_clusters, &p_start,
+						 &n_clusters);
+			if (ret || n_clusters == 0)
+				return ret;
+
+			begin_blocks = v_blkno & (bpc - 1);
+			p_blkno = p_start + begin_blocks;
+			contig_blocks = n_clusters * bpc - begin_blocks;
+			if (contig_blocks > wanted_blocks) {
+				end_blocks = contig_blocks - wanted_blocks;
+				contig_blocks = wanted_blocks;
+				p_end = p_blkno + wanted_blocks;
+			}
+
+			insert = 1;
+		} else if (extent_flags & OCFS2_EXT_UNWRITTEN) {
+			begin_blocks = v_blkno & (bpc - 1);
+			p_start = p_blkno - begin_blocks;
+			p_end = p_blkno + wanted_blocks;
+			end_blocks = (p_end & (bpc - 1)) ?
+						 bpc - (p_end & (bpc - 1 )) : 0;
+		}
+
+		if (begin_blocks) {
+			/*
+			 * The user don't write the first blocks,
+			 * so we have to empty them.
+			 */
+			ret = empty_blocks(fs, p_start, begin_blocks);
+			if (ret)
+				return ret;
+		}
+
+		if (end_blocks) {
+			/*
+			 * we don't need to write that many blocks,
+			 * so empty the blocks at the bottom.
+			 */
+			ret = empty_blocks(fs, p_end, end_blocks);
+			if (ret)
+				return ret;
+		}
+
 		ret = io_write_block(fs->fs_io, p_blkno, contig_blocks, ptr);
 		if (ret)
 			return ret;
 
+		if (insert) {
+	 		ret = ocfs2_insert_extent(fs, ci->ci_blkno,
+					ocfs2_blocks_to_clusters(fs,v_blkno),
+					p_start, n_clusters, 0);
+			if (ret) {
+				/*
+				 * XXX: We don't wan't to overwrite the error
+				 * from insert_extent().  But we probably need
+				 * to BE LOUDLY UPSET.
+				 */
+				ocfs2_free_clusters(fs, n_clusters, p_start);
+				return ret;
+			}
+
+			/*
+			 * since the inode information has been changed, we
+			 * may need to reinitialize it and test whether we can
+			 * really find the inserted extents.
+			 */
+			ocfs2_free_cached_inode(fs, ci);
+			ret = ocfs2_read_cached_inode(fs,ino, &ci);
+			ret = ocfs2_extent_map_get_blocks(ci, v_blkno, 1,
+						&p_blkno, NULL, NULL);
+			/* now we shouldn't find a hole. */
+			if (!p_blkno || p_blkno != p_start + begin_blocks)
+				ret = OCFS2_ET_INTERNAL_FAILURE;
+			if (ret)
+				return ret;
+
+			insert = 0;
+		} else if (extent_flags & OCFS2_EXT_UNWRITTEN) {
+			cluster_begin = ocfs2_blocks_to_clusters(fs, v_blkno);
+			cluster_end = ocfs2_blocks_to_clusters(fs,
+						v_blkno + contig_blocks -1);
+			n_clusters = cluster_end - cluster_begin + 1;
+			ret = ocfs2_mark_extent_written(fs, ci->ci_inode,
+					cluster_begin, n_clusters,
+					p_blkno & ~(bpc - 1));
+			if (ret)
+				return ret;
+			ocfs2_free_cached_inode(fs, ci);
+			ocfs2_read_cached_inode(fs,ino, &ci);
+		}
+
 		*wrote += (contig_blocks << bs_bits);
 		wanted_blocks -= contig_blocks;
 
@@ -240,6 +399,7 @@ errcode_t ocfs2_file_write(ocfs2_cached_inode *ci, void *buf, uint32_t count,
 				*wrote = (uint32_t) (ci->ci_inode->i_size - offset);
 			/* break */
 		}
+
 	}
 
 	return ret;
@@ -432,4 +592,3 @@ out:
 	return 0;
 }
 #endif  /* DEBUG_EXE */
-
