@@ -28,7 +28,8 @@
 #include <string.h>
 #include <inttypes.h>
 
-#include "ocfs2.h"
+#include "ocfs2/byteorder.h"
+#include "ocfs2/ocfs2.h"
 
 void ocfs2_swap_disk_heartbeat_block(struct o2hb_disk_heartbeat_block *hb)
 {
@@ -89,7 +90,7 @@ errcode_t ocfs2_fill_heartbeat_desc(ocfs2_filesys *fs,
 		goto leave;
 	}
 
-	blocks = rec->e_clusters << cluster_bits;
+	blocks = ocfs2_rec_clusters(0, rec) << cluster_bits;
 	blocks >>= block_bits;
 
 	if (blocks > O2NM_MAX_NODES)
@@ -111,23 +112,3 @@ leave:
 	return ret;
 }
 
-errcode_t ocfs2_start_heartbeat(ocfs2_filesys *fs)
-{
-	errcode_t ret;
-	struct o2cb_region_desc desc;
-
-	ret = ocfs2_fill_heartbeat_desc(fs, &desc);	
-	if (ret)
-		goto leave;
-
-        /* XXX: NULL cluster is a hack for right now */
-	ret = o2cb_start_heartbeat_region(NULL, &desc);
-
-leave:
-	return ret;
-}
-
-errcode_t ocfs2_stop_heartbeat(ocfs2_filesys *fs)
-{
-	return o2cb_stop_heartbeat_region(NULL, fs->uuid_str);
-}
