@@ -31,8 +31,8 @@
 #include <string.h>
 #include <netinet/in.h>
 
-#include "ocfs2.h"
-#include "jbd.h"
+#include "ocfs2/byteorder.h"
+#include "ocfs2/ocfs2.h"
 
 void ocfs2_swap_journal_superblock(journal_superblock_t *jsb)
 {
@@ -150,7 +150,7 @@ errcode_t ocfs2_read_journal_superblock(ocfs2_filesys *fs, uint64_t blkno,
 	if (ret)
 		return ret;
 
-	ret = io_read_block(fs->fs_io, blkno, 1, blk);
+	ret = ocfs2_read_blocks(fs, blkno, 1, blk);
 	if (ret)
 		goto out;
 
@@ -217,10 +217,6 @@ static errcode_t ocfs2_format_journal(ocfs2_filesys *fs,
 	uint64_t offset = 0;
 	uint32_t wrote, count, jrnl_blocks;
 
-	ret = ocfs2_extent_map_init(fs, ci);
-	if (ret)
-		goto out;
-
 #define BUFLEN	1048576
 	ret = ocfs2_malloc_blocks(fs->fs_io, (BUFLEN >> bs_bits), &buf);
 	if (ret)
@@ -243,7 +239,7 @@ static errcode_t ocfs2_format_journal(ocfs2_filesys *fs,
 		goto out;
 
 	/* re-use offset here for 1st journal block. */
-	ret = ocfs2_extent_map_get_blocks(ci, 0, 1, &offset, NULL);
+	ret = ocfs2_extent_map_get_blocks(ci, 0, 1, &offset, NULL, NULL);
 	if (ret)
 		goto out;
 
@@ -264,6 +260,12 @@ errcode_t ocfs2_make_journal(ocfs2_filesys *fs, uint64_t blkno,
 	ocfs2_cached_inode *ci = NULL;
 	struct ocfs2_dinode *di;
 
+	if ((clusters << OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits) <
+	    OCFS2_MIN_JOURNAL_SIZE) {
+		ret = OCFS2_ET_JOURNAL_TOO_SMALL;
+		goto out;
+	}
+
 	ret = ocfs2_read_cached_inode(fs, blkno, &ci);
 	if (ret)
 		goto out;
@@ -301,6 +303,19 @@ errcode_t ocfs2_make_journal(ocfs2_filesys *fs, uint64_t blkno,
 		ret = ocfs2_write_inode(fs, blkno, (char *)di);
 		if (ret)
 			goto out;
+	} else if (clusters < di->i_clusters) {
+		uint64_t new_size = clusters <<
+			   OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits;
+		ret = ocfs2_truncate(fs, blkno, new_size);
+		if (ret)
+			goto out;
+
+		ocfs2_free_cached_inode(fs, ci);
+		ret = ocfs2_read_cached_inode(fs, blkno, &ci);
+		if (ret) {
+			ci = NULL;
+			goto out;
+		}
 	}
 
 	ret = ocfs2_format_journal(fs, ci);
