@@ -63,7 +63,6 @@ static void format_file(State *s, SystemFileDiskRecord *rec);
 static void write_metadata(State *s, SystemFileDiskRecord *rec, void *src);
 static void write_bitmap_data(State *s, AllocBitmap *bitmap);
 static void write_directory_data(State *s, DirData *dir);
-static void write_slot_map_data(State *s, SystemFileDiskRecord *slot_map_rec);
 static void write_group_data(State *s, AllocGroup *group);
 static void format_leading_space(State *s);
 //static void replacement_journal_create(State *s, uint64_t journal_off);
@@ -84,6 +83,7 @@ static AllocGroup * initialize_alloc_group(State *s, const char *name,
 					   uint16_t bpc);
 static void create_lost_found_dir(State *s);
 static void format_journals(State *s);
+static void format_slotmap(State *s);
 static int format_backup_super(State *s);
 
 extern char *optarg;
@@ -116,6 +116,10 @@ static struct fs_type_translation ocfs2_fs_types_table[] = {
 
 enum {
 	BACKUP_SUPER_OPTION = CHAR_MAX + 1,
+	FEATURE_LEVEL,
+	FEATURES_OPTION,
+	CLUSTER_STACK_OPTION,
+	CLUSTER_NAME_OPTION,
 };
 
 static uint64_t align_bytes_to_clusters_ceil(State *s,
@@ -373,14 +377,6 @@ main(int argc, char **argv)
 		}
 	}
 
-	if (!hb_dev_skip(s, SLOT_MAP_SYSTEM_INODE)) {
-		tmprec = &(record[SLOT_MAP_SYSTEM_INODE][0]);
-		alloc_from_bitmap(s, 1, s->global_bm,
-				  &tmprec->extent_off,
-				  &tmprec->extent_len);
-		tmprec->file_size = s->cluster_size;
-	}
-
 	fsync(s->fd);
 	if (!s->quiet)
 		printf("done\n");
@@ -412,11 +408,6 @@ main(int argc, char **argv)
 
 	write_group_data(s, s->system_group);
 
-	if (!hb_dev_skip(s, SLOT_MAP_SYSTEM_INODE)) {
-		tmprec = &(record[SLOT_MAP_SYSTEM_INODE][0]);
-		write_slot_map_data(s, tmprec);
-	}
-
 	write_directory_data(s, root_dir);
 	write_directory_data(s, system_dir);
 	
@@ -443,15 +434,6 @@ main(int argc, char **argv)
 	if (!s->quiet)
 		printf("done\n");
 
-	if (!s->no_backup_super) {
-		if (!s->quiet)
-			printf("Writing backup superblock: ");
-
-		num = format_backup_super(s);
-		if (!s->quiet)
-			printf("%d block(s)\n", num);
-	}
-
 	if (!s->hb_dev) {
 		/* These routines use libocfs2 to do their work. We
 		 * don't share an ocfs2_filesys context between the
@@ -459,11 +441,28 @@ main(int argc, char **argv)
 		 * the library can use the journal for the latter in
 		 * future revisions. */
 
+		if (!s->no_backup_super) {
+			if (!s->quiet)
+				printf("Writing backup superblock: ");
+
+			num = format_backup_super(s);
+			if (!s->quiet)
+				printf("%d block(s)\n", num);
+		}
+
 		if (!s->quiet)
 			printf("Formatting Journals: ");
 
 		format_journals(s);
 
+		if (!s->quiet)
+			printf("done\n");
+
+		if (!s->quiet)
+			printf("Formatting slot map: ");
+
+		format_slotmap(s);
+
 		if (!s->quiet)
 			printf("done\n");
 
@@ -512,6 +511,8 @@ get_state(int argc, char **argv)
 	unsigned int blocksize = 0;
 	unsigned int cluster_size = 0;
 	char *vol_label = NULL;
+	char *stack_name = NULL;
+	char *cluster_name = NULL;
 	unsigned int initial_slots = 0;
 	char *dummy;
 	State *s;
@@ -523,8 +524,10 @@ get_state(int argc, char **argv)
 	uint64_t val;
 	uint64_t journal_size_in_bytes = 0;
 	enum ocfs2_fs_types fs_type = FS_DEFAULT;
-	int mount = 0;
-	int no_backup_super = 0;
+	int mount = -1;
+	int no_backup_super = -1;
+	enum ocfs2_feature_levels level = OCFS2_FEATURE_LEVEL_DEFAULT;
+	ocfs2_fs_options feature_flags = {0,0,0}, reverse_flags = {0,0,0};
 
 	static struct option long_options[] = {
 		{ "block-size", 1, 0, 'b' },
@@ -539,6 +542,10 @@ get_state(int argc, char **argv)
 		{ "force", 0, 0, 'F'},
 		{ "mount", 1, 0, 'M'},
 		{ "no-backup-super", 0, 0, BACKUP_SUPER_OPTION },
+		{ "fs-feature-level=", 1, 0, FEATURE_LEVEL },
+		{ "fs-features=", 1, 0, FEATURES_OPTION },
+		{ "cluster-stack=", 1, 0, CLUSTER_STACK_OPTION },
+		{ "cluster-name=", 1, 0, CLUSTER_NAME_OPTION },
 		{ 0, 0, 0, 0}
 	};
 
@@ -673,6 +680,58 @@ get_state(int argc, char **argv)
 			no_backup_super = 1;
 			break;
 
+		case FEATURE_LEVEL:
+			ret = ocfs2_parse_feature_level(optarg, &level);
+			if (ret) {
+				com_err(progname, ret,
+					"when parsing fs-feature-level string");
+				exit(1);
+			}
+			break;
+
+		case FEATURES_OPTION:
+			ret = ocfs2_parse_feature(optarg,
+						  &feature_flags,
+						  &reverse_flags);
+			if (ret) {
+				com_err(progname, ret,
+					"when parsing fs-features string");
+				exit(1);
+			}
+			break;
+
+		case CLUSTER_STACK_OPTION:
+			if (!optarg || !strlen(optarg)) {
+				com_err(progname, 0,
+					"Option --cluster-stack requires an argument");
+				exit(1);
+			}
+			if (strlen(optarg) != OCFS2_STACK_LABEL_LEN) {
+				com_err(progname, 0,
+					"Invalid argument to --cluster-stack");
+				exit(1);
+			}
+			if (stack_name)
+				free(stack_name);
+			stack_name = strdup(optarg);
+			break;
+
+		case CLUSTER_NAME_OPTION:
+			if (!optarg || !strlen(optarg)) {
+				com_err(progname, 0,
+					"Option --cluster-name requires an argument");
+				exit(1);
+			}
+			if (strlen(optarg) > OCFS2_CLUSTER_NAME_LEN) {
+				com_err(progname, 0,
+					"Cluster name is too long");
+				exit(1);
+			}
+			if (cluster_name)
+				free(cluster_name);
+			cluster_name = strdup(optarg);
+			break;
+
 		default:
 			usage(progname);
 			break;
@@ -735,9 +794,45 @@ get_state(int argc, char **argv)
 
 	s->fs_type = fs_type;
 
-	s->mount = mount;
+	ret = ocfs2_merge_feature_flags_with_level(&s->feature_flags,
+						   level,
+						   &feature_flags,
+						   &reverse_flags);
+	if (ret) {
+		com_err(s->progname, ret,
+			"while reconciling specified features with chosen "
+			"defaults");
+		exit(1);
+	}
 
-	s->no_backup_super = no_backup_super;
+	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT)
+		s->mount = MOUNT_LOCAL;
+	else
+		s->mount = MOUNT_CLUSTER;
+	if (s->feature_flags.opt_compat & OCFS2_FEATURE_COMPAT_BACKUP_SB)
+		s->no_backup_super = 0;
+	else
+		s->no_backup_super = 1;
+
+
+	/* Here if the user set these flags explicitly, we will use them and
+	 * discard the setting in the features set.
+	 */
+	if (mount != -1)
+		s->mount = mount;
+
+	if ((stack_name || cluster_name) && (s->mount == MOUNT_LOCAL)) {
+		com_err(progname, 0,
+			"Local mount is incompatible with specifying a cluster stack");
+		exit(1);
+	}
+	if (stack_name)
+		s->cluster_stack = stack_name;
+	if (cluster_name)
+		s->cluster_name = cluster_name;
+
+	if (no_backup_super != -1)
+		s->no_backup_super = no_backup_super;
 
 	return s;
 }
@@ -851,6 +946,8 @@ usage(const char *progname)
 	fprintf(stderr, "usage: %s [-b block-size] [-C cluster-size] "
 		"[-J journal-options]\n\t\t[-L volume-label] [-M mount-type] "
 		"[-N number-of-node-slots]\n\t\t[-T filesystem-type] [-HFqvV] "
+		"\n\t\t[--fs-feature-level=[default|max-compat|max-features]] "
+		"\n\t\t[--fs-features=[[no]sparse,...]]"
 		"[--no-backup-super] device [blocks-count]\n", progname);
 	exit(0);
 }
@@ -940,21 +1037,18 @@ static uint64_t figure_journal_size(uint64_t size, State *s)
 
 static uint32_t cluster_size_default(State *s)
 {
-	uint32_t volume_size, cluster_size, cluster_size_bits, need;
+	uint32_t cluster_size, cluster_size_bits;
+	uint64_t volume_size;
 
 	for (cluster_size = OCFS2_MIN_CLUSTERSIZE;
-	     cluster_size < AUTO_CLUSTERSIZE;
+	     cluster_size < OCFS2_MAX_CLUSTERSIZE;
 	     cluster_size <<= 1) {
 		cluster_size_bits = get_bits(s, cluster_size);
 
 		volume_size =
 			s->volume_size_in_bytes >> cluster_size_bits;
 
-		need = (volume_size + 7) >> 3;
-		need = ((need + cluster_size - 1) >>
-			cluster_size_bits) << cluster_size_bits;
-
-		if (need <= BITMAP_AUTO_MAX) 
+		if (volume_size <= CLUSTERS_MAX)
 			break;
 	}
 
@@ -1147,7 +1241,11 @@ fill_defaults(State *s)
 	printf("nr_cluster_groups = %u\n", s->nr_cluster_groups);
 	printf("tail_group_bits = %u\n", s->tail_group_bits);
 #endif
-	if (!s->initial_slots) {
+
+	if (s->hb_dev)
+		s->initial_slots = 0;
+
+	if (!s->hb_dev && !s->initial_slots) {
 		if (s->mount == MOUNT_LOCAL)
 			s->initial_slots = 1;
 		else
@@ -1245,7 +1343,7 @@ initialize_alloc_group(State *s, const char *name,
 	group->gd = do_malloc(s, s->blocksize);
 	memset(group->gd, 0, s->blocksize);
 
-	strcpy(group->gd->bg_signature, OCFS2_GROUP_DESC_SIGNATURE);
+	strcpy((char *)group->gd->bg_signature, OCFS2_GROUP_DESC_SIGNATURE);
 	group->gd->bg_generation = s->vol_generation;
 	group->gd->bg_size = (uint32_t)ocfs2_group_bitmap_size(s->blocksize);
 	group->gd->bg_bits = cpg * bpc;
@@ -1716,13 +1814,12 @@ format_superblock(State *s, SystemFileDiskRecord *rec,
 		  SystemFileDiskRecord *root_rec, SystemFileDiskRecord *sys_rec)
 {
 	struct ocfs2_dinode *di;
-	uint32_t incompat;
 	uint64_t super_off = rec->fe_off;
 
 	di = do_malloc(s, s->blocksize);
 	memset(di, 0, s->blocksize);
 
-	strcpy(di->i_signature, OCFS2_SUPER_BLOCK_SIGNATURE);
+	strcpy((char *)di->i_signature, OCFS2_SUPER_BLOCK_SIGNATURE);
 	di->i_suballoc_slot = (__u16)OCFS2_INVALID_SLOT;
 	di->i_suballoc_bit = (__u16)-1;
 	di->i_generation = s->vol_generation;
@@ -1750,16 +1847,40 @@ format_superblock(State *s, SystemFileDiskRecord *rec,
 	di->id2.i_super.s_max_slots = s->initial_slots;
 	di->id2.i_super.s_first_cluster_group = s->first_cluster_group_blkno;
 
-	incompat = 0;
-	if (s->hb_dev)
-		incompat |= OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV;
+	if (s->hb_dev) {
+		s->feature_flags.opt_incompat =
+				 	OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV;
+		s->feature_flags.opt_compat = 0;
+		s->feature_flags.opt_ro_compat = 0;
+	}
 
 	if (s->mount == MOUNT_LOCAL)
-		incompat |= OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT;
+		s->feature_flags.opt_incompat |=
+			OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT;
 
-	di->id2.i_super.s_feature_incompat = incompat;
+	if (s->cluster_stack) {
+		s->feature_flags.opt_incompat |=
+			(OCFS2_FEATURE_INCOMPAT_EXTENDED_SLOT_MAP|
+			 OCFS2_FEATURE_INCOMPAT_USERSPACE_STACK);
+		memcpy(di->id2.i_super.s_cluster_info.ci_stack,
+		       s->cluster_stack, OCFS2_STACK_LABEL_LEN);
+		memcpy(di->id2.i_super.s_cluster_info.ci_cluster,
+		       s->cluster_name, OCFS2_CLUSTER_NAME_LEN);
+	}
 
-	strcpy(di->id2.i_super.s_label, s->vol_label);
+	/*
+	 * we clear the "backup_sb" here since it should be written by
+	 * format_backup_super, not by us. And we have already set the
+	 * "s->no_backup_super" according to the features in get_state,
+	 * so it is safe to clear the flag here.
+	 */
+	s->feature_flags.opt_compat &= !OCFS2_FEATURE_COMPAT_BACKUP_SB;
+
+	di->id2.i_super.s_feature_incompat = s->feature_flags.opt_incompat;
+	di->id2.i_super.s_feature_compat = s->feature_flags.opt_compat;
+	di->id2.i_super.s_feature_ro_compat = s->feature_flags.opt_ro_compat;
+
+	strcpy((char *)di->id2.i_super.s_label, s->vol_label);
 	memcpy(di->id2.i_super.s_uuid, s->uuid, 16);
 
 	ocfs2_swap_inode_from_cpu(di);
@@ -1809,7 +1930,7 @@ format_file(State *s, SystemFileDiskRecord *rec)
 	di = do_malloc(s, s->blocksize);
 	memset(di, 0, s->blocksize);
 
-	strcpy(di->i_signature, OCFS2_INODE_SIGNATURE);
+	strcpy((char *)di->i_signature, OCFS2_INODE_SIGNATURE);
 	di->i_generation = s->vol_generation;
 	di->i_fs_generation = s->vol_generation;
 	di->i_suballoc_slot = (__u16)OCFS2_INVALID_SLOT;
@@ -1845,7 +1966,8 @@ format_file(State *s, SystemFileDiskRecord *rec)
 	if (rec->cluster_bitmap) {
 		di->id2.i_chain.cl_count = 
 			ocfs2_chain_recs_per_inode(s->blocksize);
-		di->id2.i_chain.cl_cpg = s->global_cpg;
+		di->id2.i_chain.cl_cpg =
+				 ocfs2_group_bitmap_size(s->blocksize) * 8;
 		di->id2.i_chain.cl_bpc = 1;
 		if (s->nr_cluster_groups > 
 		    ocfs2_chain_recs_per_inode(s->blocksize)) {
@@ -1899,7 +2021,7 @@ format_file(State *s, SystemFileDiskRecord *rec)
 	if (rec->extent_len) {
 		di->id2.i_list.l_next_free_rec = 1;
 		di->id2.i_list.l_recs[0].e_cpos = 0;
-		di->id2.i_list.l_recs[0].e_clusters = clusters;
+		ocfs2_set_rec_clusters(0, &di->id2.i_list.l_recs[0], clusters);
 		di->id2.i_list.l_recs[0].e_blkno =
 			rec->extent_off >> s->blocksize_bits;
 	}
@@ -1940,7 +2062,7 @@ write_bitmap_data(State *s, AllocBitmap *bitmap)
 	parent_blkno = bitmap->bm_record->fe_off >> s->blocksize_bits;
 	for(i = 0; i < s->nr_cluster_groups; i++) {
 		gd = bitmap->groups[i]->gd;
-		if (strcmp(gd->bg_signature, OCFS2_GROUP_DESC_SIGNATURE)) {
+		if (strcmp((char *)gd->bg_signature, OCFS2_GROUP_DESC_SIGNATURE)) {
 			fprintf(stderr, "bad group descriptor!\n");
 			exit(1);
 		}
@@ -1976,29 +2098,6 @@ write_directory_data(State *s, DirData *dir)
 					      dir->record->file_size);
 }
 
-static void
-write_slot_map_data(State *s, SystemFileDiskRecord *slot_map_rec)
-{
-	int i, num;
-	int16_t *slot_map;
-
-	/* we only use the 1st block of this file, the rest is zero'd
-	 * out. */
-	num = s->blocksize / sizeof(int16_t);
-
-	slot_map = do_malloc(s, slot_map_rec->extent_len);
-	memset(slot_map, 0, slot_map_rec->extent_len);
-
-	for(i = 0; i < num; i++)
-		slot_map[i] = -1;
-
-	ocfs2_swap_slot_map(slot_map, num);
-	do_pwrite(s, slot_map, slot_map_rec->extent_len,
-		  slot_map_rec->extent_off);
-
-	free(slot_map);
-}
-
 static void
 format_leading_space(State *s)
 {
@@ -2014,13 +2113,13 @@ format_leading_space(State *s)
 	memset(buf, 2, size);
 
 	hdr = buf;
-	strcpy(hdr->signature, "this is an ocfs2 volume");
-	strcpy(hdr->mount_point, "this is an ocfs2 volume");
+	strcpy((char *)hdr->signature, "this is an ocfs2 volume");
+	strcpy((char *)hdr->mount_point, "this is an ocfs2 volume");
 
 	p += 512;
 	lbl = (struct ocfs1_vol_label *)p;
-	strcpy(lbl->label, "this is an ocfs2 volume");
-	strcpy(lbl->cluster_name, "this is an ocfs2 volume");
+	strcpy((char *)lbl->label, "this is an ocfs2 volume");
+	strcpy((char *)lbl->cluster_name, "this is an ocfs2 volume");
 
 	do_pwrite(s, buf, size, 0);
 	free(buf);
@@ -2268,6 +2367,31 @@ error:
 	exit(1);
 }
 
+static void format_slotmap(State *s)
+{
+	errcode_t ret;
+	ocfs2_filesys *fs;
+
+	ret = ocfs2_open(s->device_name, OCFS2_FLAG_RW, 0, 0, &fs);
+	if (ret) {
+		com_err(s->progname, ret, "while opening new file system");
+		goto error;
+	}
+
+	ret = ocfs2_format_slot_map(fs);
+	if (ret) {
+	    com_err(s->progname, ret, "while formatting the slot map");
+	    goto error;
+	}
+
+	ocfs2_close(fs);
+	return;
+
+error:
+	clear_both_ends(s);
+	exit(1);
+}
+
 static int format_backup_super(State *s)
 {
 	errcode_t ret;
@@ -2282,9 +2406,9 @@ static int format_backup_super(State *s)
 		goto error;
 	}
 
-	len = ocfs2_get_backup_super_offset(fs, blocks, ARRAY_SIZE(blocks));
+	len = ocfs2_get_backup_super_offsets(fs, blocks, ARRAY_SIZE(blocks));
 
-	ret = ocfs2_set_backup_super(fs, blocks, len);
+	ret = ocfs2_set_backup_super_list(fs, blocks, len);
 	if (ret) {
 		com_err(s->progname, ret, "while backing up superblock.");
 		goto error;
@@ -2293,7 +2417,8 @@ static int format_backup_super(State *s)
 	OCFS2_SET_COMPAT_FEATURE(OCFS2_RAW_SB(fs->fs_super),
 				 OCFS2_FEATURE_COMPAT_BACKUP_SB);
 
-	ret = ocfs2_write_super(fs);
+	/* ocfs2_set_backup_super_list() wrote the backups */
+	ret = ocfs2_write_primary_super(fs);
 	if (ret) {
 		com_err(s->progname, ret, "while updating superblock.");
 		goto error;
