@@ -24,8 +24,7 @@
  */
 
 
-#include <ocfs2.h>
-#include <ocfs2_fs.h>
+#include "ocfs2/ocfs2.h"
 
 
 #undef offsetof
@@ -33,15 +32,15 @@
 #define ssizeof(TYPE, MEMBER) ((unsigned int) sizeof(((TYPE *)0)->MEMBER))
 
 #define START_TYPE(TYPE) do { \
-    printf("[off]\t%- 20s\t[size]\n", #TYPE); \
+    printf("[off]\t%-20s\t[size]\n", #TYPE); \
 } while (0)
 
 #define SHOW_OFFSET(TYPE, MEMBER) do { \
-    printf("0x%03X\t%- 20s\t+0x%02X\n", offsetof(TYPE, MEMBER), #MEMBER, ssizeof(TYPE, MEMBER)); \
+    printf("0x%03X\t%-20s\t+0x%02X\n", offsetof(TYPE, MEMBER), #MEMBER, ssizeof(TYPE, MEMBER)); \
 } while (0)
 
 #define END_TYPE(TYPE) do {\
-    printf("\t%- 20s\t0x%03X\n", "Total", sizeof(*((TYPE *)0))); \
+    printf("\t%-20s\t0x%03X\n", "Total", sizeof(*((TYPE *)0))); \
 } while (0)
 
 static void print_ocfs2_extent_rec(void)
@@ -49,7 +48,7 @@ static void print_ocfs2_extent_rec(void)
 	START_TYPE(ocfs2_extent_rec);
 
 	SHOW_OFFSET(struct ocfs2_extent_rec, e_cpos);
-	SHOW_OFFSET(struct ocfs2_extent_rec, e_clusters);
+	SHOW_OFFSET(struct ocfs2_extent_rec, e_int_clusters);
 	SHOW_OFFSET(struct ocfs2_extent_rec, e_blkno);
 
 	END_TYPE(struct ocfs2_extent_rec);
@@ -137,8 +136,8 @@ static void print_ocfs2_super_block(void)
 	SHOW_OFFSET(struct ocfs2_super_block, s_blocksize_bits);
 	SHOW_OFFSET(struct ocfs2_super_block, s_clustersize_bits);
 	SHOW_OFFSET(struct ocfs2_super_block, s_max_slots);
+	SHOW_OFFSET(struct ocfs2_super_block, s_tunefs_flag);
 	SHOW_OFFSET(struct ocfs2_super_block, s_reserved1);
-	SHOW_OFFSET(struct ocfs2_super_block, s_reserved2);
 	SHOW_OFFSET(struct ocfs2_super_block, s_first_cluster_group);
 	SHOW_OFFSET(struct ocfs2_super_block, s_label);
 	SHOW_OFFSET(struct ocfs2_super_block, s_uuid);
@@ -188,7 +187,7 @@ static void print_ocfs2_dinode(void)
 	SHOW_OFFSET(struct ocfs2_dinode, i_ctime_nsec);
 	SHOW_OFFSET(struct ocfs2_dinode, i_mtime_nsec);
 	SHOW_OFFSET(struct ocfs2_dinode, i_attr);
-	SHOW_OFFSET(struct ocfs2_dinode, i_reserved1);
+	SHOW_OFFSET(struct ocfs2_dinode, i_dyn_features);
 	SHOW_OFFSET(struct ocfs2_dinode, i_reserved2);
 
 	SHOW_OFFSET(struct ocfs2_dinode, id1.i_pad1);
@@ -196,7 +195,7 @@ static void print_ocfs2_dinode(void)
 	SHOW_OFFSET(struct ocfs2_dinode, id1.bitmap1.i_used);
 	SHOW_OFFSET(struct ocfs2_dinode, id1.bitmap1.i_total);
 	SHOW_OFFSET(struct ocfs2_dinode, id1.journal1.ij_flags);
-	SHOW_OFFSET(struct ocfs2_dinode, id1.journal1.ij_pad);
+	SHOW_OFFSET(struct ocfs2_dinode, id1.journal1.ij_recovery_generation);
 
 	SHOW_OFFSET(struct ocfs2_dinode, id2.i_super);
 	SHOW_OFFSET(struct ocfs2_dinode, id2.i_lab);
@@ -243,7 +242,7 @@ static void print_ocfs2_group_desc(void)
         printf("\n");
 }
 
-int main()
+int main(int argc, char **argv)
 {
 	print_ocfs2_extent_rec();
 	print_ocfs2_chain_rec();
