@@ -1,4 +1,6 @@
-/*
+/* -*- mode: c; c-basic-offset: 8; -*-
+ * vim: noexpandtab sw=8 ts=8 sts=0:
+ *
  * tune.c
  *
  * ocfs2 tune utility
@@ -23,83 +25,26 @@
  * Authors: Sunil Mushran
  */
 
-#define _LARGEFILE64_SOURCE
-#define _GNU_SOURCE /* Because libc really doesn't want us using O_DIRECT? */
+#include <tunefs.h>
 
-#include <sys/types.h>
-#include <stdio.h>
-#include <unistd.h>
-#include <errno.h>
-#include <fcntl.h>
-#include <linux/fd.h>
-#include <string.h>
-#include <sys/stat.h>
-#include <stdlib.h>
-#include <string.h>
-#include <getopt.h>
-#include <malloc.h>
-#include <time.h>
-#include <libgen.h>
-#include <netinet/in.h>
-#include <inttypes.h>
-#include <ctype.h>
-#include <signal.h>
-#include <uuid/uuid.h>
+#define WHOAMI "tunefs.ocfs2"
 
-#include <ocfs2.h>
-#include <ocfs2_fs.h>
-#include <ocfs1_fs_compat.h>
-#include <bitops.h>
-
-#include <jbd.h>
-
-#include <kernel-list.h>
-
-#define SYSTEM_FILE_NAME_MAX   40
-
-#ifndef MIN
-#define MIN(a, b) (((a) < (b)) ? (a) : (b))
-#endif
-#ifndef MAX
-#define MAX(a, b) (((a) > (b)) ? (a) : (b))
-#endif
-
-#define MOUNT_LOCAL             1
-#define MOUNT_CLUSTER           2
-#define MOUNT_LOCAL_STR         "local"
-#define MOUNT_CLUSTER_STR       "cluster"
-
-enum {
-	BACKUP_SUPER_OPTION = CHAR_MAX + 1,
-};
-
-typedef struct _ocfs2_tune_opts {
-	uint16_t num_slots;
-	uint64_t num_blocks;
-	uint64_t jrnl_size;
-	char *vol_label;
-	char *progname;
-	char *device;
-	char *vol_uuid;
-	int mount;
-	int verbose;
-	int quiet;
-	int prompt;
-	int backup_super;
-	time_t tune_time;
-	int fd;
-} ocfs2_tune_opts;
-
-static ocfs2_tune_opts opts;
-static ocfs2_filesys *fs_gbl = NULL;
+ocfs2_tune_opts opts;
+ocfs2_filesys *fs_gbl = NULL;
 static int cluster_locked = 0;
 static int resize = 0;
+static int online_resize = 0;
+static uint64_t def_jrnl_size = 0;
+static char old_uuid[OCFS2_VOL_UUID_LEN * 2 + 1];
+static char new_uuid[OCFS2_VOL_UUID_LEN * 2 + 1];
 
 static void usage(const char *progname)
 {
 	fprintf(stderr, "usage: %s [-J journal-options] [-L volume-label]\n"
-			"\t\t[-M mount-type] [-N number-of-node-slots]\n"
-			"\t\t[-qSUvV] [--backup-super] device [blocks-count]\n",
+			"\t\t[-M mount-type] [-N number-of-node-slots] [-Q query-fmt]\n"
+			"\t\t[--update-cluster-stack]\n"
+			"\t\t[-qSUvV] [--backup-super] [--list-sparse]\n"
+			"\t\t[--fs-features=[no]sparse,...]] device [blocks-count]\n",
 			progname);
 	exit(0);
 }
@@ -120,7 +65,7 @@ static void handle_signal(int sig)
 			ocfs2_release_cluster(fs_gbl);
 
 		if (fs_gbl && fs_gbl->fs_dlm_ctxt)
-			ocfs2_shutdown_dlm(fs_gbl);
+			ocfs2_shutdown_dlm(fs_gbl, WHOAMI);
 
 		exit(1);
 	}
@@ -129,7 +74,7 @@ static void handle_signal(int sig)
 }
 
 /* Call this with SIG_BLOCK to block and SIG_UNBLOCK to unblock */
-static void block_signals(int how)
+void block_signals(int how)
 {
      sigset_t sigs;
 
@@ -247,18 +192,23 @@ static void get_options(int argc, char **argv)
 	int show_version = 0;
 	int uuid = 0;
 	char *dummy;
+	errcode_t ret = 0;
 
 	static struct option long_options[] = {
 		{ "label", 1, 0, 'L' },
 		{ "node-slots", 1, 0, 'N' },
 		{ "verbose", 0, 0, 'v' },
 		{ "quiet", 0, 0, 'q' },
+		{ "query", 1, 0, 'Q' },
 		{ "version", 0, 0, 'V' },
 		{ "journal-options", 0, 0, 'J'},
 		{ "volume-size", 0, 0, 'S'},
 		{ "uuid-reset", 0, 0, 'U'},
 		{ "mount", 1, 0, 'M' },
 		{ "backup-super", 0, 0, BACKUP_SUPER_OPTION },
+		{ "update-cluster-stack", 0, 0, UPDATE_CLUSTER_OPTION },
+		{ "list-sparse", 0, 0, LIST_SPARSE_FILES },
+		{ "fs-features=", 1, 0, FEATURES_OPTION },
 		{ 0, 0, 0, 0}
 	};
 
@@ -270,7 +220,7 @@ static void get_options(int argc, char **argv)
 	opts.prompt = 1;
 
 	while (1) {
-		c = getopt_long(argc, argv, "L:N:J:M:SUvqVxb", long_options,
+		c = getopt_long(argc, argv, "L:N:J:M:Q:SUvqVxb", long_options,
 				NULL);
 
 		if (c == -1)
@@ -313,13 +263,16 @@ static void get_options(int argc, char **argv)
 					"than %d",
 					OCFS2_MAX_SLOTS);
 				exit(1);
-			} else if (opts.num_slots < 2) {
+			} else if (opts.num_slots < 1) {
 				com_err(opts.progname, 0,
 					"Number of node slots must be at "
-					"least 2");
+					"least 1");
 				exit(1);
 			}
 			break;
+		case 'Q':
+			opts.queryfmt = strdup(optarg);
+			break;
 
 		case 'J':
 			parse_journal_opts(opts.progname, optarg,
@@ -354,6 +307,26 @@ static void get_options(int argc, char **argv)
 			opts.backup_super = 1;
 			break;
 
+		case UPDATE_CLUSTER_OPTION:
+			opts.update_cluster = 1;
+			break;
+
+		case LIST_SPARSE_FILES:
+			opts.list_sparse = 1;
+			break;
+
+		case FEATURES_OPTION:
+			ret = ocfs2_parse_feature(optarg,
+						  &opts.set_feature,
+						  &opts.clear_feature);
+			if (ret) {
+				com_err(opts.progname, ret,
+					"when parsing --fs-features string");
+				exit(1);
+			}
+			opts.feature_string = strdup(optarg);
+			break;
+
 		default:
 			usage(opts.progname);
 			break;
@@ -365,7 +338,9 @@ static void get_options(int argc, char **argv)
 	 */
 	if (opts.backup_super &&
 	    (opts.vol_label || opts.num_slots ||
-	     opts.mount || opts.jrnl_size || resize)) {
+	     opts.mount || opts.jrnl_size || resize ||
+	     opts.list_sparse || opts.feature_string ||
+	     opts.update_cluster)) {
 		com_err(opts.progname, 0, "Cannot backup superblock"
 			" along with other tasks");
 		exit(1);
@@ -380,6 +355,45 @@ static void get_options(int argc, char **argv)
 		exit(0);
 	}
 
+	/*
+	 * We don't allow list-sparse to be coexist with other tunefs
+	 * options to keep things simple.
+	 */
+	if (opts.list_sparse &&
+	    (opts.vol_label || opts.num_slots ||
+	     opts.mount || opts.jrnl_size || resize || opts.backup_super ||
+	     opts.feature_string || opts.update_cluster)) {
+		com_err(opts.progname, 0, "Cannot list sparse files"
+			" along with other tasks");
+		exit(1);
+	}
+
+	/*
+	 * We don't allow feature-modifation to be coexist with other tunefs
+	 * options to keep things simple.
+	 */
+	if (opts.feature_string &&
+	    (opts.vol_label || opts.num_slots ||
+	     opts.mount || opts.jrnl_size || resize || opts.backup_super ||
+	     opts.list_sparse || opts.update_cluster)) {
+		com_err(opts.progname, 0, "Cannot modify fs features"
+			" along with other tasks");
+		exit(1);
+	}
+
+	/*
+	 * We don't allow cluster information to be modified with other
+	 * tunefs options to keep things simple.
+	 */
+	if (opts.update_cluster &&
+	    (opts.vol_label || opts.num_slots || opts.feature_string ||
+	     opts.mount || opts.jrnl_size || resize || opts.backup_super ||
+	     opts.list_sparse)) {
+		com_err(opts.progname, 0, "Cannot modify cluster stack"
+			" along with other tasks");
+		exit(1);
+	}
+
 	if (show_version)
 		exit(0);
 
@@ -421,31 +435,6 @@ static void get_options(int argc, char **argv)
 	return ;
 }
 
-static void get_vol_size(ocfs2_filesys *fs)
-{
-	errcode_t ret = 0;
-	uint64_t num_blocks;
-
-	ret = ocfs2_get_device_size(opts.device, fs->fs_blocksize,
-				    &num_blocks);
-	if (ret) {
-		com_err(opts.progname, ret, "while getting size of device %s",
-			opts.device);
-		exit(1);
-	}
-
-	if (!opts.num_blocks)
-		opts.num_blocks = num_blocks;
-
-	if (opts.num_blocks > num_blocks) {
-		com_err(opts.progname, 0, "The containing partition (or device) "
-			"is only %"PRIu64" blocks", num_blocks);
-		exit(1);
-	}
-
-	return ;
-}
-
 static int validate_mount_change(ocfs2_filesys *fs)
 {
 	if (opts.mount == MOUNT_LOCAL) {
@@ -459,40 +448,6 @@ static int validate_mount_change(ocfs2_filesys *fs)
 	return -1;
 }
 
-static int validate_vol_size(ocfs2_filesys *fs)
-{
-	uint64_t num_blocks;
-
-	if (opts.num_blocks == fs->fs_blocks) {
-		com_err(opts.progname, 0, "The filesystem is already "
-			"%"PRIu64" blocks", fs->fs_blocks);
-		return -1;
-	}
-
-	if (opts.num_blocks < fs->fs_blocks) {
-		com_err(opts.progname, 0, "Cannot shrink volume size from "
-		       "%"PRIu64" blocks to %"PRIu64" blocks",
-		       fs->fs_blocks, opts.num_blocks);
-		return -1;
-	}
-
-	num_blocks = ocfs2_clusters_to_blocks(fs, 1);
-	if (num_blocks > (opts.num_blocks - fs->fs_blocks)) {
-		com_err(opts.progname, 0, "Cannot grow volume size less than "
-		       "%d blocks", num_blocks);
-		return -1;
-	}
-
-	if (opts.num_blocks > UINT32_MAX) {
-		com_err(opts.progname, 0, "As JBD can only store block numbers "
-			"in 32 bits, %s cannot be grown to more than %"PRIu64" "
-			"blocks.", opts.device, UINT32_MAX);
-		return -1;
-	}
-
-	return 0;
-}
-
 static errcode_t add_slots(ocfs2_filesys *fs)
 {
 	errcode_t ret = 0;
@@ -633,7 +588,7 @@ bail:
 	return ret;
 }
 
-static void get_total_free_bits(struct ocfs2_group_desc *gd, int *bits)
+static void get_total_free_bits(struct ocfs2_group_desc *gd, uint16_t *bits)
 {
 	int end = 0;
 	int start;
@@ -660,7 +615,7 @@ static errcode_t validate_chain_group(ocfs2_filesys *fs, struct ocfs2_dinode *di
 	struct ocfs2_chain_rec *cr;
 	uint32_t total = 0;
 	uint32_t free = 0;
-	uint32_t bits;
+	uint16_t bits;
 
 	ret = ocfs2_malloc_block(fs->fs_io, &buf);
 	if (ret) {
@@ -692,7 +647,7 @@ static errcode_t validate_chain_group(ocfs2_filesys *fs, struct ocfs2_dinode *di
 			ret = OCFS2_ET_CORRUPT_CHAIN;
 			com_err(opts.progname, ret, " - group descriptor at "
 				"%"PRIu64" does not belong to allocator %"PRIu64"",
-				blkno, di->i_blkno);
+				blkno, (uint64_t)di->i_blkno);
 			goto bail;
 		}
 
@@ -701,7 +656,7 @@ static errcode_t validate_chain_group(ocfs2_filesys *fs, struct ocfs2_dinode *di
 			com_err(opts.progname, ret, " - group descriptor at "
 				"%"PRIu64" does not agree to the chain it "
 				"belongs to in allocator %"PRIu64"",
-				blkno, di->i_blkno);
+				blkno, (uint64_t)di->i_blkno);
 			goto bail;
 		}
 
@@ -710,7 +665,7 @@ static errcode_t validate_chain_group(ocfs2_filesys *fs, struct ocfs2_dinode *di
 			ret = OCFS2_ET_CORRUPT_CHAIN;
 			com_err(opts.progname, ret, " - group descriptor at "
 				"%"PRIu64" does not have a consistent free "
-				"bit count", blkno);
+				"bit count", (uint64_t)blkno);
 			goto bail;
 		}
 
@@ -718,7 +673,7 @@ static errcode_t validate_chain_group(ocfs2_filesys *fs, struct ocfs2_dinode *di
 			ret = OCFS2_ET_CORRUPT_CHAIN;
 			com_err(opts.progname, ret, " - group descriptor at "
 				"%"PRIu64" does not have a valid total bit "
-				"count", blkno);
+				"count", (uint64_t)blkno);
 			goto bail;
 		}
 
@@ -739,7 +694,7 @@ static errcode_t validate_chain_group(ocfs2_filesys *fs, struct ocfs2_dinode *di
 		ret = OCFS2_ET_CORRUPT_CHAIN;
 		com_err(opts.progname, ret, " - total bits for chain %u in "
 			"allocator %"PRIu64" does not match its chained group "
-			"descriptors", chain, di->i_blkno);
+			"descriptors", chain, (uint64_t)di->i_blkno);
 		goto bail;
 
 	}
@@ -748,7 +703,7 @@ static errcode_t validate_chain_group(ocfs2_filesys *fs, struct ocfs2_dinode *di
 		ret = OCFS2_ET_CORRUPT_CHAIN;
 		com_err(opts.progname, ret, " - free bits for chain %u in "
 			"allocator %"PRIu64" does not match its chained group "
-			"descriptors", chain, di->i_blkno);
+			"descriptors", chain, (uint64_t)di->i_blkno);
 		goto bail;
 	}
 
@@ -841,7 +796,8 @@ static errcode_t backup_super_check(ocfs2_filesys *fs)
 		return -1;
 	}
 
-	num = ocfs2_get_backup_super_offset(fs, blocks, ARRAY_SIZE(blocks));
+	num = ocfs2_get_backup_super_offsets(fs, blocks,
+					     ARRAY_SIZE(blocks));
 	if (!num) {
 		com_err(opts.progname, 0,
 			"Volume is too small to hold backup superblocks");
@@ -885,7 +841,7 @@ static void update_volume_label(ocfs2_filesys *fs, int *changed)
 {
   	memset (OCFS2_RAW_SB(fs->fs_super)->s_label, 0,
 		OCFS2_MAX_VOL_LABEL_LEN);
-	strncpy (OCFS2_RAW_SB(fs->fs_super)->s_label, opts.vol_label,
+	strncpy ((char *)OCFS2_RAW_SB(fs->fs_super)->s_label, opts.vol_label,
 		 OCFS2_MAX_VOL_LABEL_LEN);
 
 	*changed = 1;
@@ -903,34 +859,75 @@ static void update_volume_uuid(ocfs2_filesys *fs, int *changed)
 	return ;
 }
 
-static void update_mount_type(ocfs2_filesys *fs, int *changed)
+static errcode_t update_mount_type(ocfs2_filesys *fs, int *changed)
 {
-	if (opts.mount == MOUNT_LOCAL)
-		OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat |=
+	errcode_t ret = 0;
+	struct o2cb_cluster_desc desc;
+	struct ocfs2_super_block *sb = OCFS2_RAW_SB(fs->fs_super);
+
+	if (opts.mount == MOUNT_LOCAL) {
+		sb->s_feature_incompat |=
 			OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT;
-	else if (opts.mount == MOUNT_CLUSTER)
-		OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat &=
+		sb->s_feature_incompat &=
+			~OCFS2_FEATURE_INCOMPAT_USERSPACE_STACK;
+	} else if (opts.mount == MOUNT_CLUSTER) {
+		ret = o2cb_init();
+		if (ret)
+			goto out;
+		ret = o2cb_running_cluster_desc(&desc);
+		if (ret)
+			goto out;
+		sb->s_feature_incompat &=
 			~OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT;
-	else
-		return;
+		ret = ocfs2_set_cluster_desc(fs, &desc);
+		o2cb_free_cluster_desc(&desc);
+		if (ret)
+			goto out;
+	} else
+		goto out;
 
 	*changed = 1;
 
-	return;
+out:
+	return ret;
+}
+
+static errcode_t update_cluster(ocfs2_filesys *fs)
+{
+	errcode_t ret;
+	struct o2cb_cluster_desc desc;
+
+	ret = o2cb_running_cluster_desc(&desc);
+	if (!ret) {
+		ret = ocfs2_set_cluster_desc(fs, &desc);
+		o2cb_free_cluster_desc(&desc);
+	}
+
+	return ret;
 }
 
 static errcode_t update_slots(ocfs2_filesys *fs, int *changed)
 {
 	errcode_t ret = 0;
+	int orig_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
 
 	block_signals(SIG_BLOCK);
-	ret = add_slots(fs);
-	block_signals(SIG_UNBLOCK);
+	if (opts.num_slots > orig_slots)
+		ret = add_slots(fs);
+	else
+		ret = remove_slots(fs);
 	if (ret)
-		return ret;
+		goto unblock;
 
 	OCFS2_RAW_SB(fs->fs_super)->s_max_slots = opts.num_slots;
-	*changed = 1;
+	ret = ocfs2_format_slot_map(fs);
+	if (!ret)
+		*changed = 1;
+	else
+		OCFS2_RAW_SB(fs->fs_super)->s_max_slots = orig_slots;
+
+unblock:
+	block_signals(SIG_UNBLOCK);
 
 	return ret;
 }
@@ -976,7 +973,7 @@ static errcode_t update_journal_size(ocfs2_filesys *fs, int *changed)
 		}
 
 		di = (struct ocfs2_dinode *)buf;
-		if (num_clusters <= di->i_clusters)
+		if (num_clusters == di->i_clusters)
 			continue;
 
 		printf("Updating %s...  ", jrnl_file);
@@ -1004,263 +1001,15 @@ bail:
 	return ret;
 }
 
-static errcode_t update_volume_size(ocfs2_filesys *fs, int *changed)
-{
-	errcode_t ret = 0;
-	struct ocfs2_dinode *di;
-	uint64_t bm_blkno = 0;
-	uint64_t gd_blkno = 0;
-	uint64_t lgd_blkno = 0;
-	char *in_buf = NULL;
-	char *gd_buf = NULL;
-	char *lgd_buf = NULL;
-	struct ocfs2_chain_list *cl;
-	struct ocfs2_chain_rec *cr;
-	struct ocfs2_group_desc *gd;
-	uint32_t cluster_chunk;
-	uint32_t num_new_clusters, save_new_clusters;
-	uint32_t first_new_cluster;
-	uint16_t chain;
-	uint32_t used_bits;
-	uint32_t total_bits;
-	uint32_t num_bits;
-	int flush_lgd = 0;
-	char *zero_buf = NULL;
-
-	ret = ocfs2_malloc_block(fs->fs_io, &in_buf);
-	if (ret) {
-		com_err(opts.progname, ret, "while allocating a block during "
-			"volume resize");
-		goto bail;
-	}
-
-	ret = ocfs2_malloc_block(fs->fs_io, &gd_buf);
-	if (ret) {
-		com_err(opts.progname, ret, "while allocating a block during "
-			"volume resize");
-		goto bail;
-	}
-
-	ret = ocfs2_malloc_block(fs->fs_io, &lgd_buf);
-	if (ret) {
-		com_err(opts.progname, ret, "while allocating a block during "
-			"volume resize");
-		goto bail;
-	}
-
-	ret = ocfs2_malloc_blocks(fs->fs_io, ocfs2_clusters_to_blocks(fs, 1),
-				  &zero_buf);
-	if (ret) {
-		com_err(opts.progname, ret, "while allocating a cluster during "
-			"volume resize");
-		goto bail;
-	}
-
-	memset(zero_buf, 0, fs->fs_clustersize);
-
-	/* read global bitmap */
-	ret = ocfs2_lookup_system_inode(fs, GLOBAL_BITMAP_SYSTEM_INODE, 0,
-					&bm_blkno);
-	if (ret) {
-		com_err(opts.progname, ret, "while looking up global bitmap "
-			"inode during volume resize");
-		goto bail;
-	}
-
-	ret = ocfs2_read_inode(fs, bm_blkno, in_buf);
-	if (ret) {
-		com_err(opts.progname, ret, "while reading inode at block "
-			"%"PRIu64" during volume resize", bm_blkno);
-		goto bail;
-	}
-
-	di = (struct ocfs2_dinode *)in_buf;
-	cl = &(di->id2.i_chain);
-
-	total_bits = di->id1.bitmap1.i_total;
-	used_bits = di->id1.bitmap1.i_used;
-
-	first_new_cluster = di->i_clusters;
-	save_new_clusters = num_new_clusters =
-	       	ocfs2_blocks_to_clusters(fs, opts.num_blocks) - di->i_clusters;
-
-	/* Find the blknum of the last cluster group */
-	lgd_blkno = ocfs2_which_cluster_group(fs, cl->cl_cpg, first_new_cluster - 1);
-
-	ret = ocfs2_read_group_desc(fs, lgd_blkno, lgd_buf);
-	if (ret) {
-		com_err(opts.progname, ret, "while reading group descriptor "
-			"at block %"PRIu64" during volume resize", lgd_blkno);
-		goto bail;
-	}
-
-	gd = (struct ocfs2_group_desc *)lgd_buf;
-
-	/* If only one cluster group then see if we need to adjust up cl_cpg */
-	if (cl->cl_next_free_rec == 1) {
-		if (cl->cl_cpg < 8 * gd->bg_size)
-			cl->cl_cpg = 8 * gd->bg_size;
-	}
-
-	chain = gd->bg_chain;
-
-	/* If possible round off the last group to cpg */
-	cluster_chunk = MIN(num_new_clusters,
-			    (cl->cl_cpg - (gd->bg_bits/cl->cl_bpc)));
-	if (cluster_chunk) {
-		num_new_clusters -= cluster_chunk;
-		first_new_cluster += cluster_chunk;
-
-		num_bits = cluster_chunk * cl->cl_bpc;
-
-		gd->bg_bits += num_bits;
-		gd->bg_free_bits_count += num_bits;
-
-		cr = &(cl->cl_recs[chain]);
-		cr->c_total += num_bits;
-		cr->c_free += num_bits;
-
-		total_bits += num_bits;
-
-		fs->fs_clusters += cluster_chunk;
-		fs->fs_blocks += ocfs2_clusters_to_blocks(fs, cluster_chunk);
-
-		/* This cluster group block is written after the new */
-		/* cluster groups are written to disk */
-		flush_lgd = 1;
-	}
-
-	/* Init the new groups and write to disk */
-	/* Add these groups one by one starting from the first chain after */
-	/* the one containing the last group */
-
-	gd = (struct ocfs2_group_desc *)gd_buf;
-
-	while(num_new_clusters) {
-		gd_blkno = ocfs2_which_cluster_group(fs, cl->cl_cpg,
-						     first_new_cluster);
-		cluster_chunk = MIN(num_new_clusters, cl->cl_cpg);
-		num_new_clusters -= cluster_chunk;
-		first_new_cluster += cluster_chunk;
-
-		if (++chain >= cl->cl_count)
-			chain = 0;
-
-		ocfs2_init_group_desc(fs, gd, gd_blkno,
-				      fs->fs_super->i_fs_generation, di->i_blkno,
-				      (cluster_chunk *cl->cl_bpc), chain);
-
-		/* Add group to chain */
-		cr = &(cl->cl_recs[chain]);
-		if (chain >= cl->cl_next_free_rec) {
-			cl->cl_next_free_rec++;
-			cr->c_free = 0;
-			cr->c_total = 0;
-			cr->c_blkno = 0;
-		}
-
-		gd->bg_next_group = cr->c_blkno;
-		cr->c_blkno = gd_blkno;
-		cr->c_free += gd->bg_free_bits_count;
-		cr->c_total += gd->bg_bits;
-
-		used_bits += (gd->bg_bits - gd->bg_free_bits_count);
-		total_bits += gd->bg_bits;
-
-		fs->fs_clusters += cluster_chunk;
-		fs->fs_blocks += ocfs2_clusters_to_blocks(fs, cluster_chunk);
-
-		/* Initialize the first cluster in the group */
-		ret = io_write_block(fs->fs_io, gd_blkno,
-				     ocfs2_clusters_to_blocks(fs, 1), zero_buf);
-		if (ret) {
-			com_err(opts.progname, ret, "while initializing the "
-				"cluster starting at block %"PRIu64" during "
-				"volume resize", gd_blkno);
-			goto bail;
-		}
-
-		/* write a new group descriptor */
-		ret = ocfs2_write_group_desc(fs, gd_blkno, gd_buf);
-		if (ret) {
-			com_err(opts.progname, ret, "while writing group "
-				"descriptor at block %"PRIu64" during "
-				"volume resize", gd_blkno);
-			goto bail;
-		}
-	}
-
-	di->id1.bitmap1.i_total = total_bits;
-	di->id1.bitmap1.i_used = used_bits;
-
-	di->i_clusters += save_new_clusters;
-	di->i_size = (uint64_t) di->i_clusters * fs->fs_clustersize;
-
-	fs->fs_super->i_clusters = di->i_clusters;
-
-	block_signals(SIG_BLOCK);
-	/* Flush that last group descriptor we updated before the new ones */
-	if (flush_lgd) {
-		ret = ocfs2_write_group_desc(fs, lgd_blkno, lgd_buf);
-		if (ret) {
-			block_signals(SIG_UNBLOCK);
-			com_err(opts.progname, ret, "while flushing group "
-				"descriptor at block %"PRIu64" during "
-				"volume resize", lgd_blkno);
-			goto bail;
-		}
-	}
-
-	/* write the global bitmap inode */
-	ret = ocfs2_write_inode(fs, bm_blkno, in_buf);
-	if (ret) {
-		block_signals(SIG_UNBLOCK);
-		com_err(opts.progname, ret, "while writing global bitmap "
-			"inode at block %"PRIu64" during volume resize",
-			bm_blkno);
- 		goto bail;
-	}
-
-	block_signals(SIG_UNBLOCK);
-
-	*changed = 1;
-
-bail:
-	if (zero_buf)
-		ocfs2_free(&zero_buf);
-	if (in_buf)
-		ocfs2_free(&in_buf);
-	if (gd_buf)
-		ocfs2_free(&gd_buf);
-	if (lgd_buf)
-		ocfs2_free(&lgd_buf);
-
-	return ret;
-}
-
-static errcode_t refresh_backup_super(ocfs2_filesys *fs)
-{
-	errcode_t ret;
-	int num;
-	uint64_t blocks[OCFS2_MAX_BACKUP_SUPERBLOCKS];
-
-	num = ocfs2_get_backup_super_offset(fs, blocks, ARRAY_SIZE(blocks));
-	if (!num)
-		return 0;
-
-	ret = ocfs2_refresh_backup_super(fs, blocks, num);
-
-	return ret;
-}
-
-static errcode_t update_backup_super(ocfs2_filesys *fs, uint64_t newblocks)
+static errcode_t update_backup_super(ocfs2_filesys *fs, uint64_t startblk,
+				     uint64_t newblocks)
 {
 	errcode_t ret;
 	int num, i;
 	uint64_t *new_backup_super, blocks[OCFS2_MAX_BACKUP_SUPERBLOCKS];
-	uint64_t startblk = fs->fs_blocks - newblocks;
 
-	num = ocfs2_get_backup_super_offset(fs, blocks, ARRAY_SIZE(blocks));
+	num = ocfs2_get_backup_super_offsets(fs, blocks,
+					     ARRAY_SIZE(blocks));
 	if (!num)
 		return 0;
 
@@ -1278,7 +1027,7 @@ static errcode_t update_backup_super(ocfs2_filesys *fs, uint64_t newblocks)
 	} else
 		new_backup_super = blocks;
 
-	ret = ocfs2_set_backup_super(fs, new_backup_super, num);
+	ret = ocfs2_set_backup_super_list(fs, new_backup_super, num);
 	if (ret) {
 		com_err(opts.progname, ret, "while backing up superblock.");
 		goto bail;
@@ -1288,93 +1037,58 @@ bail:
 	return ret;
 }
 
-int main(int argc, char **argv)
+static void free_opts(void)
 {
-	errcode_t ret = 0;
-	ocfs2_filesys *fs = NULL;
-	int open_flags = OCFS2_FLAG_RW | OCFS2_FLAG_STRICT_COMPAT_CHECK;
-	int upd_label = 0;
-	int upd_uuid = 0;
-	int upd_slots = 0;
-	int upd_jrnls = 0;
-	int upd_blocks = 0;
-	int upd_mount = 0;
-	int upd_incompat = 0;
-	int upd_backup_super = 0;
-	char *tmpstr;
-	uint16_t tmp;
-	uint64_t def_jrnl_size = 0;
-	uint64_t num_clusters;
+	if (opts.vol_uuid)
+		free(opts.vol_uuid);
+	if (opts.vol_label)
+		free(opts.vol_label);
+	if (opts.queryfmt)
+		free(opts.queryfmt);
+	if (opts.device)
+		free(opts.device);
+}
+
+static errcode_t volume_check(ocfs2_filesys *fs)
+{
+	errcode_t ret;
 	int dirty = 0;
-	char old_uuid[OCFS2_VOL_UUID_LEN * 2 + 1];
-	char new_uuid[OCFS2_VOL_UUID_LEN * 2 + 1];
+	uint16_t max_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
 
-	initialize_ocfs_error_table();
-	initialize_o2dl_error_table();
-	initialize_o2cb_error_table();
-
-	setbuf(stdout, NULL);
-	setbuf(stderr, NULL);
-
-	if (signal(SIGTERM, handle_signal) == SIG_ERR) {
-		fprintf(stderr, "Could not set SIGTERM\n");
-		exit(1);
-	}
-
-	if (signal(SIGINT, handle_signal) == SIG_ERR) {
-		fprintf(stderr, "Could not set SIGINT\n");
-		exit(1);
-	}
-
-	memset (&opts, 0, sizeof(opts));
-
-	get_options(argc, argv);
-
-	/* strict compat flag check */
-	ret = ocfs2_open(opts.device, open_flags, 0, 0, &fs); //O_EXCL?
-	if (ret) {
-		com_err(opts.progname, ret, "while opening device %s",
-			opts.device);
-		goto close;
-	}
-	fs_gbl = fs;
-
-	if (OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat &
-	    OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG) {
-		fprintf(stderr, "Aborted resize detected. Run fsck.ocfs2 -f <device>.\n");
-		goto close;
-	}
-
-	if (resize)
-		get_vol_size(fs);
-
-	if (!ocfs2_mount_local(fs)) {
-		ret = o2cb_init();
-		if (ret) {
-			com_err(opts.progname, ret, "while initializing the cluster");
-			exit(1);
-		}
-
-		ret = ocfs2_initialize_dlm(fs);
-		if (ret) {
-			com_err(opts.progname, ret, "while initializing the dlm");
-			goto close;
-		}
-
-		block_signals(SIG_BLOCK);
-		ret = ocfs2_lock_down_cluster(fs);
-		if (ret) {
-			block_signals(SIG_UNBLOCK);
-			com_err(opts.progname, ret, "while locking down the cluster");
-			goto close;
-		}
-		cluster_locked = 1;
-		block_signals(SIG_UNBLOCK);
+	/*
+	 * online_resize can't coexist with other tasks, and it does't
+	 * need other checks, so we just do the check and return.
+	 */
+	if (online_resize) {
+		ret = online_resize_check(fs);
+		if (ret)
+			com_err(opts.progname, 0, "online resize check failed.");
+		goto bail;
 	}
 
 	ret = journal_check(fs, &dirty, &def_jrnl_size);
 	if (ret || dirty)
-		goto unlock;
+		goto bail;
+
+	ret = 1;
+	if (opts.list_sparse) {
+		if (!ocfs2_sparse_alloc(OCFS2_RAW_SB(fs->fs_super))) {
+			com_err(opts.progname, 0,
+				"sparse_file flag check failed. ");
+			goto bail;
+		}
+		printf("List all the sparse files in the volume\n");
+	}
+
+	if (opts.feature_string) {
+		if (feature_check(fs)) {
+			com_err(opts.progname, 0,
+				"feature check failed. ");
+			goto bail;
+		}
+		printf("Modify feature \"%s\" for the volume\n",
+			opts.feature_string);
+	}
 
 	/* If operation requires touching the global bitmap, ensure it is good */
 	/* This is to handle failed resize */
@@ -1383,18 +1097,41 @@ int main(int argc, char **argv)
 		if (global_bitmap_check(fs)) {
 			com_err(opts.progname, 0, "Global bitmap check failed. "
 				"Run fsck.ocfs2 -f <device>.");
-			goto unlock;
+			goto bail;
 		}
 	}
 
 	/* check whether the block for backup superblock are used. */
 	if (opts.backup_super) {
 		if (backup_super_check(fs))
-			goto unlock;
-		else
-			printf("Adding backup superblock for the volume\n");
+			goto bail;
 	}
 
+	/* remove slot check. */
+	if (opts.num_slots && opts.num_slots < max_slots) {
+		ret = remove_slot_check(fs);
+		if (ret) {
+			com_err(opts.progname, 0,
+				"remove slot check failed. ");
+			goto bail;
+		}
+	}
+
+	ret = 0;
+bail:
+	return ret;
+}
+
+static void validate_parameter(ocfs2_filesys *fs)
+{
+	uint16_t max_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
+	uint64_t num_clusters;
+	char *tmpstr;
+
+	/* valid backup super. */
+	if (opts.backup_super)
+		printf("Adding backup superblock for the volume\n");
+
 	/* validate volume label */
 	if (opts.vol_label) {
 		printf("Changing volume label from %s to %s\n",
@@ -1421,20 +1158,21 @@ int main(int argc, char **argv)
 	}
 
 	/* validate num slots */
+	max_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
 	if (opts.num_slots) {
-		tmp = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
-		if (opts.num_slots > tmp) {
-			printf("Changing number of node slots from %d to %d\n",
-			       tmp, opts.num_slots);
-		} else {
-			com_err(opts.progname, 0, "Node slots (%d) has to be "
-				"more than the configured node slots (%d)",
-			       opts.num_slots, tmp);
-			goto unlock;
+		if (opts.num_slots > max_slots) {
+			if (!opts.jrnl_size)
+				opts.jrnl_size = def_jrnl_size;
+
+		} else if (opts.num_slots == max_slots) {
+			printf("Giving the same number of nodes. "
+				"Ignore the change of slots.");
+			opts.num_slots = 0;
 		}
 
-		if (!opts.jrnl_size)
-			opts.jrnl_size = def_jrnl_size;
+		if (opts.num_slots)
+			printf("Changing number of node slots from %d to %d\n",
+			       max_slots, opts.num_slots);
 	}
 
 	/* validate journal size */
@@ -1445,17 +1183,9 @@ int main(int argc, char **argv)
 		opts.jrnl_size = num_clusters <<
 				OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits;
 
-		if (opts.jrnl_size > def_jrnl_size)
+		if (opts.jrnl_size != def_jrnl_size)
 			printf("Changing journal size %"PRIu64" to %"PRIu64"\n",
 			       def_jrnl_size, opts.jrnl_size);
-		else {
-			if (!opts.num_slots) {
-				com_err(opts.progname, 0, "Journal size %"PRIu64" "
-					"has to be larger " "than %"PRIu64"",
-					opts.jrnl_size, def_jrnl_size);
-				goto unlock;
-			}
-		}
 	}
 
 	/* validate volume size */
@@ -1467,10 +1197,154 @@ int main(int argc, char **argv)
 			       "%"PRIu64" blocks\n", fs->fs_blocks,
 			       opts.num_blocks);
 	}
+}
+
+int main(int argc, char **argv)
+{
+	errcode_t ret = 0;
+	ocfs2_filesys *fs = NULL;
+	int open_flags;
+	int upd_label = 0;
+	int upd_uuid = 0;
+	int upd_slots = 0;
+	int upd_jrnls = 0;
+	int upd_blocks = 0;
+	int upd_mount = 0;
+	int upd_incompat = 0;
+	int upd_backup_super = 0;
+	int upd_feature = 0;
+	uint16_t max_slots;
+	uint64_t old_blocks = 0;
+
+	initialize_ocfs_error_table();
+	initialize_o2dl_error_table();
+	initialize_o2cb_error_table();
+
+	setbuf(stdout, NULL);
+	setbuf(stderr, NULL);
+
+	memset (&opts, 0, sizeof(opts));
+
+	get_options(argc, argv);
+
+	if (signal(SIGTERM, handle_signal) == SIG_ERR) {
+		com_err(opts.progname, 0, "Could not set SIGTERM");
+		exit(1);
+	}
+
+	if (signal(SIGINT, handle_signal) == SIG_ERR) {
+		com_err(opts.progname, 0, "Could not set SIGINT");
+		exit(1);
+	}
+
+	open_flags = OCFS2_FLAG_HEARTBEAT_DEV_OK;
+	if (opts.queryfmt)
+		open_flags |= OCFS2_FLAG_RO;
+	else
+		open_flags |= OCFS2_FLAG_RW | OCFS2_FLAG_STRICT_COMPAT_CHECK;
+
+	ret = ocfs2_open(opts.device, open_flags, 0, 0, &fs); //O_EXCL?
+	if (ret) {
+		com_err(opts.progname, ret, "while opening device %s",
+			opts.device);
+		goto close;
+	}
+	fs_gbl = fs;
+
+	if (opts.queryfmt) {
+		print_query(opts.queryfmt);
+		goto close;
+	}
+
+	if (OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat &
+	    OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV) {
+		com_err(opts.progname, 0, "Heartbeat devices cannot be tuned, "
+			"only re-formatted using mkfs.ocfs2");
+		goto close;
+	}
+
+	if (OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat &
+	    OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG) {
+		com_err(opts.progname, 0, "Aborted resize detected. "
+			"Run fsck.ocfs2 -f <device>.\n");
+		goto close;
+	}
+
+	if (OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat &
+	    OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG) {
+		fprintf(stderr, "Aborted tunefs operation detected. "
+			"Run fsck.ocfs2 -f <device>.\n");
+		goto close;
+	}
+
+	if (resize)
+		get_vol_size(fs);
+
+	if (!ocfs2_mount_local(fs)) {
+		ret = o2cb_init();
+		if (ret) {
+			com_err(opts.progname, ret, "while initializing the cluster");
+			exit(1);
+		}
+
+		ret = ocfs2_initialize_dlm(fs, WHOAMI);
+		if (opts.update_cluster) {
+			/* We have the right cluster, do nothing */
+			if (!ret)
+				goto close;
+			if (ret == O2CB_ET_INVALID_STACK_NAME) {
+				/* We expected this - why else ask for
+				 * a cluster information update? */
+				printf("Updating on-disk cluster information "
+				       "to match the running cluster.\n"
+				       "DANGER: YOU MUST BE ABSOLUTELY "
+				       "SURE THAT NO OTHER NODE IS USING "
+				       "THIS FILESYSTEM BEFORE MODIFYING "
+				       "ITS CLUSTER CONFIGURATION.\n");
+				goto skip_cluster_start;
+			}
+		}
+		if (ret) {
+			com_err(opts.progname, ret, "while initializing the dlm");
+			goto close;
+		}
+
+		block_signals(SIG_BLOCK);
+		ret = ocfs2_lock_down_cluster(fs);
+		block_signals(SIG_UNBLOCK);
+		if (!ret)
+			cluster_locked = 1;
+		else if (ret == O2DLM_ET_TRYLOCK_FAILED && resize) {
+			/*
+			 * We just set the flag here and more check and
+			 * lock will be done later.
+			 */
+			online_resize = 1;
+		} else {
+			com_err(opts.progname, ret, "while locking down the cluster");
+			goto close;
+		}
+	}
+
+skip_cluster_start:
+	/*
+	 * We will use block cache in io. Now whether the cluster is locked or
+	 * the volume is mount local, in both situation we can safely use cache.
+	 * If io_init_cache failed, we will go on the tunefs work without
+	 * the io_cache, so there is no check here.
+	 */
+	io_init_cache(fs->fs_io, ocfs2_extent_recs_per_eb(fs->fs_blocksize));
+
+	ret = volume_check(fs);
+	if (ret)
+		goto unlock;
+
+	validate_parameter(fs);
 
 	if (!opts.vol_label && !opts.vol_uuid && !opts.num_slots &&
 	    !opts.jrnl_size && !opts.num_blocks && !opts.mount &&
-	    !opts.backup_super) {
+	    !opts.backup_super && !opts.list_sparse && !opts.feature_string &&
+	    !opts.update_cluster) {
 		com_err(opts.progname, 0, "Nothing to do. Exiting.");
 		goto unlock;
 	}
@@ -1484,11 +1358,42 @@ int main(int argc, char **argv)
 		}
 	}
 
+	/*
+	 * We handle online resize seperately here, since it is
+	 * not like tunefs operations.
+	 */
+	if (online_resize) {
+		ret = online_resize_lock(fs);
+		if (ret)
+			goto close;
+
+		ret = update_volume_size(fs, &upd_blocks, 1);
+		if (ret) {
+			com_err(opts.progname, ret,
+				"while updating volume size");
+			goto online_resize_unlock;
+		}
+		if (upd_blocks)
+			printf("Resized volume\n");
+
+		goto online_resize_unlock;
+	}
+
 	/* Set resize incompat flag on superblock */
-	if (opts.num_blocks) {
-		OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat |=
-			OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG;
-		ret = ocfs2_write_super(fs);
+	max_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
+	if (opts.num_blocks ||
+	    (opts.num_slots && opts.num_slots < max_slots)) {
+		if (opts.num_blocks)
+			OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat |=
+				OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG;
+		else {
+			OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat |=
+				OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG;
+			OCFS2_RAW_SB(fs->fs_super)->s_tunefs_flag |=
+				OCFS2_TUNEFS_INPROG_REMOVE_SLOT;
+		}
+
+		ret = ocfs2_write_primary_super(fs);
 		if (ret) {
 			com_err(opts.progname, ret,
 				"while writing resize incompat flag");
@@ -1519,13 +1424,25 @@ int main(int argc, char **argv)
 				"while updating node slots");
 			goto unlock;
 		}
+		/* Clear remove slot incompat flag on superblock */
+		if (opts.num_slots < max_slots) {
+			OCFS2_RAW_SB(fs->fs_super)->s_feature_incompat &=
+				~OCFS2_FEATURE_INCOMPAT_TUNEFS_INPROG;
+			OCFS2_RAW_SB(fs->fs_super)->s_tunefs_flag &=
+				~OCFS2_TUNEFS_INPROG_REMOVE_SLOT;
+		}
 		if (upd_slots)
-			printf("Added node slots\n");
+			printf("Changed node slots\n");
 	}
 
 	/* change mount type */
 	if (opts.mount) {
-		update_mount_type(fs, &upd_mount);
+		ret = update_mount_type(fs, &upd_mount);
+		if (ret) {
+			com_err(opts.progname, ret,
+				"while changing mount type");
+			goto unlock;
+		}
 		if (upd_mount)
 			printf("Changed mount type\n");
 	}
@@ -1544,7 +1461,8 @@ int main(int argc, char **argv)
 
 	/* update volume size */
 	if (opts.num_blocks) {
-		ret = update_volume_size(fs, &upd_blocks);
+		old_blocks = fs->fs_blocks;
+		ret = update_volume_size(fs, &upd_blocks, 0);
 		if (ret) {
 			com_err(opts.progname, ret,
 				"while updating volume size");
@@ -1557,13 +1475,48 @@ int main(int argc, char **argv)
 			printf("Resized volume\n");
 	}
 
+	/*
+	 * list all the files in the volume, and since list_sparse
+	 * option can coexist with others, we jump to the end after
+	 * the work.
+	 */
+	if (opts.list_sparse) {
+		ret = list_sparse(fs);
+		if (ret) {
+			com_err(opts.progname, ret,
+				"while listing sparse files");
+			goto unlock;
+		}
+	}
+
+	/* change the feature in the super block.*/
+	if (opts.feature_string) {
+		ret = update_feature(fs);
+		if (ret) {
+			com_err(opts.progname, ret,
+				"while updating feature in super block");
+			goto unlock;
+		}
+		upd_feature = 1;
+	}
+
+	/* Update the cluster configuration */
+	if (opts.update_cluster) {
+		ret = update_cluster(fs);
+		if (ret) {
+			com_err(opts.progname, ret,
+				"while updating cluster information");
+			goto unlock;
+		}
+	}
+
 	/* update the backup superblock. */
 	if (opts.backup_super ||
 	    (opts.num_blocks &&
 	    OCFS2_HAS_COMPAT_FEATURE(OCFS2_RAW_SB(fs->fs_super),
 				     OCFS2_FEATURE_COMPAT_BACKUP_SB))) {
 		block_signals(SIG_BLOCK);
-		ret = update_backup_super(fs, opts.num_blocks);
+		ret = update_backup_super(fs, old_blocks, opts.num_blocks);
 		block_signals(SIG_UNBLOCK);
 		if (ret) {
 			com_err(opts.progname, ret,
@@ -1581,43 +1534,21 @@ int main(int argc, char **argv)
 
 	/* write superblock */
 	if (upd_label || upd_uuid || upd_slots || upd_blocks || upd_incompat ||
-	    upd_mount || upd_backup_super) {
+	    upd_mount || upd_backup_super || upd_feature) {
 		block_signals(SIG_BLOCK);
 		ret = ocfs2_write_super(fs);
 		if (ret) {
-			com_err(opts.progname, ret, "while writing superblock");
+			com_err(opts.progname, ret,
+				"while writing superblock(s)");
 			goto unlock;
 		}
 		block_signals(SIG_UNBLOCK);
-		printf("Wrote Superblock\n");
-
-		/* superblock's information has changed.
-		 * We need to synchronize the backup blocks if needed.
-		 * We also have to admit that if upd_backup_super is set,
-		 * there is no need to refresh the backups since they are
-		 * written above by update_backup_super.
-		 */
-		if (!upd_backup_super &&
-		    OCFS2_HAS_COMPAT_FEATURE(OCFS2_RAW_SB(fs->fs_super),
-					     OCFS2_FEATURE_COMPAT_BACKUP_SB)) {
-			block_signals(SIG_BLOCK);
-			ret = refresh_backup_super(fs);
-			block_signals(SIG_UNBLOCK);
-			if (ret) {
-				printf("Unable to refresh backup superblocks. "
-					"Please run fsck.ocfs2 before running "
-					"tunefs.ocfs2 to re-enable "
-					"backup superblocks.");
-				OCFS2_CLEAR_COMPAT_FEATURE(
-					OCFS2_RAW_SB(fs->fs_super),
-					OCFS2_FEATURE_COMPAT_BACKUP_SB);
-				block_signals(SIG_BLOCK);
-				ocfs2_write_super(fs);
-				block_signals(SIG_UNBLOCK);
-			}
-		}
+		printf("Wrote Superblock(s)\n");
 	}
 
+online_resize_unlock:
+	if (online_resize)
+		online_resize_unlock(fs);
 unlock:
 	block_signals(SIG_BLOCK);
 	if (cluster_locked && fs->fs_dlm_ctxt)
@@ -1626,17 +1557,17 @@ unlock:
 	block_signals(SIG_UNBLOCK);
 
 close:
+	if (fs && fs->fs_io)
+		io_destroy_cache(fs->fs_io);
 	block_signals(SIG_BLOCK);
 	if (fs && fs->fs_dlm_ctxt)
-		ocfs2_shutdown_dlm(fs);
+		ocfs2_shutdown_dlm(fs, WHOAMI);
 	block_signals(SIG_UNBLOCK);
 
-	if (opts.vol_uuid)
-		free(opts.vol_uuid);
-	if (opts.vol_label)
-		free(opts.vol_label);
-	if (opts.device)
-		free(opts.device);
+	free_clear_ctxt();
+
+	free_opts();
+
 	if (fs)
 		ocfs2_close(fs);
 
